
; Assembly Editor mode flags

AEMODE_OVERWRITE   = 1
AEMODE_VERTICALSEL = 2

; Assembly Editor search flags

AEFIND_CASESENSITIVE = 1
AEFIND_WHOLEWORDS    = 2
AEFIND_BACKWARD      = 4

; Assembly Editor styles

AES_AUTOINDENT   = 0001h
AES_AUTOBRACKETS = 0002h
AES_SMARTTABS    = 0004h
AES_SECURESEL    = 0008h
AES_OPTIMALFILL  = 0010h
AES_CONSOLECARET = 0020h

; Assembly Editor messages

AEM_SETMODE            = WM_USER + 0
AEM_GETMODE            = WM_USER + 1
AEM_SETPOS             = WM_USER + 2
AEM_GETPOS             = WM_USER + 3
AEM_SETSYNTAXHIGHLIGHT = WM_USER + 4
AEM_SETRIGHTCLICKMENU  = WM_USER + 5
AEM_SETTEXTCOLOR       = WM_USER + 6
AEM_SETSELCOLOR        = WM_USER + 7
AEM_FINDFIRST          = WM_USER + 8
AEM_FINDNEXT           = WM_USER + 9
AEM_CANFINDNEXT        = WM_USER + 10
AEM_GETLINELENGTH      = WM_USER + 11
AEM_GETLINE            = WM_USER + 12
AEM_GETWORDATCARET     = WM_USER + 13

AEM_GETMODIFIED        = WM_USER + 14
AEM_SETMODIFIED        = WM_USER + 15
AEM_GETCARETXY         = WM_USER + 16
AEM_SETREADONLY        = WM_USER + 17
AEM_GETREADONLY        = WM_USER + 18
AEM_SETLINEOFFSET      = WM_USER + 19
AEM_GETLINEOFFSET      = WM_USER + 20
AEM_SETFLAGS           = WM_USER + 21
AEM_GETFLAGS           = WM_USER + 22
AEM_NEXTBOOKMARK       = WM_USER + 23
AEM_PREVBOOKMARK       = WM_USER + 24
AEM_COMMENT            = WM_USER + 25
AEM_INDENT             = WM_USER + 26
AEM_SETTHEME           = WM_USER + 27
AEM_SETICONS           = WM_USER + 28
AEM_CLEARALLFLAGS      = WM_USER + 29
AEM_CLEARACTIVELINE    = WM_USER + 30

; Assembly Editor notifications

AEN_SETFOCUS    = 01h
AEN_KILLFOCUS   = 02h
AEN_TEXTCHANGE  = 03h
AEN_POSCHANGE   = 04h
AEN_MODECHANGE  = 05h
AEN_OUTOFMEMORY = 0Fh

; Assembly Editor position structure

struct AEPOS
  selectionPosition dd ?
  selectionLine     dd ?
  caretPosition     dd ?
  caretLine         dd ?
ends

struct AECARETXY
  x0 dd ?
  y0 dd ?
  x1 dd ?
  y1 dd ?
ends

; Assembly Editor segment structure

struct LINE_SEGMENT
  Next         dd ?
  Prev         dd ?
  r1           dd ?
  r2           dd ?
  Flags        dd ? ; bookmark, breakpoint etc. flags
  Offset       dd ? ; offset from compilation
ends

; Assembly Editor segment flags

FLAG_BOOKMARK   = 01h   ; the line is bookmarked
FLAG_BREAKPOINT = 02h   ; the line is a breakpoint
FLAG_READONLY   = 04h   ; the line is read-only
FLAG_ACTIVE     = 08h   ; highlight line as active / mark as an error

; Assembly Editor internal constants

SEGMENT_LENGTH        = 100h
BLOCK_LENGTH          = 100h * SEGMENT_LENGTH
SEGMENT_HEADER_LENGTH = 24
SEGMENT_DATA_LENGTH   = SEGMENT_LENGTH - SEGMENT_HEADER_LENGTH

struc AETHEME tfg,tbg,sfg,sbg,mbg,mfg,mw,focusfg,focusbg,synproc,[syn]
{
   common
     .text_color        dd tfg
     .text_background   dd tbg
     .sel_text          dd sfg
     .sel_background    dd sbg

     .margin_background dd mbg
     .margin_border     dd mfg
     .margin_width      dd mw

     .active_color       dd focusfg
     .active_background  dd focusbg

     .syntax_proc       dd synproc
     .syntax_colors:
   forward
     if ~ syn eq
       dd syn
     end if
   common
     .syntax_colors_count = ($ - .syntax_colors) / 4
 }

macro Theme fontName, fontSize, fontCharset, asmeditStyle,      \
            tfg,tbg,sfg,sbg,mbg,mfg,mw,focusfg,focusbg,[syn]
{
  common
    local ..fontN, ..theme, ..beg, ..size
  ..beg:
    dd ..fontN
    dd fontSize
    dd fontCharset
    dd asmeditStyle
    dd ..theme
    dd ..size
  common
    ..fontN db fontName, 0
            rb 256 - ($-..fontN)
    ..theme AETHEME tfg, tbg, sfg, sbg, mbg, mfg, mw, focusfg, focusbg, 0, syn
  ..size = $ - ..beg
}


struct AEICONS
  hBookmarkIcon   dd ?
  hActiveIcon     dd ?
  hBreakpointIcon dd ?
ends

; AsmEdit offsets in window extra memory.
ofsAsmEditMemory     = 0
ofsOpenFileStruct    = $c


