;------------------------------------------------------
; proc JumpTo
; Searches for the message in the message table.
; Arguments:
;   ebx - message number to search in the table.
; Returns:
;   This procedure doesn't returns. It simply
;   jumps to the address from table or imediately
;   after the table if the message is not found.
;   in both cases esi = eip
; Uses:
;   eax, esi
;------------------------------------------------------
proc JumpTo
        pop     esi     ; get table address and remove return
                        ; address from the stack.
.loop:
        lodsd
        test    eax,eax
        jz      .exit

        cmp     bx, ax
        jne     .loop

        sar     eax, 16
        add     esi, eax
.exit:
        jmp     esi
endp



;-------------------------------------------------
; Process all messages in the queue
; then exits. if WM_QUIT is accepted, exits
; directly from application.
;
; This procedure should be called during any
; long process to avoid hanging of application.
;
; Returns CF = 0 if message queue is empty
; Returns CF = 1 is WM_QUIT message was detected.
;         in this case, ProcessMessages also returns
;         the exit code in register eax.
;-------------------------------------------------
proc ProcessMessages
local .msg:MSG
        push    esi ebx

        lea     esi, [.msg]

.msg_loop:
        xor     eax, eax
        invoke  PeekMessage, esi, eax, eax, eax, PM_REMOVE     ; peek message from queue if any.
        test    eax,eax
        jz      .finish                                 ; exit.

        cmp     [.msg.message], WM_QUIT ; if message is WM_QUIT then quit.
        je      .terminate

; Process keyboard acceleration table if any...
.accel:
if defined AppLib
        invoke  SendMessage, [hApplication], AM_TRANSACCEL, esi, 0
        test    eax,eax
        jnz     .msg_loop
else
  if (defined hAccelerators) & (defined hMainForm)
          invoke  IsWindowEnabled, [hMainForm]
          test    eax, eax
          jz      .dialog

          invoke  TranslateAccelerator, [hMainForm], [hAccelerators], esi
          test    eax,eax
          jnz     .msg_loop
  end if
end if

; Processing of standard dialog messages (standard navigation: Tab, shift Tab etc.)
.dialog:
        mov     eax,[gui.modal_window]          ; 12.10.2005 - Mateusz Tymek
        or      eax,eax                         ; a bit different way of handling
        jz      .is_dialog_message              ; dialog messages, with global
        invoke  IsDialogMessage,eax,esi         ; variable
        test    eax,eax
        jnz     .msg_loop
.is_dialog_message:
if defined hPropEditor
        invoke  GetForegroundWindow                   ;
        cmp     eax,[hPropEditor]                     ;  this is temporary solution
        jne     .translate                            ;  proably whole message processing
        invoke  IsDialogMessage,eax,esi               ;  will have to be moved outside
        or      eax,eax                               ;  this library
        jnz     .msg_loop                             ;

;else
;  if defined hMainForm
;        invoke  GetLastActivePopup, [hMainForm]
;        invoke  IsDialogMessage, eax, esi
;        test    eax, eax
;        jnz     .msg_loop
;  end if
end if

.translate:
        invoke  TranslateMessage, esi                           ; So, standard processing.
        invoke  DispatchMessage, esi
        jmp     .msg_loop

.terminate:
        mov     eax, [.msg.wParam]
        stc
        pop     ebx esi
        ret

.finish:
        clc
        pop     ebx esi
        ret
endp
