; Common properties names.
iglobal
  property propAlign, 'Align'
  property propAutoSize, 'AutoSize'
  property propOwnFont, 'OwnFont'
  property propOldWinProc, 'OldWinProc'
  property propUserWinProc, 'UserWinProc'
endg

;initialize CreateCommonProperties
;begin
;        stdcall CreateProperty, propAlign
;        stdcall CreateProperty, propAutoSize
;        stdcall CreateProperty, propOwnFont
;        stdcall CreateProperty, propOldWinProc
;        stdcall CreateProperty, propUserWinProc
;        return
;endp
;
;finalize DeleteCommonProperties
;begin
;        stdcall DestroyProperty, propAlign
;        stdcall DestroyProperty, propAutoSize
;        stdcall DestroyProperty, propOwnFont
;        stdcall DestroyProperty, propOldWinProc
;        stdcall DestroyProperty, propUserWinProc
;        return
;endp


;*********************************************************************************************
; CreateForm
; Creates form window with all children from template.
; Arguments:
;   ptrTemplate - pointer to the form template. See TWinTemplate structure and
;                 WinTemplate macro in 'form.inc' for details.
; Returns:
;   EBX - handle of the created form. It's more useful to use ebx for this purpose to
;         allow using of folowing sequence without eax preserving:
;
;         stdcall CreateForm, xxxx
;         stdcall ShowModal, ebx
;         ... some API calls ...
;         invoke DestroyWindow, ebx
;
;*********************************************************************************************
proc CreateForm .ptrTemplate, .hOwner
        push    esi edi

        mov     esi, [.ptrTemplate]
        stdcall _DoCreateForm, [.hOwner]
        invoke  SendMessage, ebx, FM_AFTERCREATE, 0, 0

        pop     edi esi
        ret
endp

;----- internal called from create form ------------------------------------

proc _DoCreateForm .hParent

local .hwnd:DWORD
local .style:DWORD

        push    edi

.loop:
        push    [esi+TWinTemplate.WinProc]      ; ???
        push    [hInstance]
        movzx   eax, [esi+TWinTemplate.ID]
        movsx   edx, [esi+TWinTemplate.height]
        movsx   ecx, [esi+TWinTemplate.width]
        push    eax
        push    [.hParent]
        push    edx
        push    ecx

        movsx   edx, [esi+TWinTemplate.top]
        movsx   eax, [esi+TWinTemplate.left]
        push    edx
        push    eax

        mov     eax, [esi+TWinTemplate.style]
        mov     [.style], eax
        test    eax, WS_CHILD
        jnz     .styleok
        and     eax, not WS_VISIBLE
.styleok:
        push    eax

        push    [esi+TWinTemplate.ptrText]
        push    [esi+TWinTemplate.ptrClass]
        push    [esi+TWinTemplate.styleEx]
        invoke  CreateWindowEx

        mov     [.hwnd], eax
        cmp     [esi+TWinTemplate.WinProc], 0
        je      .subclassok

        stdcall SubclassWindow, eax, [esi+TWinTemplate.WinProc]

.subclassok:
        movzx   eax, [esi+TWinTemplate.Align]
        stdcall SetAlign, [.hwnd], eax

        invoke  SendMessage, [.hParent], WM_GETFONT, 0, 0
        stdcall ControlSetFont, [.hwnd], eax

        movzx   edi, [esi+TWinTemplate.flags]
        add     esi, sizeof.TWinTemplate

        test    edi, wtParent
        jz      @f

        stdcall _DoCreateForm, [.hwnd]

@@:
        test    [.style], WS_CHILD
        jnz     @f

        test    [.style], WS_VISIBLE
        jz      @f

        invoke  ShowWindow, [.hwnd], SW_SHOWNORMAL
@@:
        test    edi, wtEnd
        jz      .loop

        mov     ebx, [.hwnd]
        pop     edi
        ret
endp



proc SubclassWindow .hwnd, .ptrUserProc
        invoke  GetWindowLong, [.hwnd], GWL_WNDPROC
        invoke  SetProp, [.hwnd], [propOldWinProc], eax
        invoke  SetProp, [.hwnd], [propUserWinProc], [.ptrUserProc]
        invoke  SetWindowLong, [.hwnd], GWL_WNDPROC, UniversalWinProc
        invoke  SendMessage, [.hwnd], AM_INITWINDOW, 0, 0
        ret
endp


proc UnSubclassWindow .hwnd
        invoke  GetProp, [.hwnd], [propOldWinProc]
        test    eax, eax
        jz      .exit

        invoke  SetWindowLong, [.hwnd], GWL_WNDPROC, eax
        invoke  RemoveProp, [.hwnd], propOldWinProc
        invoke  RemoveProp, [.hwnd], propUserWinProc
.exit:
        ret
endp


proc UniversalWinProc .hwnd, .wmsg, .wparam, .lparam
;        mov eax,[debug_var]
;        push    [.wmsg]
;        pop     dword[debug_history+eax*4]
;        inc [debug_var]
        invoke  GetProp, [.hwnd], [propUserWinProc]
        test    eax, eax
        jz      .oldproc

        push    ebx esi edi
        mov     ebx, [.wmsg]
        stdcall eax, [.hwnd], ebx, [.wparam], [.lparam]
        pop     edi esi ebx
        jnc     .finish

.oldproc:
        invoke  GetProp, [.hwnd], [propOldWinProc]
        invoke  CallWindowProc, eax ,[.hwnd],[.wmsg],[.wparam],[.lparam]

.finish:
;        dec [debug_var]
        ret
endp


;------------------------------------------------------------
; Set new Align for the window.
; Notifies the window about changing with AM_ALIGNCHANGED
; Realigns the parent childrens.
; Returns old align
;------------------------------------------------------------
proc SetAlign .hwnd, .align
        invoke  GetProp, [.hwnd], [propAlign]
        cmp     eax, [.align]
        je      .exit

        push    eax

        invoke  SetProp, [.hwnd], [propAlign], [.align]
        invoke  SendMessage, [.hwnd], AM_ALIGNCHANGED, [.align], 0

        stdcall AlignSiblings, [.hwnd]
        pop     eax
.exit:
        ret
endp



;----------------------------------------------------------------------
; Set the font of some window if propOwnFont = FALSE
;----------------------------------------------------------------------
proc ControlSetFont .hwnd, .hFont
        invoke  GetProp, [.hwnd], [propOwnFont]
        test    eax, eax
        jnz     .exit

        invoke  SendMessage, [.hwnd], WM_SETFONT, [.hFont], TRUE
        mov     eax, TRUE

.exit:
        ret
endp


; Various property related string procedures
proc CreateProperty .ptrProp
        push    ebx
        mov     ebx, [.ptrProp]
        invoke  IsBadReadPtr, ebx, 4
        test    eax, eax
        jnz     .exit

        invoke  IsBadStringPtr, [ebx], 256
        test    eax, eax
        jnz     .exit

        invoke  AddAtom, [ebx]
        movzx   eax, ax
        mov     [ebx], eax
.exit:
        pop     ebx
        ret
endp


proc DestroyProperty .ptrProp
        push    ebx
        mov     ebx, [.ptrProp]
        invoke  IsBadReadPtr, ebx, 4
        test    eax, eax
        jnz     .exit

        invoke  DeleteAtom, [ebx]
.exit:
        pop     ebx
        ret
endp