TFormLib:
;************************************************************
;*   Form class library                                     *
;*                                                          *
;*   This file is part of Fresh base library                *
;*   This is base window class for "Fresh" created project  *
;*                                                          *
;*   (C)2003, 2004 John Found                               *
;************************************************************

;*********************************************************************************************
; Registers form window class
; Call only once
;*********************************************************************************************
uglobal
  gui.modal_window dd ?
endg

initialize RegisterFormClass

local .wc:WNDCLASS

        xor     eax, eax
        lea     edi, [.wc]
        mov     ecx, sizeof.WNDCLASS / 4
        rep stosd
        invoke  LoadCursor, eax, IDC_ARROW
        mov     [.wc.hCursor],eax
        mov     [.wc.style],CS_OWNDC
        mov     [.wc.lpfnWndProc],FormWinProc
        mov     [.wc.cbWndExtra], 32
        mov     eax,[hInstance]
        mov     [.wc.hInstance],eax
        mov     [.wc.lpszClassName],cFormClassName
        lea     eax, [.wc]
        invoke  RegisterClass, eax
        mov     [gui.modal_window],0

        ret
endp


iglobal
  cFormClassName   db      'TForm', 0
  property propLastFocused, 'LastFocused'
  property propColor, 'Color'
  property propModalResult, 'ModalResult'
  property propFont, 'Font'
endg

uglobal
  hLastBeforeMenu dd ?
endg


;initialize CreateFormProperties
;begin
;        stdcall CreateProperty, propLastFocused
;        stdcall CreateProperty, propColor
;        stdcall CreateProperty, propModalResult
;        stdcall CreateProperty, propFont
;        return
;endp


;finalize DeleteFormProperties
;begin
;        stdcall DestroyProperty, propLastFocused
;        stdcall DestroyProperty, propColor
;        stdcall DestroyProperty, propModalResult
;        stdcall DestroyProperty, propFont
;        return
;endp


;*********************************************************************************************
; Main window procedure for 'TForm' window class.
;*********************************************************************************************
winproc FormWinProc
        push    ebx edi esi
        mov     ebx, [wmsg]

ondefault
        invoke  DefWindowProc, [hwnd],[wmsg],[wparam],[lparam]
        pop     esi edi ebx
        ret

;------------------------------------------------------------------------
onmessage FM_SETCOLOR
        mov      eax, [wparam]
        cmp      eax, $ff000000
        jne      .setcolor

        invoke  RemoveProp, [hwnd], [propColor]
        pop     esi edi ebx
        ret

.setcolor:
        or       eax, $ff000000
        invoke   SetProp, [hwnd], [propColor], eax
        invoke   InvalidateRect, [hwnd], NULL, TRUE
        pop     esi edi ebx
        ret

;------------------------------------------------------------------------
onmessage WM_SYSCOMMAND
        mov     eax, [wparam]
        and     eax, $fff0
        cmp     eax, SC_MINIMIZE
        je      .savestate

        cmp     eax, SC_KEYMENU
        jne     .ondefault

        invoke  GetParent, [hwnd]
        mov     ebx, eax
        test    eax, eax
        jnz     .resendmessage

        invoke  GetMenu, [hwnd]
        test    eax, eax
        jnz     .ondefault

; Get the main application window.
if defined AppLib
        invoke  GetProp, [hApplication], [propMainForm]
        test    eax, eax
        jz      .ondefault

        cmp     eax, [hwnd]
        je      .ondefault

        mov     ebx, eax
else
  if defined hMainForm
        mov ebx, [hMainForm]
        cmp ebx, [hwnd]
        je  .ondefault
  else
        xor ebx, ebx
  end if
end if

        invoke  GetFocus
        mov     [hLastBeforeMenu], eax

        invoke  SetFocus, ebx

.resendmessage:
        invoke  SendMessage, ebx, [wmsg], [wparam], [lparam]
        xor     eax, eax
        pop     esi edi ebx
        ret

;-------------------------------------------------
onmessage WM_ACTIVATE
        cmp     word [wparam], WA_INACTIVE
        jne     .ondefault

; Save the last focused window
.savestate:
        stdcall GetFocusedControl, [hwnd]
        test    eax, eax
        jz      .ondefault

        invoke  SetProp, [hwnd], [propLastFocused], eax
        jmp     .ondefault

;----------------------------------------------------------------
onmessage WM_SETFOCUS
; Restore the last focused window
        invoke  GetProp, [hwnd], [propLastFocused]
        test    eax, eax
        jnz     .focusit

        invoke  GetNextDlgTabItem, [hwnd], NULL, FALSE
        test    eax, eax
        jz      .endsetfocus

        push    eax
        invoke  SetProp, [hwnd], [propLastFocused], eax
        pop     eax

.focusit:
        invoke  SetFocus, eax

.endsetfocus:
        xor     eax, eax
        pop     esi edi ebx
        ret

;-------------------------------------------------------------------------
;onmessage WM_NEXTDLGCTL
;;          int3
;        cmp     [lparam], FALSE
;        je      .nextprev
;
;        invoke  SetFocus, [wparam]
;        xor     eax, eax
;        pop     esi edi ebx
;        return
;
;.nextprev:
;        stdcall GetFocusedControl, [hwnd]
;        test    eax, eax
;        jz      .focusok
;
;        invoke  GetNextDlgTabItem, [hwnd], eax, [wparam]
;        invoke  SetFocus, eax
;
;.focusok:
;        pop     esi edi ebx
;        return

;------------------------------------------------------------------------
onmessage WM_ERASEBKGND
locals
  .rect    RECT
endl
        lea     eax, [.rect]
        invoke  GetClientRect, [hwnd], eax

        invoke  GetProp, [hwnd], [propColor]
        test    eax, eax
        jnz     .colorok

        invoke  GetSysColor, COLOR_BTNFACE

.colorok:
        and     eax, $ffffff
        invoke  CreateSolidBrush, eax
        push    eax
        lea     ecx, [.rect]
        invoke  FillRect, [wparam], ecx, eax
        invoke  DeleteObject ; from the stack.
        xor     eax, eax
        inc     eax
        pop     esi edi ebx
        ret

;------------------------------------------------------------------------
onmessage WM_COMMAND

        invoke  GetParent, [hwnd]
        test    eax, eax
        jz      .process

        invoke  SendMessage, eax, [wmsg], [wparam], [lparam]
        xor     eax, eax
        pop     esi edi ebx
        ret

.process:
        movzx   eax, word [wparam]
        cmp     eax, MODAL_MIN
        jb      .ondefault
        cmp     eax, MODAL_MAX
        ja      .ondefault

        invoke  SetProp, [hwnd], [propModalResult], eax

        xor     eax, eax
        pop     esi edi ebx
        ret

;------------------------------------------------------------------------
onmessage WM_CLOSE
if defined AppLib
        invoke  GetProp, [hApplication], [propMainForm]
        cmp     eax, [hwnd]
        jne     .ondefault

        invoke  SendMessage,  [hApplication], WM_CLOSE, 0, 0
        xor     eax, eax
        pop     esi edi ebx
        ret
else
  if defined hMainForm
        mov     eax, [hwnd]
        cmp     eax, [hMainForm]
        jne     .ondefault

        invoke  PostQuitMessage, 0
        jmp     .ondefault
  end if
end if


;------------------------------------------------------------------------
onmessage WM_EXITMENULOOP
        cmp     [hLastBeforeMenu], 0
        je      @f

        invoke  SetFocus, [hLastBeforeMenu]
@@:
        xor     eax, eax
        mov     [hLastBeforeMenu], eax
        pop     esi edi ebx
        ret

;------------------------------------------------------------------------
onmessage WM_WINDOWPOSCHANGED
        mov     eax, [lparam]
        mov     ecx, [eax+WINDOWPOS.flags]
        test    ecx, SWP_SHOWWINDOW
        jnz     .realign

        test    ecx, SWP_HIDEWINDOW
        jnz     .savestate

        test    ecx, SWP_NOSIZE
        jnz     .ondefault

.realign:
        stdcall AlignChildren, [hwnd]
        jmp     .ondefault

;------------------------------------------------------------------------


if defined GUILib
;------------------------------------------------------------------------
onmessage WM_MEASUREITEM
        stdcall OnMeasureItem,[hwnd],[lparam]
        mov     eax, TRUE
        pop     esi edi ebx
        ret

;------------------------------------------------------------------------
onmessage WM_MENUCHAR
        stdcall OnMenuChar, [wparam], [lparam]
        pop     esi edi ebx
        ret

;------------------------------------------------------------------------
onmessage WM_DRAWITEM
        stdcall OnDrawItem, [lparam]
        mov     eax, TRUE
        pop     esi edi ebx
        ret

end if

;------------------------------------------------------------------------
onmessage WM_GETFONT
        invoke  GetProp, [hwnd], [propFont]
        test    eax, eax
        jnz     @f

        invoke  GetStockObject, DEFAULT_GUI_FONT
@@:
        pop     esi edi ebx
        ret

;------------------------------------------------------------------------
onmessage WM_PARENTNOTIFY
        cmp     word [wparam], WM_CREATE
        jne     @f

        invoke  GetProp, [lparam], [propOwnFont]
        test    eax, eax
        jnz     @f

        invoke  SendMessage, [hwnd], WM_GETFONT, 0, 0
        invoke  SendMessage, [lparam], WM_SETFONT, eax, TRUE

@@:
        xor     eax, eax
        pop     esi edi ebx
        ret
endwp



proc GetFocusedControl hwnd
        push    esi ebx
        invoke  GetFocus
        mov     esi, eax

.parentloop:
        mov     ebx, eax
        invoke  GetParent, eax
        test    eax, eax
        jnz     .parentloop

;        xor     eax, eax     ; eax is zero here...
        cmp     ebx, [hwnd]
        jne     .endfocused
        mov     eax, esi
.endfocused:
        pop     ebx esi
        ret
endp


;*********************************************************************************************
; Shows given form in modal form.
; Returns modal result
; This is first version so behaviour is not very proper.
;*********************************************************************************************
proc ShowModal hwnd, .flags

local .rect:RECT

        push    esi edi ebx

        invoke  GetFocus
        push    eax      ; the parameter for SetFocus at the end of the procedure.

        lea     eax, [.rect]
        invoke  GetWindowRect, [hwnd], eax

        mov     esi, [.rect.left]
        mov     edi, [.rect.top]

        test    [.flags], MSF_HCENTER
        jz      @f

        invoke  GetSystemMetrics, SM_CXSCREEN
        sub     eax, [.rect.right]
        add     eax, [.rect.left]
        sar     eax,1
        mov     esi, eax
@@:
        test    [.flags], MSF_VCENTER
        jz      @f

        invoke  GetSystemMetrics, SM_CYSCREEN
        sub     eax, [.rect.bottom]
        add     eax, [.rect.top]
        sar     eax,1
        mov     edi, eax

@@:
        invoke  SendMessage, [hwnd], WM_INITDIALOG, 0, 0
        invoke  SetWindowPos, [hwnd], HWND_TOP, esi, edi, 0, 0, SWP_NOSIZE
        invoke  ShowWindow, [hwnd], SW_SHOW
;        invoke  SendMessage, [hwnd], AM_CHILDRESIZED, 0, 0

        push    [hwnd]
        pop     [gui.modal_window]

        invoke  GetWindow, [hwnd], GW_OWNER
        mov     ebx, eax

; Disables window owner and all other windows owned by window owner.

        invoke  EnableWindow, ebx, FALSE
        stdcall EnableAllOwnedWindows, ebx, FALSE       ; This is useless if all top level windows
                                                        ; are TForm, but it is important if not...

        invoke  EnableWindow, [hwnd], TRUE              ; IMPORTANT: re-enable modal window.

        invoke  GetNextDlgTabItem, [hwnd], NULL, FALSE
        test    eax, eax
        jz      @f
        invoke  SetFocus, eax
@@:
        invoke  SetProp, [hwnd], [propModalResult], MR_NONE

; Special message loop for modal window. It not finish until window becomes invisible.
.modalloop:
        call    ProcessMessages

        invoke  IsWindowVisible, [hwnd]
        test    eax,eax
        jz      .endloop

        invoke  GetProp, [hwnd], [propModalResult]
        test    eax,eax              ; mrNone = 0
        jnz     .endloop

        invoke  SendMessage, [hwnd], AM_ONIDLE, 0, 0

        invoke  WaitMessage
        jmp     .modalloop

.endloop:
        stdcall EnableAllOwnedWindows, ebx, TRUE
        invoke  EnableWindow, ebx, TRUE
        invoke  ShowWindow, [hwnd], SW_HIDE

        invoke  SetFocus                       ; the parameter is in stack

        invoke  GetProp, [hwnd], [propModalResult]
        mov     [gui.modal_window],0
        pop     ebx edi esi
        ret
endp


;-------------------------------------------------------------
; Enables or disables all windows owned by specifyed window.
;
; fEnable - TRUE = enable
;           FALSE = disable
;-------------------------------------------------------------
proc EnableAllOwnedWindows hwnd, .fEnable

        mov     eax, _EnableOwnedProc
        cmp     [.fEnable], FALSE
        jne     @f
        mov     eax, _DisableOwnedProc
@@:
        invoke  EnumWindows, eax, [hwnd]
        ret
endp


; callback procedure used in EnableAllOwnedWindows
proc _EnableOwnedProc hwnd, .owner

        invoke  GetWindow, [hwnd], GW_OWNER
        cmp     eax, [.owner]
        jne     .ready
        invoke  EnableWindow, [hwnd], TRUE

.ready:
        mov     eax, 1
        ret
endp


; callback procedure used in EnableAllOwnedWindows
proc _DisableOwnedProc hwnd, .owner
        invoke  GetWindow, [hwnd], GW_OWNER

        cmp     eax, [.owner]
        jne     .ready
        invoke  EnableWindow, [hwnd], FALSE
.ready:
        mov     eax, 1
        ret
endp




DispSize 'TForm class library', $ - TFormLib