;**********************************************************************
; Returns string with source code for given window Template and all its
; children.
; Arguments:
;       ptrWinTemplateDT - pointer to TWinTemplateDT structure.
;
; Returns:
;       eax - handle to the string with source code.
;**********************************************************************
proc CreateSourceTemplate .ptrWinTemplateDT
        stdcall StrNew
        push    eax
        stdcall _DoCreateSourceTemplate, [.ptrWinTemplateDT], eax, FALSE
        pop     eax
        ret
endp


iglobal
  cFormMacro  db 'Window ',0
  cFormComma  db ', ', 0
  cFormQuote  db "'",0
  cFormNULL   db '0',0
  cFormReturn db ';',13, 10,0
  cFormNameNone db 'NONE',0
endg

uglobal
  numbuff rb 256
endg


proc _DoCreateSourceTemplate .ptrWinTemplateDT, .res, .fNotLast
        push    esi edi ebx
        mov     esi, [.ptrWinTemplateDT]

; create the line for parent window
;-------------------------------------------------------
        stdcall StrCat, [.res], cFormMacro

        mov     eax, [esi+TWinTemplateDT.hName] ; Or maybe empty string?
        test    eax, eax
        jnz     @f
        mov     eax, cFormNameNone
@@:
        stdcall StrCat, [.res], eax
        stdcall StrCat, [.res], cFormComma

; Add wtXXX flags
        xor     eax, eax
        cmp     [esi+TWinTemplateDT.children.Count], 0
        je      @f
        or      eax, wtParent
@@:
        cmp     [.fNotLast], 0
        jne     @f
        or      eax, wtEnd
@@:
        stdcall NumToStr, eax, numbuff, 0, ntsDec or ntsUnsigned or ntsZeroTerminated
        stdcall StrCat, [.res], numbuff
        stdcall StrCat, [.res], cFormComma

; Add Align:
        movzx   eax, [esi+TWinTemplateDT.Align]
        stdcall NumToStr, eax, numbuff, 0, ntsDec or ntsUnsigned or ntsZeroTerminated
        stdcall StrCat, [.res], numbuff
        stdcall StrCat, [.res], cFormComma
;
; Add Class name
        stdcall StrCat, [.res], cFormQuote
        stdcall StrCat, [.res], [esi+TWinTemplateDT.ptrClass]
        stdcall StrCat, [.res], cFormQuote
        stdcall StrCat, [.res], cFormComma

; Add window text
        stdcall StrCat, [.res], cFormQuote
        stdcall StrCat, [.res], [esi+TWinTemplateDT.ptrText]
        stdcall StrCat, [.res], cFormQuote
        stdcall StrCat, [.res], cFormComma

; Add Style and StyleEx
        mov     [numbuff], '$'

        stdcall NumToStr, [esi+TWinTemplateDT.style], numbuff, 1, ntsHex or ntsUnsigned or ntsZeroTerminated
        stdcall StrCat, [.res], numbuff
        stdcall StrCat, [.res], cFormComma

        stdcall NumToStr, [esi+TWinTemplateDT.styleEx], numbuff, 1, ntsHex or ntsUnsigned or ntsZeroTerminated
        stdcall StrCat, [.res], numbuff
        stdcall StrCat, [.res], cFormComma

; Add ID
        movzx   eax, [esi+TWinTemplateDT.ID]
        stdcall NumToStr, eax, numbuff, 0, ntsDec or ntsUnsigned or ntsZeroTerminated
        stdcall StrCat, [.res], numbuff
        stdcall StrCat, [.res], cFormComma

; Add coordinates/size
        movsx   eax, [esi+TWinTemplateDT.left]
        stdcall NumToStr, eax, numbuff, 0, ntsDec or ntsZeroTerminated
        stdcall StrCat, [.res], numbuff
        stdcall StrCat, [.res], cFormComma

        movsx   eax, [esi+TWinTemplateDT.top]
        stdcall NumToStr, eax, numbuff, 0, ntsDec or ntsZeroTerminated
        stdcall StrCat, [.res], numbuff
        stdcall StrCat, [.res], cFormComma

        movsx   eax, [esi+TWinTemplateDT.width]
        stdcall NumToStr, eax, numbuff, 0, ntsDec or ntsZeroTerminated
        stdcall StrCat, [.res], numbuff
        stdcall StrCat, [.res], cFormComma

        movsx   eax, [esi+TWinTemplateDT.height]
        stdcall NumToStr, eax, numbuff, 0, ntsDec or ntsZeroTerminated
        stdcall StrCat, [.res], numbuff
        stdcall StrCat, [.res], cFormComma

; Add user WinProc
        mov     ecx, [esi+TWinTemplateDT.WinProc]
        stdcall StrLen, ecx
        test    eax, eax
        jnz     @f
        mov     ecx, cFormNULL
@@:
        stdcall StrCat, [.res], ecx
        stdcall StrCat, [.res], cFormReturn

; recurcive creation of children
        mov     ebx, [esi+TWinTemplateDT.children.Count]
        mov     esi, [esi+TWinTemplateDT.children.ptrItems]

.loop:
        dec     ebx
        js      .endloop
        stdcall _DoCreateSourceTemplate, [esi], [.res], ebx
        add     esi, 4
        jmp     .loop

.endloop:
        pop     ebx edi esi
        ret
endp




;------------------------------------
; Retrives the TBaseWin structure
; pointer from the given window.
; Returns NULL if the window have no
; atached such structure.
;------------------------------------
;proc GetBasePointer, hwnd
;begin
;        push    esi
;        invoke  GetWindowLong, [.hwnd], GWL_USERDATA
;        test    eax, eax
;        jz      .exit
;        mov     esi, eax
;        invoke  IsBadReadPtr, esi, sizeof.TAlignData
;        test    eax, eax
;        jnz     .false
;
;        mov     esi, [esi+TAlignData.UserData]
;        invoke  IsBadReadPtr, esi, sizeof.TBaseWin
;        test    eax, eax
;        jnz     .false
;
;        mov     eax, esi
;        pop     esi
;        return
;
;.false:
;        xor     eax, eax
;.exit:
;        pop     esi
;        return
;endp
;




;------------------------------------------------------------------------
; Create binary structure TWinTemplateDT structure from
; source text in FASM syntax.
;
; Arguments:
;   ptrSource: pointer to the source text.
;
; Returns:
;   eax: TWinTemplateDT structure containing all children etc.
;   edx: pointer to the next source line after the last processed.
;------------------------------------------------------------------------
proc CompileSourceTemplate .ptrSource
        push    esi edi ebx
        mov     esi, [.ptrSource]
        invoke  IsBadStringPtr, esi, -1
        test    eax, eax
        jnz     .errorformat

.compileloop:
        invoke  HeapAlloc, [hHeap], HEAP_ZERO_MEMORY, sizeof.TWinTemplateDT
        mov     edi, eax

        lea     eax, [edi+TWinTemplateDT.children]
        stdcall ListCreate, eax

        cmp     dword [esi], 'Wind'
        jne     .errorformat

        cmp     word [esi+4], 'ow'
        jne     .errorformat

        cmp     byte [esi+6], ' '
        jne     .errorformat

        add     esi, 7
        mov     ebx, esi        ; begin of the label

.search:
        call    SearchComma
        jc      .errorformat

.getname:
        mov     byte [esi-1], 0

        stdcall StrDup, ebx     ; creates string with name of the window.
        mov     [edi+TWinTemplateDT.hName], eax

; get TWinTemplateDT.flags value
        call    SearchNonSpace
        jc      .errorformat
        mov     ebx, esi
        call    SearchComma
        jc      .errorformat
        mov     byte [esi-1], 0

        stdcall StrToFasmNum, ebx
        jc      .errorformat

        mov     [edi+TWinTemplateDT.flags], al

; get TWinTemplateDT.Align value
        call    SearchNonSpace
        jc      .errorformat
        mov     ebx, esi
        call    SearchComma
        jc      .errorformat
        mov     byte [esi-1], 0

        stdcall StrToFasmNum, ebx
        jc      .errorformat

        mov     [edi+TWinTemplateDT.Align], al

; get TWinTemplateDT.ptrClass
        call    SearchNonSpace
        jc      .errorformat
        cmp     byte [esi], "'"
        jne     .errorformat

        inc     esi
        mov     ebx, esi
        call    SearchQuote
        jc      .errorformat

        mov     byte [esi-1], 0

        stdcall StrDup, ebx
        mov     [edi+TWinTemplateDT.ptrClass], eax

; get TWinTemplateDT.ptrText
        call    SearchQuote
        jc      .errorformat

        mov     ebx, esi
        call    SearchQuote
        jc      .errorformat

        mov     byte [esi-1], 0
        stdcall StrDup, ebx
        mov     [edi+TWinTemplateDT.ptrText], eax

; get TWinTemplateDT.style
        add     esi, 2
        call    SearchNonSpace
        jc      .errorformat

        mov     ebx, esi
        call    SearchComma
        jc      .errorformat

        mov     byte [esi-1], 0
        stdcall StrToFasmNum, ebx
        jc      .errorformat

        mov     [edi+TWinTemplateDT.style], eax

; get TWinTemplateDT.styleEx

        call    SearchNonSpace
        jc      .errorformat
        mov     ebx, esi

        call    SearchComma
        jc      .errorformat

        mov     byte [esi-1], 0
        stdcall StrToFasmNum, ebx
        jc      .errorformat

        mov     [edi+TWinTemplateDT.styleEx], eax

; get  TWinTemplateDT.ID

        call    SearchNonSpace
        jc      .errorformat
        mov     ebx, esi

        call    SearchComma
        jc      .errorformat

        mov     byte [esi-1], 0
        stdcall StrToFasmNum, ebx
        jc      .errorformat

        mov     [edi+TWinTemplateDT.ID], ax

; get  TWinTemplateDT.left
        call    SearchNonSpace
        jc      .errorformat
        mov     ebx, esi

        call    SearchComma
        jc      .errorformat

        mov     byte [esi-1], 0
        stdcall StrToFasmNum, ebx
        jc      .errorformat

        mov     [edi+TWinTemplateDT.left], ax

; get  TWinTemplateDT.top

        call    SearchNonSpace
        jc      .errorformat
        mov     ebx, esi

        call    SearchComma
        jc      .errorformat

        mov     byte [esi-1], 0
        stdcall StrToFasmNum, ebx
        jc      .errorformat

        mov     [edi+TWinTemplateDT.top], ax

; get  TWinTemplateDT.width

        call    SearchNonSpace
        jc      .errorformat
        mov     ebx, esi

        call    SearchComma
        jc      .errorformat

        mov     byte [esi-1], 0
        stdcall StrToFasmNum, ebx
        jc      .errorformat

        mov     [edi+TWinTemplateDT.width], ax

; get  TWinTemplateDT.height

        call    SearchNonSpace
        jc      .errorformat
        mov     ebx, esi

        call    SearchComma
        jc      .errorformat

        mov     byte [esi-1], 0
        stdcall StrToFasmNum, ebx
        jc      .errorformat

        mov     [edi+TWinTemplateDT.height], ax

; get TWinTemplateDT.WinProc, i.e. the label to the user procedure.
        stdcall StrNew
        mov     [edi+TWinTemplateDT.WinProc], eax

        call    SearchNonSpace
        jc      .errorformat
        cmp     byte [esi], '0'
        je      .nowinproc

        mov     ebx, esi
        call    SearchComment
        jc      .errorformat

        mov     byte [esi-1], 0
        stdcall StrCopy, [edi+TWinTemplateDT.WinProc], ebx

.nowinproc:
        call    SkipToNextLine  ; goto to the next line to get next window.

; the window template is created, so what to do?
        test    [edi+TWinTemplateDT.flags], wtParent
        jz      .childrenok

.childrenloop:
        stdcall CompileSourceTemplate, esi
        mov     esi, edx                     ; next line of source
        mov     ebx, eax                     ; TWinTemplateDT of child.

        lea     ecx, [edi+TWinTemplateDT.children]
        stdcall ListInsert, ecx, -1
        mov     [eax], ebx

        test    [ebx+TWinTemplateDT.flags], wtEnd
        jz      .childrenloop

.childrenok:
        mov     eax, edi
        mov     edx, esi

        pop     ebx edi esi
        ret

.errorformat:
        xor     eax, eax
        pop     ebx edi esi
        ret
endp


proc SearchComma
.search:
        lodsb
        cmp     al, ','
        je      .found

        cmp     al, $0d
        je      .error
        cmp     al, $0a
        je      .error
        cmp     al, ' '
        je      .error
        cmp     al, $09
        jne     .search

.error:
        stc
        ret

.found:
        clc
        ret
endp


proc SkipToNextLine
.searchloop:
        lodsb
        test    al,al
        jz      .error
        cmp     al, $0d         ; 0000 1101
        je      .found
        xor     al, $0a         ; 0000 1010
        jne     .searchloop
.found:
        xor     al, $07         ; CR <-> LF
        cmp     [esi], al
        jne     .exit
        inc     esi
.exit:
        clc
        ret

.error:
        stc
        dec     esi
        ret
endp


proc SearchNonSpace
.loop:
        lodsb
        test    al,al
        jz      .error
        cmp     al, $0d
        je      .error
        cmp     al, $0a
        je      .error

        cmp     al, ' '
        je      .loop
        cmp     al, $09
        je      .loop

        dec     esi
        clc
        ret

.error:
        dec     esi
        stc
        ret
endp


proc SearchQuote
.search:
        lodsb
        cmp     al, "'"
        je      .found

        cmp     al, $0d
        je      .error
        cmp     al, $0a
        jne     .search

.error:
        stc
        ret

.found:
        clc
        ret
endp

proc SearchComment
.search:
        lodsb
        cmp     al, ";"
        je      .found

        cmp     al, $0d
        je      .error
        cmp     al, $0a
        je      .error
        cmp     al, ' '
        je      .error
        cmp     al, $09
        jne     .search

.error:
        stc
        ret

.found:
        clc
        ret
endp

;--------------------------------------------------
; Creates the form from design window template
;--------------------------------------------------
proc CreateDesignWindow, .ptrTemplate
        push    esi
        mov     esi, [.ptrTemplate]
        cmp     [esi+TWinTemplateDT.hwnd], 0
        je      @f
        invoke  DestroyWindow, [esi+TWinTemplateDT.hwnd]
@@:
        cmp     [esi+TWinTemplateDT.hSizer], 0
        je      @f

; deal with sizer window

@@:
        mov     edx, [hApplication]
        mov     eax, [.ptrTemplate]
        mov     esi, [esi+TWinTemplateDT.parent]
        test    esi, esi
        jz      @f
        mov     edx, [esi+TWinTemplateDT.hwnd]
@@:
        stdcall _DoCreateDesignWindow, edx, [.ptrTemplate]
        push    eax

        mov     eax, [.ptrTemplate]
        stdcall GetSizer
        invoke  SendMessage, eax, SZM_UPDATESIZE, 0, 0

        mov     eax, [.ptrTemplate]
        stdcall ReorderChildren, [eax+TWinTemplateDT.parent]

        pop     eax
        pop     esi
        ret
endp


proc ReorderChildren .ptrTemplate
        push    esi edi ebx
        cmp     [.ptrTemplate], 0
        je      .finish

        mov     esi, [.ptrTemplate]
        mov     ebx, [esi+TWinTemplateDT.children.Count]
        mov     esi, [esi+TWinTemplateDT.children.ptrItems]

        mov     edi, HWND_BOTTOM

.loop:
        dec     ebx
        js      .endloop

        mov     eax, dword [esi]
        push    SWP_NOSIZE or SWP_NOMOVE
        push    0 0 0 0
        push    edi
        push    [eax+TWinTemplateDT.hwnd]
        mov     edi, [eax+TWinTemplateDT.hwnd]
        invoke  SetWindowPos
        add     esi, 4
        jmp     .loop

.endloop:
        mov     eax, [.ptrTemplate]
        stdcall AlignChildren, [eax+TWinTemplateDT.hwnd]

.finish:
        pop     ebx edi esi
        ret
endp



;----- internally called from CreateDesignWindow -----
proc _DoCreateDesignWindow .hParent, .ptrTemplate
        push    esi edi ebx

        mov     esi, [.ptrTemplate]
        movsx   eax, [esi+TWinTemplateDT.left]
        movsx   ebx, [esi+TWinTemplateDT.top]
        movsx   ecx, [esi+TWinTemplateDT.width]
        movsx   edx, [esi+TWinTemplateDT.height]
        movzx   edi, [esi+TWinTemplateDT.ID]

        push    NULL
        push    [hInstance]
        push    edi
        push    [.hParent]
        push    edx ecx ebx eax
        mov     eax, [esi+TWinTemplateDT.style]
        or      eax, WS_VISIBLE
        mov     ecx, [.hParent]
        cmp     ecx, [hApplication]
        jne     @f
        or      eax, WS_SIZEBOX
        and     eax, not WS_CHILD
@@:
        push    eax

        stdcall StrPtr, [esi+TWinTemplateDT.ptrText]
        push    eax
        stdcall StrPtr, [esi+TWinTemplateDT.ptrClass]
        push    eax
        push    [esi+TWinTemplateDT.styleEx]
        invoke  CreateWindowEx
        test    eax, eax
        jz      .errorcreate

        mov     [esi+TWinTemplateDT.hwnd], eax

; disable all internal windows
        invoke  EnumChildWindows, [esi+TWinTemplateDT.hwnd], procDisableCallback, 0

        cmp     [esi+TWinTemplateDT.parent], 0
        jne     .sizerok

; This is not proper way of handling sizer window in all cases.
; for example when there is already a sizer in the window and it
; have some windows selected.
        invoke  CreateWindowEx, WS_EX_TRANSPARENT, cSizerClassName, NULL,        \
                                WS_CHILD, 0, 0, 0, 0, [esi+TWinTemplateDT.hwnd], \
                                0, [hInstance], 0
        mov     [esi+TWinTemplateDT.hSizer], eax

.sizerok:
        stdcall SubclassWindow, [esi+TWinTemplateDT.hwnd], procDesignControl
        invoke  SetProp, [esi+TWinTemplateDT.hwnd], [propTemplateDT], esi
        movzx   eax, [esi+TWinTemplateDT.Align]
        stdcall SetAlign, [esi+TWinTemplateDT.hwnd], eax

        invoke  SendMessage, [.hParent], WM_GETFONT, 0, 0
        stdcall ControlSetFont, [esi+TWinTemplateDT.hwnd], eax

        cmp     [esi+TWinTemplateDT.children.Count], 0
        je      .childrenOK

        mov     ebx, [esi+TWinTemplateDT.children.Count]
        mov     edi, [esi+TWinTemplateDT.children.ptrItems]

.childrenloop:
        dec     ebx
        js      .childrenOK

        mov     eax, [edi]
        mov     [eax+TWinTemplateDT.parent], esi

        stdcall _DoCreateDesignWindow, [esi+TWinTemplateDT.hwnd], eax
        add     edi, 4
        jmp     .childrenloop

.childrenOK:
        stdcall SearchForDesignTimeInfo, [esi+TWinTemplateDT.ptrClass], [esi+TWinTemplateDT.style]
        mov     [esi+TWinTemplateDT.DesignInfo], eax

        mov     eax, esi
        pop     ebx edi esi
        ret

.errorcreate:
        stdcall LastError
        jmp     .childrenOK
endp


proc procDisableCallback .hwnd, .dummy
        invoke  EnableWindow, [.hwnd], FALSE
        mov     eax, 1
        ret
endp


proc CountDesignWindows .ptrWinTemplateDT
        push    esi edi ebx
        mov     eax, [.ptrWinTemplateDT]
        mov     ebx, [eax+TWinTemplateDT.children.Count]
        mov     esi, [eax+TWinTemplateDT.children.ptrItems]
        xor     edi, edi

.childrenloop:
        dec     ebx
        js      .endchildren
        stdcall CountDesignWindows, [esi+4*ebx]
        add     edi, eax
        jmp     .childrenloop

.endchildren:
        lea     eax, [edi+1]
        pop     ebx edi esi
        ret
endp


iglobal
  property propTemplateDT, 'TemplateDT'
endg