; This module is responsible for creating/reading/saving of common for
; the whole IDE parameters from the main .ini file.
;

iglobal
  cParamSection db 'IDE settings',0
  cDebugerKey db 'Debugger',0
  cHelpKey    db 'HelpFile',0

  cEditorSection  db 'Editor settings', 0
  cCodeCompletion db 'CodeCompletion', 0

  cErrStr2    db 'ERR!'
              db 0
endg

uglobal
  EditorOptions.fCodeCompletion dd ?
endg



proc CheckParams
local buf[1024]:BYTE
        invoke  SendMessage, [frmMsg], MWM_CLEARMESSAGES, 0, 0

; Check ini file consistency...
        stdcall StrPtr, [hIniFileName]
        mov     esi, eax

        stdcall FileExists, esi
        jnz     .doCheckParams

        stdcall FirstTimeInstall       ; try to create .ini file automatically

.doCheckParams:

        ; 14.10.2005 - Mateusz Tymek
        ;   When this lines are commented, Fresh won't complain about
        ;   unset paths to linux emulator, ext. debugger, etc.
        ;   I think that it is a better behaviour, as they are
        ;   only options.

;        stdcall CheckFileParam, cHelpKey, cHelpFileNotSet, cHelpFileNotExists
;        stdcall CheckFileParam, cDebugerKey, cDebuggerNotSet, cDebuggerNotExists
;        stdcall CheckFileParam, cLineKey, cLinePathNotSet, cLineNotExists

        stdcall CheckPaths
        stdcall StrPtr, [hIniFileName]
        invoke  GetPrivateProfileInt, cEditorSection,cCodeCompletion,1,eax
        mov     [EditorOptions.fCodeCompletion],eax
        ret

;  constStr cHelpFileNotExists, "Specified help file doesn't exists."
;  constStr cHelpFileNotSet, "No help file is set in IDE options."

;  constStr cDebuggerNotExists, "Specified external debugger doesn't exists."
;  constStr cDebuggerNotSet, "No external debugger is set in IDE options."

;  constStr cLineNotExists, "Specified Linux emulator doesn't exists."
;  constStr cLinePathNotSet, "No Linux emulator is set in IDE options."

endp



proc CheckFileParam .ptrKeyName, .msgNotSet, .msgNotExists
        stdcall GetParamFromINI, [.ptrKeyName]
        jc      .filenotset

        push    eax
        stdcall FileExists, eax
        stdcall StrDel ; from the stack

        test    eax, eax
        jz      .notexists
        ret

.notexists:
        stdcall AddCheckMessage, [.msgNotExists], mtError
        ret

.filenotset:
        stdcall AddCheckMessage, [.msgNotSet], mtWarning
        ret
endp




proc CheckPaths
        ret
endp



;-------------------------------------
; It tries to create proper ini file
; in the case that ini file does not
; exists.
;-------------------------------------
proc FirstTimeInstall
locals
.str rb 1024
.ptr dd ?
endl
        mov     ebx, cPathsTableInc

.findloop:
        mov     ecx, [ebx]
        test    ecx, ecx
        jz      .notfound1
        lea     edi, [.str]
        lea     eax, [.ptr]
        invoke  GetFullPathName, ecx, 1024, edi, eax
        stdcall FileExists, edi
        lea     ebx, [ebx+4]
        test    eax, eax
        jz      .findloop
        mov     eax, [.ptr]
        mov     byte [eax-1], 0
        invoke  WritePrivateProfileString, _section_environment, cIncludeKey, edi, esi
        lea     eax,[.str]
        mov     dword[eax],'1'
        invoke  WritePrivateProfileString, cEditorSection, cCodeCompletion, eax, esi
        mov     [EditorOptions.fCodeCompletion],1
.findmain:
        mov     ebx, cPathsTableMain
.findloop2:
        mov     ecx, [ebx]
        test    ecx, ecx
        jz      .notfound2

        lea     edi, [.str]
        lea     eax, [.ptr]
        invoke  GetFullPathName, ecx, 1024, edi, eax
        stdcall FileExists, edi
        lea     ebx, [ebx+4]
        test    eax, eax
        jz      .findloop2

        mov     eax, [.ptr]
        mov     byte [eax-1], 0
        invoke  WritePrivateProfileString, _section_environment, cFreshKey, edi, esi
        ret

.notfound1:
        invoke  MessageBox, [hApplication], cNotFoundMsg1, cNotFoundTtl, MB_ICONWARNING or MB_OK
        jmp     .findmain

.notfound2:
        invoke  MessageBox, [hApplication], cNotFoundMsg2, cNotFoundTtl, MB_ICONWARNING or MB_OK
        ret
endp


iglobal
  cPathsTableInc dd cAutoPath1, cAutoPath2, 0
  cAutoPath2  db '..\'
  cAutoPath1  db 'include'
  cFincTestFile db '\finc.txt',0

  cIncludeKey db 'finc',0

  cPathsTableMain dd cAutoPath3, cAutoPath4, 0
  cAutoPath3  db '..\'
  cAutoPath4  db 'Fresh.exe',0

  cFreshKey db 'Fresh',0

  cNotFoundMsg1 db "First time run procedure can't find include path automanically. Try to set it manualy in 'options|IDE params' menu."
                db "You have to use 'finc' name for this path.",0
  cNotFoundMsg2 db "First time run procedure can't find main program directory. Try to set it manually in 'options|IDE params' menu."
                db "You have to use 'Fresh' name for this path.",0
  cNotFoundTtl db 'Auto setup error',0
endg




proc AddCheckMessage .ptrMessage, .msgType
local .msg:TFreshMsg
        mov     eax, [.msgType]
        mov     ecx, [.ptrMessage]
        mov     [.msg.type], eax
        mov     [.msg.hMessage], ecx
        mov     [.msg.link.hFileName], 0

        lea     eax, [.msg]
        invoke  SendMessage, [frmMsg], MWM_ADDMESSAGE, eax, TVI_ROOT
        ret
endp




;------------------------------------------------
; Parces command line and opens the files.
; Still not parse the name with spaces.
;------------------------------------------------
proc ParseCommandLine
local .num:DWORD
        stdcall StrNew
        mov     ebx, eax

        invoke  GetCommandLine
        stdcall StrCopy, ebx, eax
        stdcall StrLen, ebx
        mov     ecx, eax

        stdcall StrPtr, ebx
        mov     edi, eax

        mov     [.num], 0
.foundspc:
        call    .skipspc
        call    .findquoted
        jc      .found

        call    .findnotquoted
        jnc     .finish

.found:
        inc     [.num]
        cmp     [.num], 1
        je      .foundspc
        push    ecx
        stdcall OpenFileByName, esi, NULL, TRUE
        pop     ecx
        jmp     .foundspc

.finish:
        stdcall StrDel, ebx
        ret

.findquoted:
        jecxz   .false
        mov     al, '"'
        cmp     [edi], al
        jne     .false

        inc     edi
        dec     ecx
        jecxz   .false

        mov     esi, edi
        repnz   scasb
        jnz     .false

        mov     byte [edi-1], 0
        stc
        retn

.false:
        clc
        retn

.findnotquoted:
        jecxz   .false
        mov     al, ' '

        mov     esi, edi
        repnz scasb
        jnz     .ok

        mov     byte [edi-1], 0
.ok:
        stc
        retn

.skipspc:
        cmp     byte [edi], ' '
        jne     .endskip

        inc     edi
        loop    .skipspc
.endskip:
        retn
endp


; These wrappers are only for better readability of the source in other modules...

proc GetExternalDebuggerName
        stdcall GetParamFromINI, cDebugerKey
        ret
endp

proc GetHelpFileName
        stdcall GetParamFromINI, cHelpKey
        ret
endp

proc GetLineFileName
        stdcall GetParamFromINI, cLineKey
        ret
endp


proc GetParamFromINI .ptrKeyName
local .buff[1025]:BYTE
        stdcall StrPtr, [hIniFileName]

        lea     ecx, [.buff]
        invoke  GetPrivateProfileString, cParamSection, [.ptrKeyName], cErrStr2, ecx, 1024, eax
        cmp     dword [.buff], 'ERR!'
        je      .error

        lea     eax, [.buff]
        stdcall StrDup, eax
        clc
        ret

.error:
        xor     eax, eax
        stc
        ret
endp
