
uglobal
  hProjManager    dd 0  ; handle of project manager
  ptrProjectFile  dd 0  ; pointer to TOpenFile with ftProject, currently open in the project manager.
  fProjModified   dd 0  ; flag "modified" for current open project.

  poMemory       dd 0  ; The memory for the compiler
  poBinNameAuto  dd 0  ; The flag whether the binary name will be auto created of fixed
  poBinName      dd 0  ; Handle to the binary name.

  imlCategories   dd 0  ; Image list for tree-view control

  ptrMRUProj      dd 0  ; pointer to MRU list for projects.

  hTreePopup      dd 0
  hListPopup      dd 0
endg


iglobal
  cProjTitle      db      'Project manager:', 0

  CoolMenu mnuProjTree,           \
    mfNormal, actAddCategory,     \
    mfSeparator, NONE,            \
    mfNormal, actMoveCategoryUp,  \
    mfNormal, actMoveCategoryDn,  \
    mfSeparator, NONE,            \
    mfNormal, actProjectOptions

  CoolMenu mnuProjList,         \
    mfNormal, actAddFile

endg


;----------------------------------------
; Creates project manager common objects.
; Call this only once.
;----------------------------------------
proc InitProjectManager
; Read project MRU list
        stdcall LoadMRUFromIni, 16, [hIniFileName], cMRUProjSection
        mov     [ptrMRUProj], eax

        stdcall ImageList_LoadGif, [hInstance], resBmpProjMan, 16, FALSE
        mov     [imlCategories], eax

        stdcall CreateCoolMenu, mnuProjTree, TRUE
        mov     [hTreePopup], eax

        stdcall CreateCoolMenu, mnuProjList, TRUE
        mov     [hListPopup], eax

        mov     [poBinNameAuto], TRUE
        mov     [poMemory], 16384

        stdcall StrNew
        mov     [poBinName], eax

        ret
endp


proc FinalizeProjectManager
        cmp     [ptrProjectFile], NULL
        je      .closeproj

        stdcall CloseFile, [ptrProjectFile]

.closeproj:
        stdcall SaveMRUToIni, [ptrMRUProj], [hIniFileName], cMRUProjSection
        stdcall DestroyMRUList, [ptrMRUProj]
        invoke  ImageList_Destroy, [imlCategories]

        invoke  DestroyMenu, [hTreePopup]
        invoke  DestroyMenu, [hListPopup]
        ret
endp


proc CreateProjectManager .ptrOpenFile, .hTemplate
        push    ebx esi
        xor     ebx, ebx

        mov     esi, [.ptrOpenFile]

        cmp     [ptrProjectFile], ebx
        je      .onlyone

        stdcall CloseFile, [ptrProjectFile]
        cmp     eax, cfrDestroyed
        jne     .fault

.onlyone:
        mov     [ptrProjectFile], esi

; Initial project options... they could be changed after CEM_LOADFILE
        mov     [poMemory], 16384
        mov     [poBinNameAuto], TRUE

        cmp     [poBinName], 0
        je      @f
        stdcall StrDel, [poBinName]
@@:
        stdcall StrNew
        mov     [poBinName], eax

        invoke   CreateWindowEx, WS_EX_TOOLWINDOW, cFormClassName, cProjTitle,  \
                                 WS_VISIBLE or WS_CAPTION or WS_CLIPCHILDREN or \
                                 WS_SIZEBOX or WS_SYSMENU,                      \
                                 ebx, ebx , 160, 400,                           \
                                 [hApplication], ebx, [hInstance], ebx

        mov     [hProjManager], eax
        mov     [esi+TOpenFile.hEditor], eax
        mov     [esi+TOpenFile.projectID], ebx
        stdcall SubclassWindow, eax, ProjManProc
        invoke  SendMessage, [hProjManager], CEM_LOADFILE, [.hTemplate], ebx
        push    eax
        stdcall RearrangeWindows
        pop     eax

.finish:
        pop     esi ebx
        ret

.fault:
        mov     eax, errCanceled
        pop     esi ebx
        ret
endp



iglobal
  msgPUC1 db 'The file: "',0
  msgPUC2 db '" is part of saved project, but it doesn''t exists on the disk.', 13, 10
          db 'Do you want to add it to the project anyway?',0
  msgPUCtitle db  'Error adding file to the project.',0
endg


proc ProjectUnConsistent .ptrOpenFile
        push    ebx
        mov     ebx, [.ptrOpenFile]

        stdcall StrDup, msgPUC1
        stdcall StrCat, eax, [ebx+TOpenFile.hFileName]
        stdcall StrCat, eax, msgPUC2
        push    eax
        stdcall StrPtr, eax

        invoke  MessageBox, [hApplication], eax, msgPUCtitle, MB_YESNO or MB_ICONWARNING
        stdcall StrDel ; from the stack
        pop     ebx
        ret
endp




;-----------------------------------------------------
; Dialog proc for Project manager.
;-----------------------------------------------------
winproc ProjManProc
locals
.tvi    TVINSERTSTRUCT
.lvi    LVITEM
.col    LVCOLUMN
.count  dd ?
.str    rb 1024
.str2   rb 256
.ptr    dd ?
.pnt    POINT
endl

ondefault
        stc
        ret


onmessage PMM_FILECHANGED

        stdcall SearchForFile, [hwnd], [wparam]
        cmp     eax, -1
        je      .endchanged

        push    eax

        invoke  SendMessage, [hwnd], PMM_REMOVEFILE, [wparam], 0
        invoke  SendMessage, [hwnd], PMM_ADDFILE, [wparam] ; lparam from the stack

.endchanged:
        clc
        ret

onmessage PMM_ADDFILE
        mov     [fProjModified], TRUE
        mov     eax, [wparam]
        cmp     [eax+TOpenFile.projectID], FALSE
        je      .projok

        invoke  SendMessage, [hwnd], PMM_REMOVEFILE, eax, 0

.projok:
        cmp     [lparam], 0
        jne     .addtocategory

; Add as main file.
        mov     eax, [wparam]
        cmp     [eax+TOpenFile.ptrType], ftMainSource
        je      .typeok
        cmp     [eax+TOpenFile.ptrType], ftCommonSource
        jne     .erroradd
        mov     [eax+TOpenFile.ptrType], ftMainSource   ; change file type.
.typeok:
        mov     [eax+TOpenFile.projectID], TRUE

        stdcall AddToListView, [hwnd], ProjMainCtrlID, [wparam]
        invoke  SendDlgItemMessage, [hwnd], ProjMainCtrlID, LVM_SETCOLUMNWIDTH, 0, LVSCW_AUTOSIZE

        stdcall FileChanged, [wparam], fcEditor
        jmp     .endadd

.addtocategory:
        mov     eax, [wparam]
        cmp     [eax+TOpenFile.ptrType], ftMainSource
        jne     .typeok2

        mov     [eax+TOpenFile.ptrType], ftCommonSource

.typeok2:
        cmp     [eax+TOpenFile.projectID], FALSE
        jne     .erroradd

        mov     eax, [lparam]
        cmp     eax, -1
        jne     .addit

.getcurrent:
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CARET, 0
        test    eax, eax
        jnz     .additfocused

        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_ROOT, 0
        test    eax, eax
        jnz     .additfocused

        stdcall OnAddCategory, eax, eax
        test    eax, eax
        jz      .erroradd
        jmp     .getcurrent

.additfocused:
        mov     [lparam], eax

.addit:
        stdcall GetTVItemPtr, [hwnd], eax
        stdcall ListInsert, eax, -1
        test    eax, eax
        jz      .erroradd

        mov     ecx, [wparam]
        mov     dword [eax], ecx
        mov     [ecx+TOpenFile.projectID], TRUE


        stdcall FileChanged, ecx, fcEditor

        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CARET, 0
        cmp     eax, [lparam]
        jne     .endadd

        stdcall ReloadListView, [hwnd], [lparam]

.endadd:
        xor     eax, eax
        clc
        ret

.erroradd:
        xor     eax, eax
        inc     eax
        clc
        ret

onmessage PMM_REMOVEFILE

        mov     [fProjModified], TRUE

        mov     esi, [wparam]
        mov     [esi+TOpenFile.projectID], FALSE
        stdcall FileChanged, esi, fcEditor

        stdcall SearchForFile, [hwnd], esi
        test    eax, eax
        jz      .mainfile
        js      .errdel

        mov     edi, eax
        mov     ebx, edx

        stdcall GetTVItemPtr, [hwnd], edi
        stdcall ListDelete, eax, ebx

        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CARET, 0
        cmp     eax, edi
        jne     .enddel

        stdcall ReloadListView, [hwnd], edi

.enddel:
        xor     eax, eax
        clc
        ret

.mainfile:
        invoke  SendDlgItemMessage, [hwnd], ProjMainCtrlID, LVM_DELETEITEM, 0, 0
        jmp     .enddel

.errdel:
        xor     eax, eax
        inc     eax
        clc
        ret

onmessage CEM_LOADFILE
        mov     ebx, [ptrProjectFile]

        cmp     [wparam], 0
        jne     .loadit

        cmp     [ebx+TOpenFile.fNeverSaved], FALSE
        je      .loadit

        mov     [fProjModified], FALSE
        xor     eax, eax
        clc
        ret

.loadit:
        stdcall ChangeDir, [ebx+TOpenFile.hFileName]

        mov     eax, [wparam]
        test    eax, eax
        jnz     @f
        mov     eax, [ebx+TOpenFile.hFileName]
@@:
        stdcall LoadBinaryFile, eax
        mov     [.ptr], eax
        mov     esi, eax

; Check the file singature
        lodsd
        cmp     eax, 'FRPR'   ; Fresh project?
        je      .getoptions

.invalidfile:
        invoke  MessageBox, [hApplication], cErrorNotProject, NULL, MB_ICONERROR or MB_OK
        jmp     .endopenproj

; Get the project options from the project file. (For now it is only
; the memory for the compiler.
.getoptions:
        lodsd
        cmp     eax, -1
        jne     .oldformat

; New format options

; Get memory size:
        lodsd
        mov     [poMemory], eax

; Get bin name check.
        xor     eax, eax
        lodsb
        mov     [poBinNameAuto], eax

; Get bin name...
        stdcall StrLen, esi
        mov     ebx, eax
        stdcall StrCopy, [poBinName], esi
        lea     esi, [esi+ebx+1]
        jmp     .getmainfile

.oldformat:
        mov     [poMemory], eax

; Create the main file of the project.
.getmainfile:
        stdcall StrLen, esi
        mov     ebx, eax

        test    ebx, ebx
        jz      .getcategories

        stdcall CreateFStruct, esi, ftMainSource
        mov     edi, eax

        cmp     [wparam], 0
        jne     .addmain        ; loaded from template so someone else have to bother for consistency

        stdcall GetModified, edi
        test    eax, eax
        jz      .addmain

        stdcall ProjectUnConsistent, edi
        cmp     eax, IDNO
        jne     .addmain

        stdcall FreeFStruct, edi
        mov     [fProjModified], TRUE
        jmp     .getcategories

.addmain:
        invoke  SendMessage, [hwnd], PMM_ADDFILE, edi, NULL
        mov     [fProjModified], FALSE

.getcategories:
        lea     esi, [esi+ebx+1]        ; goto to the first category

; Main category loop:
.maincatloop:
        lodsd   ; image index
        cmp     eax, -1  ; Is this is the end of file?
        je      .endopenproj

        lea     esi, [esi+4]    ;skip the flags for now.

        stdcall GetCategoryTree, esi, TVI_ROOT, eax
        mov     ebx, eax                ; Current category tree item.

.filesloop:
        stdcall StrLen, esi
        lea     esi, [esi+eax+1]        ; goto to next file

        cmp     byte [esi], 0
        je      .endoffiles

        stdcall DetectFileType, esi
        stdcall CreateFStruct, esi, eax
        mov     edi, eax

        cmp     [wparam], 0
        jne     .addtoproject    ; loaded from template.

        stdcall GetModified, edi
        test    eax, eax
        jz      .addtoproject

        stdcall ProjectUnConsistent, edi
        cmp     eax, IDNO
        je      .adderror

.addtoproject:
        invoke  SendMessage, [hwnd], PMM_ADDFILE, edi, ebx
        mov     [fProjModified], FALSE

        test    eax, eax
        jz      .filesloop

.adderror:
        mov     [fProjModified], TRUE
        stdcall FreeFStruct, edi   ; error add file...
        jmp     .filesloop

.endoffiles:
        inc     esi     ; next category.
        jmp     .maincatloop

.endopenproj:
        invoke  HeapFree, [hHeap], 0, [.ptr]
        stdcall SetMainTitle

        mov     ebx, [ptrProjectFile]
        stdcall AddToMRUList, [ptrMRUProj], [ebx+TOpenFile.hFileName]

        invoke  SendDlgItemMessage, [hwnd], ProjMainCtrlID, LVM_SETCOLUMNWIDTH, 0, LVSCW_AUTOSIZE

        xor     eax, eax
        clc
        ret

iglobal
  cErrorNotProject db 'Error: this file is not a project file in Fresh project format.',13
                   db 'New project created.',0
endg


onmessage _CEM_GETTEXT
        mov     eax, [ptrProjectFile]
        stdcall ChangeDir, [eax+TOpenFile.hFileName]

        mov     esi, [lparam]

        invoke  HeapAlloc, [hHeap], 0, 65536    ; Maximal size of the project file.
        test    eax, eax
        jz      .errorgettext

        mov     [esi+TInMemoryFile.ptrBuffer], eax
        mov     edi, eax

        mov     eax, 'FRPR'
        stosd

; New project format save
        xor     eax, eax
        dec     eax
        stosd

; save memory...
        mov     eax, [poMemory]
        stosd
; save binary name auto flag.
        mov     al, byte [poBinNameAuto]
        stosb

; save binary name
        stdcall GetRelativePath, [poBinName]
        push    eax
        call    .StrToStream

        stdcall StrDel ; from the stack

        xor     al, al
        stosb

; Main project file:
        invoke  SendMessage, [hwnd], PMM_GETMAINFILE, 0, 0
        test    eax, eax
        jz      .mainok

        stdcall GetRelativePath, [eax+TOpenFile.hFileName]
        push    eax
        call    .StrToStream
        stdcall StrDel ; from the stack
        xor     al,al

.mainok:
        stosb   ; end of main file filename.

; Save categories...
;        proc CategoriesScan, hTree, tvRoot, lparam, ptrCallback
        invoke  GetDlgItem, [hwnd], ProjTreeCtrlID
        stdcall CategoriesScan, eax, TVI_ROOT, eax, SaveOneCategory
        xor     eax, eax
        dec     eax
        stosd                   ; end of categories.

        sub     edi, [esi+TInMemoryFile.ptrBuffer]
        mov     [esi+TInMemoryFile.fileSize], edi
        xor     eax, eax
        mov     [esi+TInMemoryFile.fAscii], eax

.finish:
        clc
        ret

.errorgettext:
        xor     eax, eax
        inc     eax
        clc
        ret

;--------------------------------------
; edi - pointer to the stream buffer
; eax - handle/pointer to the string
;--------------------------------------
.StrToStream:
        stdcall StrPtr, eax
        xor     ecx, ecx

.copyname:
        mov     cl, [eax]
        inc     eax
        jecxz   .endofname
        mov     [edi], cl
        inc     edi
        jmp     .copyname

.endofname:
        retn


onmessage  _CEM_CHANGENOTIFY
; change directory to the new name...
        mov     ebx, [ptrProjectFile]
        stdcall ChangeDir, [ebx+TOpenFile.hFileName]
        stdcall SetMainTitle
        stdcall AddToMRUList, [ptrMRUProj], [ebx+TOpenFile.hFileName]
        clc
        ret


onmessage _CEM_FOCUSFILE
        invoke  SetFocus, [hwnd]
        clc
        ret


onmessage _CEM_CANACTION
        xor     eax, eax
        clc
        ret

onmessage _CEM_EDITACTION
        clc
        ret


onmessage _CEM_GETFILE
        mov     eax, [ptrProjectFile]
        clc
        ret


onmessage _CEM_GETMODIFIED
        mov     eax, [fProjModified]
        clc
        ret


onmessage _CEM_SETMODIFIED
        push    [wparam]
        pop     [fProjModified]
        clc
        ret


onmessage WM_ACTIVATE
        stdcall FocusFile, [ptrProjectFile]
        stc
        ret

onmessage WM_DROPFILES
        lea     esi, [.str]
        invoke  DragQueryFile, [wparam], -1, esi, 1024
        mov     ebx, eax
        xor     edi, edi

.opendroploop:
        invoke  DragQueryFile, [wparam], edi, esi, 1024
        stdcall OpenFileByName, esi, NULL, FALSE
        test    eax, eax
        jz      .next
        invoke  SendMessage, [hwnd], PMM_ADDFILE, eax, -1
.next:
        inc     edi
        dec     ebx
        jnz     .opendroploop

        invoke  DragFinish, [wparam]

        xor     eax, eax
        clc
        ret

onmessage WM_COMMAND
        cmp     [wparam], IDOK
        jne     .tomain

        invoke  GetFocus
        invoke  GetDlgCtrlID, eax
        cmp     eax, ProjTreeCtrlID
        je      .ondefault

        cmp     eax, ProjListCtrlID
        je      .listviewenter
        cmp     eax, ProjMainCtrlID
        jne     .ondefault

.listviewenter:
        mov     [wparam], eax
        jmp     .itemactivate

.tomain:
        invoke  PostMessage, [hMainWindow], [wmsg], [wparam], [lparam]
        jmp     .ondefault

onmessage PMM_GETMAINFILE

        mov     [.lvi.mask], LVIF_PARAM
        mov     [.lvi.iItem], 0
        lea     eax, [.lvi]
        invoke  SendDlgItemMessage, [hwnd], ProjMainCtrlID, LVM_GETITEM, 0, eax
        test    eax, eax
        jz      .endgetmain
        mov     eax, [.lvi.lParam]
.endgetmain:
        clc
        ret

onmessage WM_NOTIFY
        mov     eax, [wparam]
        mov     esi, [lparam]
        cmp     eax, ProjTreeCtrlID
        je      .treenotify
        cmp     eax, ProjMainCtrlID
        je      .listnotify
        cmp     eax, ProjListCtrlID
        je      .listnotify
        clc
        ret

;-----------ListView notification messages-----------------

.listnotify:
        cmp     [esi+NMHDR.code], LVN_ITEMACTIVATE
        je      .itemactivate
        cmp     [esi+NMHDR.code], LVN_KEYDOWN
        je      .listkey
        cmp     [esi+NMHDR.code], NM_RCLICK
        je      .lvrightclick
        cmp     [esi+NMHDR.code], LVN_ITEMCHANGED
        jmp     .qfalse

.lvrightclick:
        lea     eax, [.pnt]
        invoke  GetCursorPos, eax
        invoke  TrackPopupMenu, [hListPopup], TPM_LEFTALIGN or TPM_RIGHTBUTTON, [.pnt.x], [.pnt.y], 0, [hwnd], NULL
        jmp     .qfalse

.itemactivate:
        xor     ebx, ebx
        dec     ebx

.openloop:
        invoke  SendDlgItemMessage, [hwnd], [wparam], LVM_GETNEXTITEM, ebx, LVNI_SELECTED
        cmp     eax, -1
        je      .qfalse

        mov     ebx, eax

        mov     [.lvi.mask], LVIF_PARAM
        mov     [.lvi.iItem], eax
        lea     eax, [.lvi]
        invoke  SendDlgItemMessage, [hwnd], [wparam], LVM_GETITEM, 0, eax
        stdcall LoadFile, [.lvi.lParam], NULL
        jmp     .openloop


;---------- Del key is pressed in the list view -------------------
.listkey:
        cmp     [esi+LVKEYDOWN.wVKey], VK_DELETE
        jne     .qfalse

.delitems:
        xor     ebx, ebx
        push    ebx             ; marker for the end of the list
        dec     ebx
.delloop:
        invoke  SendDlgItemMessage, [hwnd], [wparam], LVM_GETNEXTITEM, ebx, LVNI_SELECTED
        cmp     eax, -1
        je      .delit

        mov     ebx, eax

        mov     [.lvi.mask], LVIF_PARAM
        mov     [.lvi.iItem], eax
        lea     eax, [.lvi]
        invoke  SendDlgItemMessage, [hwnd], [wparam], LVM_GETITEM, 0, eax
        push    [.lvi.lParam]
        jmp     .delloop

.delit:
        cmp     dword [esp], 0
        je      .delitloop

        invoke  MessageBox, [hApplication], cDelFileQuery, cDelFileTitle, MB_ICONQUESTION or MB_YESNO
        cmp     eax, IDYES
        je      .delitloop

.clearstack:
        pop     ecx
        jecxz   .endfiledel
        jmp     .clearstack

.delitloop:
        pop     ecx
        jecxz   .endfiledel
        invoke  SendMessage, [hwnd], PMM_REMOVEFILE, ecx, 0
        jmp     .delitloop

.endfiledel:
        xor     eax, eax
        clc
        ret

iglobal
  cDelFileQuery db 'Do you really want to remove selected file(s) from the project?',0
  cDelFileTitle db 'Remove file from project.',0
endg


;--- Category tree notification messages ---------
.treenotify:
        cmp     [esi+NMHDR.code], TVN_SELCHANGED
        je      .selchanged

        cmp     [esi+NMHDR.code], TVN_KEYDOWN
        je      .keydown

        cmp     [esi+NMHDR.code], NM_RCLICK
        je      .tvrightclick

        cmp     [esi+NMHDR.code], TVN_ENDLABELEDIT
        je      .endlabeledit

        cmp     [esi+NMHDR.code], TVN_DELETEITEM
        jne     .qfalse

.deleteitem:
        mov     edi, [esi+NMTREEVIEW.itemOld.lParam]
        test    edi, edi
        jz      .qfalse

        test    [esi+NMTREEVIEW.itemOld.state], TVIS_FOCUSED
        jz      .freelist

        invoke  SendDlgItemMessage, [hwnd], ProjListCtrlID, LVM_DELETEALLITEMS, 0, 0

.freelist:
        stdcall ListFree, edi, FileListFreeCallback
        invoke  HeapFree, [hHeap], 0, edi
        jmp     .qfalse

;----------- end of label editing ------------------------
.endlabeledit:
        mov     esi, [lparam]
        cmp     [esi+TVDISPINFO.item.pszText], 0
        je      .endedit

        mov     [fProjModified], TRUE
        mov     [esi+TVDISPINFO.item.mask], TVIF_TEXT
        lea     esi, [esi+TVDISPINFO.item]
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID, TVM_SETITEM, 0, esi

.endedit:
       clc
       ret

;---------- right click on treeview control --------------
.tvrightclick:
        lea     eax, [.pnt]
        invoke  GetCursorPos, eax
        invoke  TrackPopupMenu, [hTreePopup], TPM_LEFTALIGN or TPM_RIGHTBUTTON, [.pnt.x], [.pnt.y], 0, [hwnd], NULL
        jmp     .qfalse


;---------- selection changed notification ---------------
.selchanged:
; load the new selected category in the listview...
        stdcall ReloadListView, [hwnd], [esi+NMTREEVIEW.itemNew.hItem]
        xor     eax, eax
        clc
        ret

;------------- key down notification -----------------
.keydown:
        cmp     [esi+TVKEYDOWN.wVKey], VK_DELETE
        jne     .qfalse

; delete key on category...
        invoke  MessageBox, [hApplication], cDelProjQuery, cDelProjTitle, MB_ICONQUESTION or MB_YESNO
        cmp     eax, IDYES
        jne     .qfalse

        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CARET, 0
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID, TVM_DELETEITEM, 0, eax
        jmp     .qfalse

iglobal
  cDelProjQuery db 'Do you really want to remove this category and all category files from the project?', 0
  cDelProjTitle db 'Remove category.', 0
endg


onmessage WM_CLOSE

        stdcall CloseFile, [ptrProjectFile]
        cmp     eax, cfrCanceled
        je      .qfalse

        mov     [ptrProjectFile], 0
        jmp     .qfalse

onmessage WM_DESTROY

        invoke  SendMessage, [hwnd], PMM_GETMAINFILE, 0, 0
        test    eax, eax
        jz      .mainfree

        stdcall FileListFreeCallback, eax

.mainfree:
        stdcall DeleteAllTVItems, [hProjManager] ; clear all items in the project and frees the memory.
        mov     [hProjManager], 0

        stdcall StrDel, [poBinName]
        mov     [poBinNameAuto], TRUE

        jmp     .qfalse

iglobal
  cMRUProjSection db 'MRU projects',0
endg

iglobal
  cMainButtonTxt db '>',0
endg

onmessage AM_INITWINDOW

        xor     ebx, ebx

        invoke  CreateWindowEx, WS_EX_CLIENTEDGE, cListviewClassName, ebx,      \
                                WS_CHILD or WS_VISIBLE or LVS_REPORT or         \
                                LVS_SHAREIMAGELISTS or LVS_SORTASCENDING or     \
                                LVS_NOSCROLL or WS_TABSTOP, \
                                ebx, ebx, ebx, 20,                              \
                                [hwnd], ProjMainCtrlID, [hInstance], ebx

        mov     edi, eax
        stdcall SetAlign, eax, waTop

        mov     [.col.mask], LVCF_FMT or LVCF_WIDTH
        mov     [.col.fmt], LVCFMT_LEFT
        mov     [.col.cx], 1024
        lea     eax, [.col]
        invoke  SendMessage, edi, LVM_INSERTCOLUMN, ebx, eax
        invoke  SendMessage, edi, LVM_SETIMAGELIST, LVSIL_SMALL, [hTabIml]

        invoke  CreateWindowEx, ebx, cStaticClassName, ebx,              \
                                WS_CHILD or WS_VISIBLE or SS_LEFT,       \
                                ebx, ebx, ebx, 4,                        \
                                [hwnd], ebx, [hInstance], ebx
        stdcall SetAlign, eax, waTop

        invoke  CreateWindowEx, WS_EX_CLIENTEDGE,                                                       \
                                cTreeViewClassName, ebx,                                                \
                                WS_CHILD or WS_VISIBLE or TVS_DISABLEDRAGDROP or TVS_HASBUTTONS or      \
                                TVS_SHOWSELALWAYS or TVS_HASLINES or TVS_LINESATROOT or TVS_EDITLABELS or WS_TABSTOP, \
                                ebx, ebx, ebx, 100,                                                     \
                                [hwnd], ProjTreeCtrlID, [hInstance], ebx
        mov     edi, eax
        stdcall SetAlign, eax, waTop
        invoke  SendMessage, edi, TVM_SETIMAGELIST, TVSIL_NORMAL, [imlCategories]

        stdcall CreateSplitter, edi, TRUE, resVSplit, resHSplit


        invoke  CreateWindowEx, WS_EX_CLIENTEDGE, cListviewClassName, ebx,              \
                                WS_CHILD or WS_VISIBLE or LVS_REPORT or                 \
                                LVS_SHAREIMAGELISTS or LVS_SORTASCENDING or             \
                                LVS_NOCOLUMNHEADER or LVS_SHOWSELALWAYS or WS_TABSTOP,  \
                                ebx, ebx, ebx, ebx,                                     \
                                [hwnd], ProjListCtrlID, [hInstance], ebx

        mov     edi, eax
        stdcall SetAlign, eax, waClient

        invoke  SendMessage, edi, LVM_SETIMAGELIST, LVSIL_SMALL, [hTabIml]
        invoke  SendMessage, edi, LVM_SETEXTENDEDLISTVIEWSTYLE, 0, \
                             LVS_EX_FULLROWSELECT or LVS_EX_LABELTIP
;        mov     [.col.mask], LVCF_FMT or LVCF_WIDTH
;        mov     [.col.fmt], LVCFMT_LEFT
;        mov     [.col.cx], 1024
        lea     eax, [.col]
        invoke  SendMessage, edi, LVM_INSERTCOLUMN, ebx, eax

        invoke  DragAcceptFiles, [hwnd], TRUE

.qfalse:
        xor     eax,eax
        clc
        ret
endwp




; it should return -1 if the result is still not found.
; arguments are [ptrList] and [hTVItem] and [lparam]
proc SaveOneCategory .ptrList, .hItem, .hTree
locals
.tvi TVITEM
.count dd ?
endl
        push    ebx esi

        stdcall GetFullItemName, [.hTree], [.hItem]
        mov     ebx, eax

        mov     [.tvi.mask], TVIF_IMAGE or TVIF_PARAM
        mov     eax, [.hItem]
        mov     [.tvi.hItem], eax
        lea     eax, [.tvi]
        invoke  SendMessage, [.hTree], TVM_GETITEM, 0, eax
        test    eax, eax
        jz      .finish

        mov     eax, [.tvi.iImage]         ; image index.
        stosd
        xor     eax, eax                        ; reserved.
        stosd

        mov     eax, ebx
        call    ProjManProc.StrToStream
        xor     al,al
        stosb

        mov     esi, [.tvi.lParam]
        mov     edx, [esi+TList.Count]
        mov     [.count], edx
        mov     esi, [esi+TList.ptrItems]

.looplist:
        dec     [.count]
        js      .finish

        lodsd   ; ptr to TOpenFile

        stdcall GetRelativePath, [eax+TOpenFile.hFileName]
        push    eax
        call    ProjManProc.StrToStream
        stdcall StrDel ; from the stack
        xor     al,al
        stosb
        jmp     .looplist

.finish:
        stdcall StrDel, ebx
        xor     eax, eax
        stosb
        dec     eax
        pop     esi ebx
        ret
endp



proc FileListFreeCallback .ptrOpenFile
        push    esi
        mov     esi, [.ptrOpenFile]

        mov     [esi+TOpenFile.projectID], FALSE
        stdcall FileChanged, esi, fcEditor
        pop     esi
        ret
endp


;***********************************************************
; Set the filename of the project and updates the title of
; the main window.
;***********************************************************
proc SetMainTitle
        push    esi

        xor     eax, eax
        mov     esi, [ptrProjectFile]
        test    esi, esi
        jz      @f

        stdcall StrPtr, [esi+TOpenFile.hFileName]

@@:
        invoke  SendMessage, [hMainWindow], WM_SETTEXT, 0, eax
        pop     esi
        ret
endp



;***********************************************************
; List view utils.
;***********************************************************
proc ReloadListView .hProj, .hCategory
        push    edi ebx

        invoke  SendDlgItemMessage, [.hProj], ProjListCtrlID, LVM_DELETEALLITEMS, 0, 0
        stdcall GetTVItemPtr, [.hProj], [.hCategory]
        mov     edi, eax

        mov     eax, [edi+TList.Count]
        mov     ebx, eax
        invoke  SendDlgItemMessage, [.hProj], ProjListCtrlID, LVM_SETITEMCOUNT, ebx, 0

        mov     edi, [edi+TList.ptrItems]

.addnewlist:
        dec     ebx
        js      .finish

        stdcall AddToListView, [.hProj], ProjListCtrlID, [edi]
        lea     edi, [edi+4]
        jmp     .addnewlist

.finish:
        invoke  SendDlgItemMessage, [.hProj], ProjListCtrlID, LVM_SETCOLUMNWIDTH, 0, LVSCW_AUTOSIZE
        pop     ebx edi
        ret
endp


proc AddToListView .hProject, .ListID, .ptrOpenFile
local .lvi:LVITEM
        push    esi
        mov     esi, [.ptrOpenFile]

        stdcall GetScreenProperties, esi
        test    eax, eax
        jz      .finish

        push    eax     ; for StrDel
        stdcall StrPtr, eax

        mov     [.lvi.mask], LVIF_IMAGE or LVIF_PARAM or LVIF_TEXT
        mov     [.lvi.iItem], 0
        mov     [.lvi.iSubItem], 0
        mov     [.lvi.iImage], edx
        mov     [.lvi.lParam], esi
        mov     [.lvi.pszText], eax

        lea     eax, [.lvi]
        invoke  SendDlgItemMessage, [.hProject], [.ListID], LVM_INSERTITEM, 0, eax
        stdcall StrDel ; from the stack.
.finish:
        pop     esi
        ret
endp


;------------------------------------------------------------;
; Searches the project for the file and returns:             ;
; eax - handle of the category that contain this file.       ;
;   eax = 0 if the file is main file.                        ;
;   eax = -1 if the file was not found.                      ;
; edx - index in TList containing the file pointer.          ;
;------------------------------------------------------------;
proc SearchForFile .hProject, .ptrOpenFile
        invoke  SendMessage, [.hProject], PMM_GETMAINFILE, 0, 0
        test    eax, eax
        jz      .nomain

        cmp     eax, [.ptrOpenFile]
        je      .main

.nomain:
        invoke  GetDlgItem, [.hProject], ProjTreeCtrlID
        mov     ebx, eax
        stdcall CategoriesScan, ebx, TVI_ROOT, [.ptrOpenFile], _SearchCategory
        ret

.main:
        mov     edx, eax
        xor     eax, eax
        ret
endp


proc _SearchCategory .ptrList, .hTVItem, .ptrOpenFile
        stdcall ListIndexOf, [.ptrList], [.ptrOpenFile]
        cmp     eax, -1
        je      .finish

        mov     edx, [.hTVItem]
        xchg    eax, edx

.finish:
        ret
endp


;---------------------------------------------------------------
; Recursively loops through all categories and calls callback
; procedure for every category.
;---------------------------------------------------------------
proc CategoriesScan .hTree, .tvRoot, lparam, .ptrCallback
local .tvi:TVITEM
        push    esi ebx
        invoke  SendMessage, [.hTree], TVM_GETNEXTITEM, TVGN_CHILD, [.tvRoot]

.loop:
        mov     ebx, eax
        test    ebx, ebx
        jz      .notfound

        mov     [.tvi.mask], TVIF_PARAM
        mov     [.tvi.hItem], ebx
        lea     eax, [.tvi]
        invoke  SendMessage, [.hTree], TVM_GETITEM, 0, eax

        mov     esi, [.tvi.lParam]
        test    esi, esi
        jz      .listdone

; it should return -1 if the result is still not found.
; arguments are [ptrList] and [hTVItem] and [lparam]
        stdcall [.ptrCallback], esi, ebx, [lparam]
        test    eax, eax
        jge     .found

.listdone:
        stdcall CategoriesScan, [.hTree], ebx, [lparam], [.ptrCallback]
        test    eax, eax
        jge     .found

        invoke  SendMessage, [.hTree], TVM_GETNEXTITEM, TVGN_NEXT, ebx
        jmp     .loop

.notfound:
        dec     eax     ; eax := -1 here
        pop     ebx esi
        ret

.found:
        pop     ebx esi
        ret
endp


;---------------------------------------------
; Returns handle to string contains full path
; name for given tree-view item.
;---------------------------------------------
proc GetFullItemName .hTree, .hItem
locals
.tvi TVITEM
.str rb 256
endl
        push    esi ebx

        stdcall StrNew
        mov     esi, eax
        mov     eax, [.hItem]

.buildloop:
        mov     ebx, eax
        test    ebx, ebx
        jz      .rootoftree

        mov     [.tvi.mask], TVIF_TEXT
        mov     [.tvi.hItem], ebx
        lea     eax, [.str]
        mov     [.tvi.pszText], eax
        mov     [.tvi.cchTextMax], 256
        lea     eax, [.tvi]
        invoke  SendMessage, [.hTree], TVM_GETITEM, 0, eax

        stdcall StrInsertChar, esi, '\', 0
        stdcall StrInsert, esi, [.tvi.pszText], 0

        invoke  SendMessage, [.hTree], TVM_GETNEXTITEM, TVGN_PARENT, ebx
        jmp     .buildloop

.rootoftree:
        mov     eax, esi
        pop     ebx esi
        ret
endp


;***********************************************************
; TreeView utils. This is small library for easy work with
; TreeView control.
;***********************************************************

;**************************************************************
; Returns lParam of the TreeView item.
;**************************************************************
proc GetTVItemPtr .hProject, .hItem
local .item:TVITEM
        push    ecx edx
        push    [.hItem]
        pop     [.item.hItem]
        mov     [.item.mask], TVIF_PARAM
        lea     eax, [.item]
        invoke  SendDlgItemMessage, [.hProject], ProjTreeCtrlID, TVM_GETITEM, 0, eax
        test    eax, eax
        jz      .error
        mov     eax, [.item.lParam]
        clc
        jmp     .finish
.error:
        stc
.finish:
        pop     edx ecx
        ret
endp



;***********************************************
; Handler of action actMoveCategory
;***********************************************
proc OnMoveCategory wparam, lparam
        xor     ebx, ebx
        cmp     [wparam], actMoveCategoryUp
        je      @f
        inc     ebx
@@:
        invoke  GetDlgItem, [hProjManager], ProjTreeCtrlID
        mov     esi, eax

        invoke  SendMessage, esi, TVM_GETNEXTITEM, TVGN_CARET, 0
        stdcall MoveTVItem, esi, eax, ebx
        mov     [fProjModified], TRUE
        ret
endp


;****************************************************************
; Moves one hItem up or down in the treeview, depending of the
; fMoveDown:
;    FALSE - move up
;    TRUE  - move down
;****************************************************************
proc MoveTVItem .hTreeView, .hItem, .fMoveDown
locals
.ptr dd ?
.obj dd ?
.tvi TVITEM
lparam dd ?
.tvs TVSORTCB
endl
        push    esi edi ebx
        invoke  SendMessage, [.hTreeView], TVM_GETCOUNT, 0, 0
        shl     eax, 2
        invoke  HeapAlloc, [hHeap], 0, eax
        mov     [.ptr], eax
        mov     edi, eax
        mov     eax, [.hItem]
        mov     [.tvi.hItem], eax
        mov     [.tvi.mask], TVIF_PARAM
        lea     eax, [.tvi]
        invoke  SendMessage, [.hTreeView], TVM_GETITEM, 0, eax
        mov     eax, [.tvi.lParam]
        mov     [lparam], eax

        invoke  SendMessage, [.hTreeView], TVM_GETNEXTITEM, TVGN_PARENT, [.hItem]
        mov     ebx, eax

        invoke  SendMessage, [.hTreeView], TVM_GETNEXTITEM, TVGN_CHILD, ebx
.getloop:
        test    eax, eax
        jz      .endofchildren

        mov     [.tvi.hItem], eax
        mov     [.tvi.mask], TVIF_PARAM or TVIF_HANDLE
        lea     eax, [.tvi]
        invoke  SendMessage, [.hTreeView], TVM_GETITEM, 0, eax
        mov     eax, [.tvi.lParam]
        cmp     eax, [lparam]
        jne     .storeit
        mov     [.obj], edi
.storeit:
        stosd

        invoke  SendMessage, [.hTreeView], TVM_GETNEXTITEM, TVGN_NEXT, [.tvi.hItem]
        jmp     .getloop

.endofchildren:
        stosd
        cmp     [.fMoveDown], FALSE
        je      .moveup

.movedown:
        mov     edi, [.obj]
        mov     eax, [edi+4]
        test    eax, eax
        jz      .cantmove       ; it is the last element.

        mov     ecx, [edi]
        mov     [edi], eax
        mov     [edi+4], ecx
        jmp     .sortit

.moveup:
        mov     edi, [.obj]
        cmp     edi, [.ptr]
        je      .cantmove               ; it is the first element

        mov     eax, [edi-4]
        mov     ecx, [edi]
        mov     [edi], eax
        mov     [edi-4], ecx

.sortit:
        mov     [.tvs.hParent], ebx
        mov     [.tvs.lpfnCompare], CompareTwoItems
        mov     eax, [.ptr]
        mov     [.tvs.lParam], eax
        lea     eax, [.tvs]
        invoke  SendMessage, [.hTreeView], TVM_SORTCHILDRENCB, 0, eax

.cantmove:
        invoke  HeapFree, [hHeap], 0, [.ptr]

        pop     ebx edi esi
        ret
endp


proc CompareTwoItems .lParam1, .lParam2, .ptrList
        push    esi

        mov     esi, [.ptrList]
        mov     ecx, [.lParam1]
        mov     edx, [.lParam2]
.loop:
        lodsd
        test    eax, eax
        jz      .notfound

        cmp     eax, ecx
        je      .first

        cmp     eax, edx
        jne     .loop

.second:
        xor     eax, eax
        inc     eax
        pop     esi
        ret

.first:
        xor     eax, eax
        dec     eax
        pop     esi
        ret

.notfound:
        xor     eax, eax
        pop     esi
        ret
endp


;****************************************************************
; Returns in ebx the handle of parent treeview item.
;****************************************************************
proc GetItemParent .hItem
        push    eax ecx edx

        invoke  SendDlgItemMessage, [hProjManager], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_PARENT, [.hItem]
        test    eax,eax
        jz      .error
        clc
        jmp     .finish

.error:
        stc
.finish:
        mov     ebx, eax
        pop     edx ecx eax
        ret
endp


proc DeleteAllTVItems .hProj
.loop:
        invoke  SendDlgItemMessage, [.hProj], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_ROOT, 0
        test    eax, eax
        jz      .finish

        stdcall DeleteOneTVItem, [.hProj], eax
        jmp     .loop
.finish:
        ret
endp


proc DeleteOneTVItem .hProj, .hItem
.clearloop:
        invoke  SendDlgItemMessage, [.hProj], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CHILD, [.hItem]
        test    eax, eax
        jz      .childrendeleted

        stdcall DeleteOneTVItem, [.hProj], eax
        jmp     .clearloop

.childrendeleted:
        invoke  SendDlgItemMessage, [.hProj], ProjTreeCtrlID, TVM_DELETEITEM, 0, [.hItem]
        ret
endp


;-------------------------------------------------------------------------
; Returns the handle of treeview item that follow the tree of direcories
; of [hFileName]. If there is no some (or any) of the directories of the
; path, it will be created.
; The function checks for "/" and "\" as items separator.
; Arguments:
; hFileName - handle or pointer to the path.
; hRoot     - is root TVItem for category. For examples if category
;             have to be created in the root of the TreeView, hParent
;             should be TVI_ROOT
; Returns:
;   eax - Handle to treeview item that should be parent of the file.
;   edx - pointer to the name of the file without path.
;-------------------------------------------------------------------------
proc GetCategoryTree .hFileName, .hRoot, .iImage
        push    esi edi ebx

        stdcall StrPtr, [.hFileName]
        mov     esi, eax

        stdcall StrLen, [.hFileName]
        mov     ecx, eax

; find the filename:
.findloop:
        cmp     byte [esi+ecx], '/'
        je      .foundname
        cmp     byte [esi+ecx], '\'
        je      .foundname
        dec     ecx
        jns     .findloop

; Returns the [hRoot] if there is no directories names.
        mov     eax, [.hRoot]
        mov     edx, esi
        pop     ebx edi esi
        ret

.foundname:
        lea     edi, [esi+ecx+1]
        pushd   [edi]
        and     dword [edi], 0
        stdcall _DoCategoryTree, esi, [.hRoot], [.iImage]    ; track/create path items.
        popd    [edi]
        mov     edx, edi
        pop     ebx edi esi
        ret
endp


proc _DoCategoryTree .ptrName, .hTVRoot, .iImage
        push    esi edi ebx
        mov     esi, [.ptrName]
        mov     ebx, [.hTVRoot]

.outloop:
        cmp     byte [esi], 0   ; is it the end?
        je      .endloop
        mov     edi, esi
.loop:
        lodsb
        cmp     al,'\'
        je      .foundslash
        cmp     al,'/'
        jne     .loop

.foundslash:
        mov     byte [esi-1], 0
        stdcall _DoGetOneItem, edi, ebx, [.iImage]
        mov     ebx, eax
        mov     byte [esi-1], '/'
        jmp     .outloop

.endloop:
        mov     eax, ebx
        pop     ebx edi esi
        ret
endp


proc _DoGetOneItem .ptrItem, .hParent, .iImage
locals
.tvi TVITEM
.tvins TVINSERTSTRUCT
.str rb 256
endl
        push    esi edi ebx
        invoke  SendDlgItemMessage, [hProjManager], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CHILD, [.hParent]
        test    eax, eax
        jz      .notfound
.itemscan:
        mov     ebx, eax
        mov     [.tvi.hItem], eax
        lea     esi, [.str]
        mov     [.tvi.pszText], esi
        mov     [.tvi.cchTextMax], 256
        mov     [.tvi.mask], TVIF_HANDLE or TVIF_TEXT
        lea     ecx, [.tvi]
        invoke  SendDlgItemMessage, [hProjManager], ProjTreeCtrlID, TVM_GETITEM, 0, ecx

        stdcall StrCompNoCase, esi, [.ptrItem]
        jc      .found

        invoke  SendDlgItemMessage, [hProjManager], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_NEXT, ebx
        test    eax, eax
        jnz     .itemscan

.notfound:
        mov     eax, [.hParent]
        mov     [.tvins.hParent], eax
        mov     [.tvins.hInsertAfter], TVI_LAST
        mov     [.tvins.item.mask], TVIF_TEXT or TVIF_IMAGE or TVIF_SELECTEDIMAGE or TVIF_PARAM
        mov     eax, [.iImage]
        mov     [.tvins.item.iImage], eax
        mov     [.tvins.item.iSelectedImage], eax

        mov     eax, [.ptrItem]
        mov     [.tvins.item.pszText], eax

        stdcall ListCreate, NULL
        mov     [.tvins.item.lParam], eax

        lea     eax, [.tvins]
        invoke  SendDlgItemMessage, [hProjManager], ProjTreeCtrlID, TVM_INSERTITEM, 0, eax
        pop     ebx edi esi
        ret

.found:
        mov     eax, ebx
        pop     ebx edi esi
        ret
endp


;--------------------------------------------------------------------
; Recursively scans given source file [hFileName] for "include"
; statements and creates treeview items with filenames.
;--------------------------------------------------------------------
proc ScanFileForIncludes .hFileName, .hTVParent
locals
.tvi TVINSERTSTRUCT
.hthis dd ?
.ptr   dd ?
.cnt   dd ?
endl
        push    esi edi ebx

; Scan the file for includes.
        stdcall LoadBinaryFile, [.hFileName]
        test    eax, eax
        jz      .quit

        mov     [.ptr], eax
        mov     [.cnt], ecx
        test    ecx, ecx
        jz      .finish

; Add the file to the project tree.
        mov     eax, [.hTVParent]
        mov     [.tvi.hParent], eax
        mov     [.tvi.hInsertAfter], TVI_SORT

        stdcall DetectFileType, [.hFileName]
        test    eax, eax
        jz      .finish

        mov     ecx, [eax+TFileType.iFileIcon]
        mov     [.tvi.item.iImage], ecx
        mov     [.tvi.item.iSelectedImage], ecx

        stdcall CreateFStruct, [.hFileName], eax
        mov     [.tvi.item.lParam], eax
        mov     [.tvi.item.mask], TVIF_TEXT or TVIF_IMAGE or TVIF_PARAM or TVIF_SELECTEDIMAGE

        stdcall GetRelativePath, [.hFileName]
        push    eax
        stdcall StrPtr, eax
        mov     [.tvi.item.pszText], eax

        lea     eax, [.tvi]
        invoke  SendDlgItemMessage, [hProjManager], ProjTreeCtrlID, TVM_INSERTITEM, 0, eax
        mov     [.hthis], eax

        stdcall StrDel ; from the stack

        mov     ecx, [.tvi.item.lParam]
        mov     [ecx+TOpenFile.projectID], eax

        mov     esi, [.ptr]
        mov     ecx, [.cnt]

.scan:
        mov     eax, [esi]
        cmp     al, ';'
        je      .comment
        or      eax, $20202020
        cmp     eax, 'incl'
        je      .maybe
.nextscan:
        inc     esi
        loop    .scan
.finish:
        invoke  HeapFree, [hHeap], 0, [.ptr]
.quit:
        pop     ebx edi esi
        ret

.comment:
        cmp     byte [esi], $0d
        je      .scan
        cmp     byte [esi], $0a
        je      .scan
        inc     esi
        loop    .comment
        jmp     .finish

.maybe:
        add     esi, 4
        sub     ecx, 4
        jbe     .finish

        mov     eax, [esi]
        and     eax, $ffffff
        or      eax, $202020
        cmp     eax, 'ude'
        je      .found
        add     esi, 3
        sub     ecx, 3
        jbe     .finish
        jmp     .scan

; Include statement
.found:
        add     esi, 3
        sub     ecx, 3
        jbe     .finish

.foundloop:
        mov     al, [esi]
        cmp     al, '"'
        je      .begin
        cmp     al, "'"
        je      .begin

        cmp     al, ' '
        je      .next
        cmp     al, $9
        jne     .nextscan

.next:
        inc     esi
        loop    .foundloop
        jmp     .finish

.begin:
        inc     esi
        dec     ecx
        jz      .finish
        mov     edi, esi
.scanend:
        cmp     [esi], al
        je      .nameok
        inc     esi
        loop    .scanend
        jmp     .finish
.nameok:
        push    ecx
        mov     byte [esi], 0
;        stdcall ConvertPath, edi
;        push    eax
;        stdcall StrPtr, eax
        stdcall ScanFileForIncludes, edi, [.hthis]
;        stdcall StrDel ; from the stack
        pop     ecx
        inc     esi
        dec     ecx
        jnz     .scan
        jmp     .finish
endp

