;------------------------------------------------------------------
; Creates TList dynamic list
; or init existing one.
; if ptrList = NULL, the memory for the list will be allocated.
; Returns: pointer to TList structure, properly initialized.
;------------------------------------------------------------------
proc ListCreate .ptrList
	push	ebx esi
	xor	ebx, ebx
	mov	esi, [.ptrList]
	test	esi,esi
	jnz	.listok

	invoke	HeapAlloc, [hHeap], ebx, sizeof.TList
	mov	esi, eax

.listok:
	mov	[esi+TList.Count], ebx

	invoke	HeapAlloc, [hHeap], HEAP_ZERO_MEMORY, 4*cMinListCapacity
	mov	[esi+TList.ptrItems], eax

	mov	eax, esi
	pop	esi ebx
	ret
endp


;--------------------------------------------
; Frees all elements of the list calling
; [FreeProc] on every element.
; (proc FreeProc, ptrItem)
; Frees the array of items
;--------------------------------------------
proc ListFree .ptrList, .FreeProc
	push	esi edi ebx

	mov	edi, [.ptrList]
	xor	ebx, ebx
	xor	esi, esi
	xchg	esi, [edi+TList.ptrItems]
	xchg	ebx, [edi+TList.Count]

	cmp	[.FreeProc], 0
	je	.endwhile

	test	esi, esi
	jz	.finish

.freeloop:
	dec	ebx
	js	.endwhile

	lea	eax, [esi+4*ebx]
	stdcall [.FreeProc], [eax]
	jmp	.freeloop

.endwhile:
	invoke	HeapFree, [hHeap], 0, esi

.finish:
	pop	ebx edi esi
	ret
endp



;-----------------------------------------
; Inserts new element in the list and
; returns it's address.
; If [Index] is bigger than List.Count
; the new element will be appended as last
; element.
;-----------------------------------------
proc ListInsert .ptrList, .Index
	push	esi edi ebx

	mov	esi, [.ptrList]

	invoke	HeapSize, [hHeap], 0, [esi+TList.ptrItems]
	cmp	eax, -1
	je	.exiterror

	shr	eax, 2		; get capacity
	mov	ebx, eax

	mov	edi, [.Index]
	cmp	edi, [esi+TList.Count]
	jb	.indexok
	mov	edi, [esi+TList.Count]

.indexok:
	cmp	[esi+TList.Count], ebx
	jb	.capacityok

	mov	eax, ebx
	shr	eax, 1
	add	eax, 2

	cmp	eax, 1024
	jbe	.incok
	mov	eax, 1024

.incok:
	add	ebx, eax
	shl	ebx, 2
	invoke	HeapReAlloc, [hHeap], 0, [esi+TList.ptrItems], ebx
	mov	[esi+TList.ptrItems], eax

.capacityok:
	mov	ecx, [esi+TList.Count]
	sub	ecx, edi ; index
	jz	.dontmove

; Move the items.
	std
	mov	eax, [esi+TList.Count]
	mov	esi, [esi+TList.ptrItems]
	lea	edi, [esi+4*eax]
	lea	esi, [edi-4]
	rep movsd
	cld

	mov	eax, edi
	jmp	.exit

.exiterror:
	xor	eax, eax
	jmp	.exit

.dontmove:
	inc	[esi+TList.Count]
	mov	eax, [esi+TList.ptrItems]
	lea	eax, [eax+4*edi]
.exit:
	pop	ebx edi esi
	ret
endp



proc ListDelete .ptrList, .Index
	push	esi edi ebx

	mov	esi, [.ptrList]
	mov	edi, [.Index]

	cmp	edi, [esi+TList.Count]
	jae	.exit

	dec	[esi+TList.Count]
	mov	ecx, [esi+TList.Count]
	sub	ecx, edi
	jz	.exit

	mov	eax, [esi+TList.ptrItems]
	lea	edi, [eax+4*edi]
	lea	esi, [edi+4]

	rep movsd

.exit:
	pop	ebx edi esi
	ret
endp


proc ListIndexOf .ptrList, .Item
	push	esi edi
	mov	esi, [.ptrList]
	mov	ecx, [esi+TList.Count]
	jecxz	.empty
	mov	edi, [esi+TList.ptrItems]
	lea	edi, [edi+4*ecx-4]		; from the end
	mov	eax, [.Item]
	std
	repnz scasd
	cld
	jz	.found
.empty:
	dec	ecx
.found:
	mov	eax, ecx
	pop	edi esi
	ret
endp