;
;  LabelsList.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;

cNamesInitialSize = 65536 ;131072

uglobal
  ptrLabels dd ?	; ptr to Root items TArray structure.
  ptrNames  dd ?
  NamesCurrent dd ?
  NamesSize dd ?
endg


proc AddToNames .ptrString
	push	esi edi ebx
	mov	esi, [.ptrString]
	mov	edi, [ptrNames]
	mov	ebx, [NamesCurrent]
	mov	eax, [NamesSize]
	sub	eax, ebx
	cmp	eax, $100
	ja	.sizeok
	add	[NamesSize], $1000
	invoke	HeapReAlloc, [hHeap], 0, [ptrNames], [NamesSize]
	mov	[ptrNames], eax
	mov	edi, eax
.sizeok:
	mov	eax, ebx
.copyloop:
	mov	cl, [esi]
	inc	esi
	mov	[edi+ebx], cl
	inc	ebx
	test	cl,cl
	jnz	.copyloop
	mov	[NamesCurrent], ebx
	pop	ebx edi esi
	ret
endp

;---------------------------------------------------------------
; Returns:
;    eax - The name of the label without the first parent name.
;    edx - First parent.
;---------------------------------------------------------------
proc GetLabelParent .ptrLabel
	push	esi ebx
	mov	esi, [.ptrLabel]
	push	esi
	cmp	word [esi], '..'	; Begins with ".."
	jne	.search
	add	esi, 2
.search:
	mov	al, [esi]
	test	al,al
	jz	.notfound
	cmp	al, '.'
	je	.found
	inc	esi
	jmp	.search
.found:
	mov	byte [esi], 0
	lea	eax, [esi+1]
	pop	edx
	pop	ebx esi
	ret
.notfound:
	pop	eax
	xor	edx, edx
	pop	ebx esi
	ret
endp

proc SearchCreateLabel .ptrVarArray, .ptrLabel
local .hash:DWORD
	push	esi edi ebx edx
	stdcall StrHash, [.ptrLabel], 0
	mov	[.hash], eax
	mov	edx, [.ptrVarArray]
	mov	esi, [edx]
	test	esi, esi
	jnz	.arrayok
	stdcall CreateArray, sizeof.TLabel
	mov	edx, [.ptrVarArray]
	mov	[edx],eax
	mov	esi, eax
	stdcall CreateArray, 8		; Array for search tree.
	mov	[esi+TArray.lParam], eax
	lea	eax, [esi+TArray.lParam]
	stdcall AddArrayItem, eax	; First empty element of the tree.
	mov	dword [eax], 0
	mov	dword [eax+4], 0
.arrayok:
	mov	edi, [esi+TArray.lParam]
	mov	ecx, 32 	; the depth of the tree
	mov	ebx, TArray.Items
.treeseek:
	xor	eax, eax
	ror	[.hash], 1
	adc	eax, 0
	lea	edx, [ebx+4*eax]
	mov	edx, [edi + edx]	; eax is offset from the begining of the array.
	test	edx, edx
	jz	.notfound
	mov	ebx, edx
	loop	.treeseek
.found: 			; the offset of the item is in the ebx register
	mov	eax, [edi+edx]
	lea	eax, [esi + eax]
	pop	edx ebx edi esi
	ret
.addloop:
	xor	eax, eax
	ror	[.hash],1
	adc	eax, 0
.notfound:	; edi - pointer to the tree array,  ebx - offset of the last found element.
		; eax - 0 or 1 depending of the last bit checked. ecx - remaining bits count.
		; edx = 0
	push	ecx eax
	lea	edx, [esi+TArray.lParam]
	stdcall AddArrayItem, edx
	mov	dword [eax], 0
	mov	dword [eax+4], 0
	mov	edi, edx		; new base for the array.
	sub	eax, edx		; offset of the new element
	xchg	eax, [esp]		; this is the old eax
	lea	edx, [ebx+ 4*eax]
	pop	ebx
	mov	[edi+edx], ebx		; the new offset is properly set.
	pop	ecx
	loop	.addloop
	push	eax
	mov	edx, [edi+edx]
	lea	ebx, [edi+edx]	; address of the last element.
	stdcall AddArrayItem, [.ptrVarArray]
	mov	esi, eax
	sub	eax, edx	; offset of the TLabel item.
	pop	ecx
	mov	[ebx+4*ecx], eax  ; stored.
	xor	eax, eax
	mov	[esi+TLabel.Children], eax
	mov	[esi+TLabel.type], al
	mov	[esi+TLabel.flags], ax
	mov	[esi+TLabel.ValueLo], eax
	mov	[esi+TLabel.ValueHi], eax
; This should be the only place where the .hName and .hashName of TLabel is set.
	stdcall AddToNames, [.ptrLabel]
	mov	[esi+TLabel.iName], eax
	push	[.hash]
	pop	[esi+TLabel.hashName]
	mov	eax, esi
	pop	edx ebx edi esi
	ret

endp

;------------------------------------------------------
; Returns: eax - pointer to new created TLabel
;          [hLabelName] content of the string will be
;          destroyed.
;------------------------------------------------------
proc AddToTree uses ebx ecx edx esi edi, .ptrLabelName, .ptrVarTree
	mov	edi, [.ptrVarTree]
	mov	ebx, [.ptrLabelName]
.addloop:
	stdcall GetLabelParent, ebx
	test	edx, edx
	jz	.addhere
	mov	ebx, eax
	stdcall SearchCreateLabel, edi, edx
	lea	edi, [eax+TLabel.Children]
	jmp	.addloop
.addhere:
	stdcall SearchCreateLabel, edi, eax
	ret
endp


proc SearchLabel .ptrVarArray, .hLabel
local .hash:DWORD
	push	esi edi ebx edx
	mov	edx, [.ptrVarArray]
	mov	esi, [edx]
	test	esi, esi
	jz	.notfound
	stdcall StrHash, [.hLabel], 0
	mov	[.hash], eax
	mov	edi, [esi+TArray.lParam]
	mov	ecx, 32 	; the depth of the tree
	mov	ebx, TArray.Items
.treeseek:
	xor	eax, eax
	ror	[.hash], 1
	adc	eax, 0
	lea	edx, [ebx+4*eax]
	mov	edx, [edi + edx]	; eax is offset from the begining of the array.
	test	edx, edx
	jz	.notfound
	mov	ebx, edx
	loop	.treeseek
.found: 			; the offset of the item is in the ebx register
	mov	eax, [edi+edx]
	lea	eax, [esi + eax]
	pop	edx ebx edi esi
	ret
.notfound:
	xor	eax, eax
	pop	edx ebx edi esi
	ret
endp

proc SearchTree .hLabelName, .ptrVarTree
	push	edi ebx ecx
	mov	edi, [.ptrVarTree]
	mov	ebx, [.hLabelName]
.searchloop:
	stdcall GetLabelParent, ebx
	test	edx, edx
	jz	.here
	mov	ebx, eax
	stdcall SearchLabel, edi, edx
	test	eax, eax
	jz	.finish
	lea	edi, [eax+TLabel.Children]
	jmp	.searchloop
.here:	mov	edx, eax
	mov	eax, edi
.finish:
	pop	ecx ebx edi
	ret
endp

; returns pointer to TLabel or NULL
proc SearchTreeExact .hLabelName, .ptrVarTree
	push	edi ebx ecx
	mov	edi, [.ptrVarTree]
	mov	ebx, [.hLabelName]
.searchloop:
	stdcall GetLabelParent, ebx
	test	edx, edx
	jz	.here
	mov	ebx, eax
	stdcall SearchLabel, edi, edx
	test	eax, eax
	jz	.finish
	lea	edi, [eax+TLabel.Children]
	jmp	.searchloop
.here:	stdcall SearchLabel, edi, eax
.finish:
	pop	ecx ebx edi
	ret
endp


proc DestroyLabelsTree .ptrVarArray
	push	esi ebx
	mov	eax, [.ptrVarArray]
	mov	esi, [eax]		; pointer to TArray
	test	esi, esi
	jz	.finish
	mov	ebx, [esi+TArray.Count]
	lea	esi, [esi+TArray.Items]
.clearlist:
	cmp	dword [esi], 0
	je	.next
	stdcall DestroyLabelsTree, esi		; for .Children field of the TLabel
.next:	add	esi, sizeof.TLabel
	dec	ebx
	jnz	.clearlist
	mov	eax, [.ptrVarArray]
	xchg	ebx, [eax]
	invoke	HeapFree, [hHeap], 0, [ebx+TArray.lParam]
	invoke	HeapFree, [hHeap], 0, ebx
.finish:
	cmp	[ptrNames], 0
	je	.quit
	invoke	HeapFree, [hHeap], 0, [ptrNames]
	mov	[ptrNames], 0
.quit:
	pop	ebx esi
	ret
endp


proc CaptureSymbols uses ebx ecx edx esi edi, pLabelsList, pListEnd
locals
  szName rb 256
endl
;        cmp     [.ptrVarArray],0
;        je      .finish
	cmp	[ptrLabels], 0
	je	@f
	stdcall DestroyLabelsTree, ptrLabels
@@:
; Allocate memory for label names.
	invoke	HeapAlloc, [hHeap], 0, cNamesInitialSize
	mov	[ptrNames], eax
	mov	[NamesCurrent], 0
	mov	[NamesSize], cNamesInitialSize
	lea	ebx,[szName]
	mov	edx,[pListEnd]
  .capture_symbols:
	cmp	edx,[pLabelsList]
	jbe	.finish
	sub	edx,LABEL_STRUCTURE_SIZE
	cmp	dword [edx+24],0
	je	.capture_symbols    ; do not dump anonymous symbols
	test	byte [edx+8],1
	jz	.capture_symbols    ; do not dump symbols that didn't get defined
	mov	ax,[current_pass]
	cmp	ax,[edx+16]
	jne	.capture_symbols
	test	byte [edx+8],4
	jnz	.capture_symbols    ; do not dump assembly-time variables
	mov	edi,ebx
	mov	esi,[edx+24]
	movzx	ecx,byte[esi-1]
     .copy_symbol:
	mov	al,[esi]
	cmp	al,'?'
	je	.capture_symbols    ; do not capture local symbols
	mov	[edi],al
	inc	esi
	inc	edi
	dec	ecx
	jnz	.copy_symbol
	mov	byte[edi],0
	stdcall AddToTree, ebx,ptrLabels
	mov	edi,eax
	; move symbol data
	mov	eax,dword[edx+TIntLabel.Value]
	mov	[edi+TLabel.ValueLo],eax
	mov	eax,dword[edx+TIntLabel.Value+4]
	mov	[edi+TLabel.ValueHi],eax
	mov	ax,word[edx+TIntLabel.Flags]
	mov	[edi+TLabel.flags], ax
	mov	al,byte[edx+TIntLabel.type]
	mov	[edi+TLabel.type],al
	mov	al,[edx+TIntLabel.DataSize]
	mov	[edi+TLabel.DataSize],al
	jmp	.capture_symbols
.finish:
	ret
endp
