;
;  DbgServer.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;

;-------------------------------------------------------------------------------
; The idea is to make everything concerning debugger in  one separate thread
; and to command it through some commands.
;-------------------------------------------------------------------------------

;-----------------------------------------------------------
; Global variables for synchronization between Fresh and
; debugger.
;-----------------------------------------------------------
uglobal
  idDbgThread dd ?

  dbgStatus  dd ?

  align 16
  dbgEvent   TDebugEvent

  align 4
  dbgContext CONTEXT
  align 4
  dbgSInfo   STARTUPINFO
  align 4
  dbgPInfo   PROCESS_INFORMATION

  dbgCommand  dd ?
endg



;-----------------------------------------------------------
; Thread procedure for debug server.
;-----------------------------------------------------------
;
; TODO: The debug server have to be implemented to get
;       commands from the message queue of the thread
;       not to poll the command variable.
;       this will ensure low resource consumpsion and
;       will not need wait states while server is busy.
;

proc DebuggerThread .hFileName
; Run the debugged program.
	mov	ecx, sizeof.STARTUPINFO shr 2
	mov	edi, dbgSInfo
	xor	eax, eax
	rep stosd

	mov	[dbgSInfo.cb], sizeof.STARTUPINFO

	stdcall StrPtr, [CompiledFileName]
	invoke	CreateProcess, eax, NULL, NULL, NULL, FALSE,				\
			       NORMAL_PRIORITY_CLASS or DEBUG_PROCESS or DEBUG_ONLY_THIS_PROCESS, \
			       NULL, NULL, dbgSInfo, dbgPInfo
	test	eax,eax
	jz	.StartError

	stdcall SetDbgStatus, dsRunning
	mov	[dbgCommand], 0

	stdcall TDumper.DumpAddress, $4013fd, [dbgPInfo.hProcess]


.debuggerloop:
	invoke	WaitForDebugEvent, dbgEvent, 100
	test	eax, eax
	jnz	.event_raised

.commandloop:
	cmp	[dbgCommand], 0
	je	.debuggerloop

	call	ExecuteCommand

	cmp	[dbgStatus], dsPaused
	je	.commandloop

	cmp	[dbgStatus], dsRunning
	je	.debuggerloop

	xor	eax, eax
	ret

;------------- On debugger event raised --------------------------
.event_raised:
	mov	[dbgCommand], 0
	stdcall SetDbgStatus, dsPaused

	invoke	SendMessage, [hMainWindow], DM_EVENTRAISED, 0, 0	; Notify main window

	mov	[dbgCommand], eax
	call	ExecuteCommand

	cmp	[dbgStatus], dsFinished
	jne	.debuggerloop

	xor	eax, eax
	ret

.StartError:
	mov	eax, 1
	ret
endp


ExecuteCommand:
	cmp	[dbgCommand], 0
	jne	.command
	ret

.command:
	mov	ebx, [dbgCommand]

.docommand:
	cmp	ebx, dbgcContinue
	je	.continue

	cmp	ebx, dbgcContinueNotHandled
	je	.continuenh

	cmp	ebx, dbgcKillProcess
	je	.killprocess

	cmp	ebx, dbgcFinish
	je	.finish

	cmp	ebx, dbgcPause
	je	.pause

.endcommand:
	mov	[dbgCommand], 0
	ret

.continue:
	invoke	ContinueDebugEvent, [dbgEvent.dwProcessId], [dbgEvent.dwThreadId], DBG_CONTINUE
	stdcall SetDbgStatus, dsRunning
	jmp	.endcommand

.continuenh:
	invoke	ContinueDebugEvent, [dbgEvent.dwProcessId], [dbgEvent.dwThreadId], DBG_EXCEPTION_NOT_HANDLED
	stdcall SetDbgStatus, dsRunning
	jmp	.endcommand

.getcontext:
	invoke	GetThreadContext, [dbgPInfo.hThread], dbgContext
	jmp	.endcommand

.pause:
	invoke	SuspendThread, [dbgPInfo.hThread]
	mov	[dbgContext.ContextFlags], CONTEXT_CONTROL
	invoke	GetThreadContext, [dbgPInfo.hThread], dbgContext
	or	[dbgContext.regFlag], $100
	invoke	SetThreadContext, [dbgPInfo.hThread], dbgContext
	invoke	ResumeThread, [dbgPInfo.hThread]
	jmp	.endcommand


.killprocess:
	invoke	TerminateProcess, [dbgPInfo.hProcess], -1
	mov	[dbgEvent.dwDebugEventCode], EXIT_PROCESS_DEBUG_EVENT
	mov	[dbgEvent.ExitProcess.dwExitCode], -1
	invoke	SendMessage, [hMainWindow], DM_EVENTRAISED, 0, 0

.finish:
	invoke	ContinueDebugEvent, [dbgEvent.dwProcessId], [dbgEvent.dwThreadId], DBG_CONTINUE
	stdcall SetDbgStatus, dsFinished

	invoke	CloseHandle, [dbgPInfo.hThread]
	invoke	CloseHandle, [dbgPInfo.hProcess]

	ret



proc SetDbgStatus .status
	mov	eax, [.status]
	cmp	[dbgStatus], eax
	je	.finish

	mov	[dbgStatus], eax
	invoke	PostMessage, [hMainWindow], DM_STATUSCHANGED, 0, 0

.finish:
	ret
endp