;
;  Debug.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;

uglobal
align 4
  LastAddr	dd ?
  LastPtrLine	dd ?
  LastVirt	db ?

align 4
  ptrLastFilename dd ?
  iLastFilename   dd ?

  ptrDebugInfo	dd ?   ; Dynamic array with TLineInfo elements.
  ptrFileNames	dd ?   ; Dynamic array containing TDebugFiles elements.
  ptrListing	dd ?   ; Pointer to memory containing listing. The first dword of the array
		       ; contains current capacity of the

  AsmLines	dd ?
  TotalLines	dd ?
endg

cInitListingSize = 65536

proc InitDebugInfo
	stdcall ClearDebugInfo
	stdcall CreateArray, sizeof.TDebugLine
	mov	[ptrDebugInfo], eax
	invoke	HeapAlloc, [hHeap], HEAP_ZERO_MEMORY, 32768 * 2
	mov	[ptrFileNames], eax
	invoke	HeapAlloc, [hHeap], 0, cInitListingSize
	mov	[ptrListing], eax
	mov	dword [eax], cInitListingSize
	mov	dword [eax+4], 8
	ret
endp



proc ClearDebugInfo
	stdcall DestroyDebugInfo, ptrDebugInfo
	stdcall DestroyFilesList, ptrFileNames
	ret
endp


proc DestroyDebugInfo .ptrVarArray
	push	esi ebx
	mov	esi, [.ptrVarArray]
	cmp	dword [esi], 0
	je	.finish
	xor	ebx, ebx
	xchg	ebx, [esi]
	test	ebx, ebx
	jz	.debugok
	invoke	HeapFree, [hHeap], 0, ebx
.debugok:
	invoke	HeapFree, [hHeap], 0, [ptrListing]
.finish:
	pop	ebx esi
	ret
endp



proc DestroyFilesList .ptrVarArray
	push	esi ebx

	mov	esi, [.ptrVarArray]
	mov	esi, [esi]
	test	esi, esi
	jz	.finish

	mov	ebx, 32768

.freeloop:
	movzx	eax, word [esi]
	add	esi, 2
	test	eax, eax
	jz	.next
	stdcall StrDel, eax
.next:
	dec	ebx
	jnz	.freeloop

.endloop:
	mov	esi, [.ptrVarArray]
	xchg	ebx, [esi]
	invoke	HeapFree, [hHeap], 0, ebx

.finish:
	pop	ebx esi
	ret
endp


proc FileNameSearchAdd .hFileName
	push	esi edi ebx

	stdcall HashFilename, [.hFileName]
	mov	edi, eax
	and	eax, $7fff
	shr	edi, 17
	xor	edi, eax	; direct index in the filenames array.

	mov	esi, [ptrFileNames]

	movzx	eax, word [esi+2*edi]
	test	eax, eax
	jnz	.found

	stdcall StrNew
	mov	[esi+2*edi], ax
	stdcall StrCopy, eax, [.hFileName]
.found:
	mov	eax, edi
	pop	ebx edi esi
	ret
endp


;----------------------------------------------------------------------
; This procedure make the additional last pass in assembly stage
; due to create a list with debug information for source level
; debuger.
;----------------------------------------------------------------------
proc lastpass
	mov	[TotalLines], 0
	mov	[AsmLines], 0
	inc	[current_pass]
	mov	eax,[labels_list]
	mov	[display_buffer],eax
	mov	eax,[additional_memory]
	mov	[free_additional_memory],eax
	mov	eax,[additional_memory_end]
	mov	[structures_buffer],eax
	mov	esi,[source_start]
	mov	edi,[code_start]
	xor	eax,eax
	mov	dword [org_origin],edi
	mov	dword [org_origin+4], eax
	mov	[org_start],edi
	mov	[org_registers],eax
	mov	[org_symbol],eax
	mov	[error_line],eax
	mov	[counter],eax
	mov	[format_flags],eax
	mov	[number_of_relocations],eax
	mov	[undefined_data_end],eax
	mov	[next_pass_needed],al
	mov	[output_format],al
	mov	[virtual_data],al
	mov	[code_type],16

	mov	[LastPtrLine], 0
	mov	[LastAddr], -1

	mov	[ptrLastFilename], 0
	mov	[iLastFilename], 0

.pass_loop:
	mov	eax, edi				 ; Capturing the address of the current
	mov	ecx, [current_line]
	mov	dl, [virtual_data]
	sub	eax, dword [org_origin] 		 ; row of the program.
	mov	[LastPtrLine], ecx
	mov	[LastVirt], dl
	mov	[LastAddr], eax

	call	assemble_line
	pushf

	inc	[TotalLines]

	mov	eax, edi
	mov	dl, [virtual_data]
	sub	eax, dword [org_origin]
	add	dl, [LastVirt]
	jnz	.nodata
	cmp	eax, [LastAddr]
	je	.nodata

; Get the current line...
	push	ebx edi
	mov	ebx, [current_line]

; For description of internal line header format, read "source/doc/FASM internals.ASM"
; Get the real compiled line:
	stdcall CreateLineString, ebx
	push	eax

.findbaseline:
	mov	eax, [ebx+4]	; Line number in the source file. If MSB is 1 then
				; it is a line number in the macro definition.
	test	eax, eax
	jns	.lineok

	mov	ebx, [ebx+8]	; get the line where the macro was invoked.
	jmp	.findbaseline

.lineok:
	stdcall AddArrayItem, ptrDebugInfo
	mov	edi, eax

	pop	[edi+TDebugLine.strLine]
	mov	eax, [LastAddr]
	mov	ecx, dword [ebx+4] ; line number.
	mov	[edi+TDebugLine.Address], eax
	mov	[edi+TDebugLine.LineNum], cx

	mov	eax, [iLastFilename]
	mov	ecx, [ebx]
	cmp	[ptrLastFilename], ecx
	je	.filenameok

	mov	[ptrLastFilename], ecx
	stdcall FileNameSearchAdd, ecx	     ; filename, returns index in the filename array.
	mov	[iLastFilename], eax

.filenameok:
	mov	[edi+TDebugLine.iFile], ax
	pop	edi ebx

	inc	[AsmLines]

.nodata:
	popf
	jnc	.pass_loop

.endpass:
	mov	eax,[structures_buffer]
	cmp	eax,[additional_memory_end]
	jne	unexpected_end_of_file
	call	close_pass
	mov	eax,[labels_list]
	jmp	check_symbols
endp




;------------------------------------------------------------
; Returns: ebx - base addr
;          edi - offset of the string with max size $1000.
;------------------------------------------------------------

proc GetStrListPtr .ptrSLVar
local .newsz:DWORD
	push	esi

	mov	esi, [.ptrSLVar]
	mov	ebx, [esi]		; pointer to the array

	mov	edi, [ebx+TStringList.offset]
	mov	eax, [ebx+TStringList.size]
	sub	eax, edi
	cmp	eax, $1000	; 4k tolerance
	jge	.sizeok

	mov	eax, [ebx+TStringList.size]
	shr	eax,2
	add	eax, [ebx+TStringList.size]
	push	eax

	invoke	HeapReAlloc, [hHeap], 0, ebx, eax
	mov	ebx, eax
	mov	[esi], eax
	popd	[ebx]

.sizeok:
	pop	esi
	ret
endp


proc CreateLineString .ptrLine
local .char:DWORD
	push	esi edi ebx

	stdcall GetStrListPtr, ptrListing
	push	edi				; this will be returned.
	add	edi, ebx

	mov	esi, [.ptrLine]
	add	esi, 16 	  ; the size of line header.

	mov	[.char], $09

.getloop:
	lodsb
	test	al, al
	jz	.eol
	cmp	al, $1a 	; prefix of token
	je	.token

	stosb
	cmp	al,','
	jne	.getloop
	mov	al,$20
	stosb
	jmp	.getloop

.token:
	movzx	ecx, byte [esi]
	inc	esi
	rep movsb		; copy the token to the output string.

	cmp	byte [esi], '['
	je	.tab
	cmp	byte [esi], $1a
	jne	.getloop
.tab:
	mov	al, byte [.char]
	mov	byte [.char], $20
	stosb
	jmp	.getloop

.eol:
	stosb

	sub	edi, ebx
	mov	[ebx+TStringList.offset], edi

	pop	eax
	pop	ebx edi esi
	ret
endp




;--------------------------------------------------------------
; Find the file and line of the source, where some address
; is generate.
; Returns:
;   eax: handle of filename string - 0 means not found
;   edx: pointer to the TDebugLine record.
;--------------------------------------------------------
proc FindAddressPosition .addr
	push	esi edi ebx

	mov	esi, [ptrDebugInfo]
	mov	edi, [ptrFileNames]

	mov	ebx, [esi+TArray.Count]
	lea	esi, [esi+TArray.Items]

	mov	eax, [.addr]

.search:
	cmp	eax, [esi+TDebugLine.Address]
	jbe	.found
	add	esi, sizeof.TDebugLine
	dec	ebx
	jnz	.search

.notfound:
	xor	eax, eax
	xor	edx, edx
	jmp	.finish

.found:
	movzx	eax, word [esi+TDebugLine.iFile]
	and	eax, $7fff
	movzx	eax, word [edi+2*eax]
	mov	edx, esi

.finish:
	pop	ebx edi esi
	ret
endp

;--------------------------------------------------
; Positions the editor to the file/line
;--------------------------------------------------
uglobal
  fSetCurrent dd ?
endg


proc PosEditor .hFileName, .LineNum
local .aepos:AEPOS
	push	eax ebx ecx edx
	stdcall ConvertPath, [.hFileName]
	push	eax
	stdcall OpenFileByName, eax, NULL, TRUE
	stdcall StrDel ; from the stack.
	test	eax, eax
	jz	.finish
	mov	ebx, [eax+TOpenFile.hEditor]
	mov	[fSetCurrent], ebx
	invoke	SendMessage, ebx, AEM_GETFLAGS,[.LineNum],0
	or	eax,FLAG_ACTIVE
	invoke	SendMessage, ebx, AEM_SETFLAGS,[.LineNum],eax
;        invoke  RedrawWindow, ebx, NULL, NULL, RDW_UPDATENOW

; workaround, because AEM_SETFOCUSLINE do not move the caret.
	mov	[.aepos.selectionPosition], 1
	mov	[.aepos.caretPosition], 1
	mov	eax, [.LineNum]
	mov	[.aepos.selectionLine], eax
	mov	[.aepos.caretLine], eax
	lea	eax, [.aepos]
	invoke	SendMessage, ebx, AEM_SETPOS, eax, 0

; Work around, because of some bug in AsmEdit, that make caret to dissapear, even when
; the control have an input focus.
	invoke	SetFocus, 0
	invoke	SetFocus, ebx

	mov	[fSetCurrent], 0
.finish:
	pop	edx ecx ebx eax
	ret
endp



;----------------------------------------------------------
; Returns the address in binary of some line from file
;----------------------------------------------------------
proc FindAddress .hFilename, .linenum
   ret
endp