;
;  Debugger.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;

uglobal
  hDebugThread dd ?
  hDebugFileImage dd ?
endg

; helper routine for running applications with
; programs such debuggers/emulators
proc RunIndirect .pszApplication
local .sinfo:STARTUPINFO
local .pinfo:PROCESS_INFORMATION
	stdcall FileExists, [.pszApplication]
	test	eax,eax
	jz	.finish
	mov	ecx, sizeof.STARTUPINFO/4
	lea	edi,[.sinfo]
	xor	eax, eax
	rep	stosd
	mov	[.sinfo.cb], sizeof.STARTUPINFO
	mov	[.sinfo.dwFlags], STARTF_USESHOWWINDOW
	mov	[.sinfo.wShowWindow], SW_HIDE
	lea	ecx,[.pinfo]
	lea	eax,[.sinfo]
	push	ecx eax
	stdcall StrDup, [.pszApplication]
	mov	ebx,eax
	stdcall StrCharCat, ebx,' '
	stdcall StrCat, ebx, [CompiledFileName]
	stdcall StrPtr,ebx
	invoke	CreateProcess,NULL,eax,NULL,NULL,FALSE,NORMAL_PRIORITY_CLASS ,NULL,NULL
	test	eax,eax
	jnz	.ok
	stdcall LastError
   .ok:
	invoke	CloseHandle,[.pinfo.hThread]
	invoke	CloseHandle,[.pinfo.hProcess]
   .finish:
	ret
endp


proc OnRun .wparam, .lparam
	cmp	[idDbgThread], 0
	jne	.rundebugged

	stdcall OnCompile, 0, 0

	cmp	eax, MR_OK
	jne	.finish 	; There are errors during compilation.

	cmp	[output_format],5
	je	.runELF

	cmp	[output_format],4
	jae	.runObject

	cmp	[output_format], 3
	jne	.dos

	invoke	SendMessage, [frmMsg], MWM_CLEARMESSAGES, 0, 0

	invoke	CreateThread, NULL, $4000, DebuggerThread, [output_file], 0, idDbgThread
	mov	[hDebugThread], eax
	test	eax, eax
	jz	.errorthread
	ret

.errorthread:
	call	LastError
.finish:
	ret

.runObject:
	invoke	MessageBox, [hApplication], cRunObjectError, NULL, MB_ICONERROR+MB_OK
	ret

.runELF:
	stdcall GetPref, prefLine,0
	mov	ebx,eax
	stdcall FileExists, eax
	test	eax,eax
	jz	.noemulator
	stdcall RunIndirect, ebx
	stdcall StrDel, ebx
	ret

.noemulator:
	invoke	MessageBox, [hApplication], cRunLinuxError, NULL, MB_ICONERROR+MB_OK
	ret

constStr cRunLinuxError, "Linux application can't be started, because Linux emulator is not set properly in Options|IDE options"

.dos:
	stdcall OnRunDirect, 3, 0
	ret

.rundebugged:
	cmp	[dbgStatus], dsPaused
	jne	.finish

	call	ResetSingleStep
	ret
endp

iglobal
  cRunObjectError db "Can't execute object file.",0
endg


; --------------------------------------------------------------
; if wparam = 3 then the source is compiled, so simply run it.
; --------------------------------------------------------------
proc OnRunDirect .wparam, .lparam
local .sinfo:STARTUPINFO
local .pinfo:PROCESS_INFORMATION
	cmp	[.wparam], 3
	je	.compileok

	stdcall OnCompile, 0, 0
	cmp	eax, MR_OK
	jne	.finish 	; There are errors during compilation.

.compileok:
	cmp	[output_format],5
	je	OnRun.runELF

	cmp	[output_format],4
	jae	OnRun.runObject

	mov	ecx, sizeof.STARTUPINFO
	shr	ecx, 2
	lea	edi,[.sinfo]
	xor	eax, eax
	rep stosd
	mov	[.sinfo.cb], sizeof.STARTUPINFO
	mov	[.sinfo.dwFlags],0
	lea	ecx,[.pinfo]
	lea	eax,[.sinfo]
	push	ecx eax
	stdcall StrPtr, [CompiledFileName]
	invoke	CreateProcess,eax,NULL,NULL,NULL,FALSE,NORMAL_PRIORITY_CLASS,NULL,NULL
	test	eax,eax
	jnz	.ok

	stdcall LastError
.ok:
	invoke	CloseHandle,[.pinfo.hThread]
	invoke	CloseHandle,[.pinfo.hProcess]

.finish:
	ret
endp





proc OnStepInto .wparam, .lparam
	cmp	[dbgStatus], dsPaused
	jne	.finish

	call	SetSingleStep
.finish:
	ret
endp


proc OnStepOver .wparam, .lparam
	cmp	[dbgStatus], dsPaused
	jne	.finish

	call	SetSingleStep
.finish:
	ret
endp


proc OnPause .wparam, .lparam
	call	WaitForCommandOrTimeout
	jc	@f
	mov	[dbgCommand], dbgcPause
@@:
	ret
endp


proc OnStop .wparam, .lparam
	call	 WaitForCommandOrTimeout
	jc	 @f
	mov	 [dbgCommand], dbgcKillProcess
@@:
	ret
endp



proc SetSingleStep
	mov	[dbgContext.ContextFlags], CONTEXT_CONTROL
	invoke	GetThreadContext, [dbgPInfo.hThread], dbgContext
	or	[dbgContext.regFlag], $100
	invoke	SetThreadContext, [dbgPInfo.hThread], dbgContext
	call	WaitForCommandOrTimeout
	jc	@f
	mov	[dbgCommand], dbgcContinue
@@:
	ret
endp



proc ResetSingleStep
	mov	[dbgContext.ContextFlags], CONTEXT_CONTROL
	invoke	GetThreadContext, [dbgPInfo.hThread], dbgContext
	and	[dbgContext.regFlag], $fffffeff
	invoke	SetThreadContext, [dbgPInfo.hThread], dbgContext
	call	WaitForCommandOrTimeout
	jc	@f
	mov	[dbgCommand], dbgcContinue
@@:
	ret
endp



iglobal
cDebugNumberHex = ntsHex or ntsUnsigned or ntsFixedWidth or 8
DebuggerProperties:
      PropInfo									    \
	'EAX', ptDword, CONTEXT.regEax, IntPropEditor,	  cDebugNumberHex, 0,	    \
	'EBX', ptDword, CONTEXT.regEbx, IntPropEditor,	  cDebugNumberHex, 0,	    \
	'ECX', ptDword, CONTEXT.regEcx, IntPropEditor,	  cDebugNumberHex, 0,	    \
	'EDX', ptDword, CONTEXT.regEdx, IntPropEditor,	  cDebugNumberHex, 0,	    \
	'ESI', ptDword, CONTEXT.regEsi, IntPropEditor,	  cDebugNumberHex, 0,	    \
	'EDI', ptDword, CONTEXT.regEdi, IntPropEditor,	  cDebugNumberHex, 0,	    \
	'EBP', ptDword, CONTEXT.regEbp, IntPropEditor,	  cDebugNumberHex, 0,	    \
	'ESP', ptDword, CONTEXT.regEsp, IntPropEditor,	  cDebugNumberHex, 0,	    \
	'EIP', ptDword, CONTEXT.regEip, IntPropEditor,	  cDebugNumberHex, 0,	    \
	'flags', ptDword, CONTEXT.regFlag, IntPropEditor, cDebugNumberHex, 0
;        'CS:', ptDword, CONTEXT.regCs, IntPropEditor,    cDebugNumberHex, 0,       \
;        'SS:', ptDword, CONTEXT.regSs, IntPropEditor,    cDebugNumberHex, 0,       \
;        'DS:', ptDword, CONTEXT.regDs, IntPropEditor,    cDebugNumberHex, 0,       \
;        'ES:', ptDword, CONTEXT.regEs, IntPropEditor,    cDebugNumberHex, 0,       \
;        'FS:', ptDword, CONTEXT.regFs, IntPropEditor,    cDebugNumberHex, 0,       \
;        'GS:', ptDword, CONTEXT.regGs, IntPropEditor,    cDebugNumberHex, 0
endg



proc WaitForCommandOrTimeout

.wait:
	cmp	[dbgCommand], 0
	je	.finish

	invoke	GetCurrentThread
	invoke	WaitForSingleObject, eax, 100

	cmp	[dbgCommand], 0
	jne	.timeout

.finish:
	clc
	ret
.timeout:
	stc
	ret
endp



proc UpdateThreadContext .ptrContext
	cmp	[dbgStatus], dsPaused
	jne	@f

	mov	[dbgContext.ContextFlags], CONTEXT_FULL
	invoke	SetThreadContext, [dbgPInfo.hThread], [.ptrContext]
@@:
	ret
endp




iglobal
  TblDebugEvents dd  DbgEventUnknown
		 dd  DbgEventException
		 dd  DbgEventCreateThread
		 dd  DbgEventCreateProcess
		 dd  DbgEventExitThread
		 dd  DbgEventExitProcess
		 dd  DbgEventLoadDll
		 dd  DbgEventUnloadDll
		 dd  DbgEventOutputString
		 dd  DbgEventRIP

  cDbgEventUnknown	  db	  'Unknown debug event.', 0
  cDbgEventException	  db	  'Exception:',0
  cDbgEventCreateThread   db	  'New thread created',0
  cDbgEventCreateProcess  db	  'New process created',0
  cDbgEventExitThread	  db	  'Thread terminated',0
  cDbgEventExitProcess	  db	  'Process terminated',0
  cDbgEventLoadDll	  db	  'DLL loaded: ',0
  cDbgEventUnloadDll	  db	  'DLL unloaded: ',0
  cDbgEventOutputString   db	  'String output: ',0
  cDbgEventRIP		  db	  'Internal debuger error.',0
endg


proc DbgEventUnknown
	stdcall StrNew
	stdcall StrCopy, eax, cDbgEventUnknown
	ret
endp


proc DbgEventException
	mov	ecx, [dbgEvent.Exception.ExceptionRecord.ExceptionCode]
	cmp	ecx, EXCEPTION_SINGLE_STEP
	jne	.logexception
	xor	eax, eax
	ret

.logexception:
	stdcall StrNew
	stdcall StrCopy, eax, cDbgEventException

	stdcall GetExceptionText, ecx
	stdcall StrCat, eax, ecx
	push	eax
	stdcall NumToStr, [dbgEvent.Exception.ExceptionRecord.ExceptionAddress], cExceptionAt, 5, ntsHex or ntsFixedWidth or ntsUnsigned or 8
	pop	eax
	stdcall StrCat, eax, cExceptionAt
	ret
endp


proc GetExceptionText .number
	push	eax esi
	mov	esi, tblExceptions
.loop:
	mov	eax, dword [esi]
	test	eax, eax
	jz	.notfound

	cmp	eax, [.number]
	je	.found

	add	esi, 4
	jmp	.loop

.found:
	mov	ecx, [esi+4]
	jmp	.finish

.notfound:
	stdcall NumToStr, [.number], cExceptionNum, 1, ntsHex or ntsUnsigned or ntsFixedWidth or 8
	mov	ecx, cExceptionNum

.finish:
	pop	esi eax
	ret
endp


macro KeyString [key, string] {
  forward
    local  ptrstr
    dd	   key, ptrstr
  common
    dd	0
  forward
ptrstr db  string, 0
}


iglobal

cExceptionAt  db ' at $00000000',0
cExceptionNum db '$00000000',0

tblExceptions:
     KeyString								  \
	  EXCEPTION_ACCESS_VIOLATION,		'Access violation',	  \
	  EXCEPTION_DATATYPE_MISALIGNMENT,	'Datatype misalignment',  \
	  EXCEPTION_BREAKPOINT, 		'Breakpoint',		  \
	  EXCEPTION_SINGLE_STEP,		'Single step',		  \
	  EXCEPTION_ARRAY_BOUNDS_EXCEEDED,	'Array bounds exceeded',  \
	  EXCEPTION_FLT_DENORMAL_OPERAND,	'FPU Denormal operand',   \
	  EXCEPTION_FLT_DIVIDE_BY_ZERO, 	'FPU Divide by zero',	  \
	  EXCEPTION_FLT_INEXACT_RESULT, 	'FPU Inexact result',	  \
	  EXCEPTION_FLT_INVALID_OPERATION,	'FPU Invalid operation',  \
	  EXCEPTION_FLT_OVERFLOW,		'FPU Overflow', 	  \
	  EXCEPTION_FLT_STACK_CHECK,		'FPU Stack check',	  \
	  EXCEPTION_FLT_UNDERFLOW,		'FPU Underflow',	  \
	  EXCEPTION_INT_DIVIDE_BY_ZERO, 	'Divide by zero',	  \
	  EXCEPTION_INT_OVERFLOW,		'Overflow',		  \
	  EXCEPTION_ILLEGAL_INSTRUCTION,	'Illegal instruction',	  \
	  EXCEPTION_PRIV_INSTRUCTION,		'Privileged instruction', \
	  EXCEPTION_IN_PAGE_ERROR,		'Page error'
endg





proc DbgEventCreateThread
	stdcall StrNew
	stdcall StrCopy, eax, cDbgEventCreateThread
	ret
endp

proc DbgEventCreateProcess
	stdcall StrNew
	stdcall StrCopy, eax, cDbgEventCreateProcess
	ret
endp

proc DbgEventExitThread
	stdcall StrNew
	stdcall StrCopy, eax, cDbgEventExitThread
	ret
endp

proc DbgEventExitProcess

	stdcall StrNew
	stdcall StrCopy, eax, cDbgEventExitProcess
	ret
endp


proc DbgEventLoadDll
locals
.name rb 512
.buffer rd 128
endl
	stdcall StrNew
	mov	edi, eax
	stdcall StrCopy, edi, cDbgEventLoadDll

	lea	ecx, [.buffer]
	invoke	ReadProcessMemory, [dbgPInfo.hProcess], [dbgEvent.LoadDll.lpImageName], ecx, 4, NULL
	test	eax, eax
	jz	.finish

	mov	esi, [.buffer]
	test	esi, esi
	jz	.finish

	lea	ecx, [.buffer]
	invoke	ReadProcessMemory, [dbgPInfo.hProcess], esi, ecx, 512, NULL
	test	eax, eax
	jz	.finish

	push	edi

	lea	esi, [.buffer]
	lea	edi, [.name]
	xor	ecx, ecx
	cmp	[dbgEvent.LoadDll.fUnicode], 0
	jz	@f
	inc	ecx
@@:
	lodsb
	stosb
	add	esi, ecx
	test	al,al
	jnz	@b

	pop    edi

	lea	eax, [.name]
	stdcall StrCat, edi, eax

.finish:
	mov	eax, edi
	ret
endp

proc DbgEventUnloadDll
	stdcall StrNew
	stdcall StrCopy, eax, cDbgEventUnloadDll
	ret
endp

proc DbgEventOutputString
	stdcall StrNew
	stdcall StrCopy, eax, cDbgEventOutputString
	stdcall StrCat, eax, [dbgEvent.DebugString.lpDebugStringData]
	ret
endp

proc DbgEventRIP
	stdcall StrNew
	stdcall StrCopy, eax, cDbgEventRIP
	ret
endp



;---------------------------------------------------------------------
; Goto address dialog form.
;---------------------------------------------------------------------
iglobal
  Window dlgGotoAddr, $3, $0, 'TForm', 'Goto address', $6C80000, $10001, 0, 373, 322, 159, 113, NULL
  Window NONE, $0, $0, 'Static', 'Input the address:', $50000000, $0, 0, 16, 8, 88, 13, NULL
  Window NONE, $0, $0, 'Edit', 'h', $50800080 or WS_TABSTOP, $0, 100, 16, 24, 121, 18, NULL
  Window NONE, $2, $0, 'Button', 'Go', $50000001 or WS_TABSTOP, $0, MR_OK, 40, 56, 64, 24, NULL

  cAddressNotFound db 'Address not found. Check the value.', 0
endg

;---------------------------------------------------------------------
; Goto address function action procedure.
;---------------------------------------------------------------------
proc OnGotoAddr .wparam, .lparam
locals
.pos AEPOS
.buff rb 256
endl
	cmp	[ptrDebugInfo],0
	je	.finish

	stdcall CreateForm, dlgGotoAddr, [hApplication]
	stdcall ShowModal, ebx, MSF_CENTER
	cmp	eax, MR_OK
	jne	.finish
	lea	esi, [.buff]
	invoke	SendDlgItemMessage, ebx, 100, WM_GETTEXT, 256, esi
	invoke	DestroyWindow, ebx
	stdcall StrToFasmNum, esi
	stdcall FindAddressPosition, eax
	test	eax, eax
	jz	.notfound
	movzx	ecx, [edx+TDebugLine.LineNum]
	stdcall PosEditor, eax, ecx
.finish:
	ret
.notfound:
	invoke	MessageBox, [hApplication], cAddressNotFound, NULL, MB_OK or MB_ICONERROR
	ret
endp