;
;  Dumper.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;

uglobal
  hSplitterDumper dd ?
  Dumper  TDumper
  hDumperPopupMenu dd ?
endg

cDefaultDumperFontWidth = 6
cDefaultDumperFontHeight = 8

iglobal
  cDefaultDumperFont db 'Terminal', 0

  CoolMenu  DumperPopupMenu,	\
    mfNormal, actDumpAddress
endg

proc InitDumperWindow .hParent
local .DumperFont:DWORD
	xor	ebx, ebx

	invoke	CreateFont, cDefaultDumperFontHeight, cDefaultDumperFontWidth, ebx, ebx, FW_DONTCARE,		     \
			    ebx, ebx, ebx,						     \
			    1, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,     \
			    FIXED_PITCH or FF_DONTCARE, cDefaultDumperFont
	mov	esi, eax

	invoke	 CreateWindowEx, ebx, cDataGridClassName, NULL, 		 \
		 WS_CHILD or WS_HSCROLL or WS_VSCROLL,				 \
		 ebx, ebx , 400, ebx,						 \
		 [.hParent],							 \
		 ebx, [hInstance],						 \
		 Dumper

	mov	ebx, eax

	invoke	SendMessage, ebx, WM_SETFONT, esi, FALSE
	stdcall SetAlign, ebx, waLeft
	stdcall SubclassWindow, ebx, DumperWinProc

	mov	[Dumper.Flags], dgfGridLinesFixed
	mov	[Dumper.OnDrawItem], TDumper.DrawCell

	stdcall TDataGrid.SetCounts, Dumper, 18, 100
	stdcall TDataGrid.SetColumnsWidth, Dumper, 1, 16, 16
	stdcall TDataGrid.SetColumnsWidth, Dumper, 17, 17, 100

	invoke	CreateWindowEx, 0, cSplitterClassName, NULL,		\
		WS_CLIPSIBLINGS or WS_CHILD,				\
		0, 0, 0, 0, [.hParent], NULL, [hInstance], [Dumper.hwnd]
	mov	[hSplitterDumper], eax
	invoke	SendMessage, eax, SPM_SETCURSORS, resVSplit, resHSplit
	invoke	SendMessage, [hSplitterDumper], SPM_SETMAXMIN, 16, 9999

	stdcall CreateCoolMenu, DumperPopupMenu, TRUE
	mov	[hDumperPopupMenu], eax

	ret
endp



winproc DumperWinProc

ondefault
	stc
	ret

onmessage WM_DESTROY
	invoke	DestroyMenu, [hDumperPopupMenu]

	cmp	[Dumper.ptrData], 0
	je	.ondefault

	invoke	HeapFree, [hHeap], 0, [Dumper.ptrData]
	stc
	ret

onmessage WM_RBUTTONDOWN
locals
  .pnt POINT
endl
	lea	eax, [.pnt]
	invoke	GetCursorPos, eax

	invoke	TrackPopupMenu, [hDumperPopupMenu], TPM_LEFTBUTTON or TPM_LEFTBUTTON,	\
		[.pnt.x],[.pnt.y], 0, [hEditorsHost], NULL
	xor	eax, eax
	clc
	ret


onmessage DMM_SHOW
	invoke	ShowWindow, [hSplitterDumper], [wparam]
	invoke	ShowWindow, [hwnd], [wparam]
	invoke	GetParent, [hwnd]
	stdcall AlignChildren, eax
	clc
	ret
endwp



proc OnDumpAddress .wparam, .lparam
local .buff[0x100]:BYTE
  iglobal
    cDumpTitle db 'Memory dump', 0
  endg

	stdcall CreateForm, dlgGotoAddr, [hApplication]

	invoke	SendMessage, ebx, WM_SETTEXT, 0, cDumpTitle

	stdcall ShowModal, ebx, MSF_CENTER
	cmp	eax, MR_OK
	jne	.finish

	lea	esi, [.buff]
	invoke	SendDlgItemMessage, ebx, 100, WM_GETTEXT, 256, esi

	invoke	DestroyWindow, ebx

	stdcall StrToFasmNum, esi
	stdcall TDumper.DumpAddress, eax, [dbgPInfo.hProcess]

	invoke	InvalidateRect, [Dumper.hwnd], 0, 0
	mov	[Dumper.fShadowReady], 0

.finish:
	ret
endp





; todo: Check if the [.addr] is inside existing buffer and to not allocate
; new buffer.

proc TDumper.DumpAddress .addr, .hProcess
local .buffer:DWORD
	push	esi edi ebx

	mov	esi, [.addr]
	and	esi, $fffff000	; address of the begining of the page.
	mov	[Dumper.pointer], esi

; determine the size of the memory block.
	lea	edi, [.buffer]
	lea	ebx, [esi-4096]
.getsz:
	add	ebx, 4096
	invoke	ReadProcessMemory, [.hProcess], ebx, edi, 1, NULL
	test	eax, eax
	jnz	.getsz

; set grid size.
	sub	ebx, esi

	mov	ecx, ebx
	sar	ecx, 4
	inc	ecx
	stdcall TDataGrid.SetCounts, Dumper, [Dumper.ColCount], ecx

	cmp	[Dumper.ptrData], 0
	je	.bufferok

	invoke	HeapFree, [hHeap], 0, [Dumper.ptrData]
	mov	[Dumper.ptrData], 0

.bufferok:
; allocate memory buffer
	invoke	HeapAlloc, [hHeap], 0, ebx
	test	eax, eax
	jz	.dataok

	mov	edi, eax

; read memory from the debugged process.
	invoke	ReadProcessMemory, [.hProcess], esi, edi, ebx, NULL
	test	eax, eax
	jz	.dataok

	mov	[Dumper.ptrData], edi
	mov	[Dumper.DataSize], ebx

.dataok:
; Set focused cell.
	mov	eax, [.addr]
	mov	ecx, eax
	sub	eax, [Dumper.pointer]
	and	ecx, $0f
	sar	eax, 4
	inc	ecx
	inc	eax

	mov	[Dumper.focused.x], ecx
	mov	[Dumper.focused.y], eax

	stdcall TDataGrid.GetCellXY, Dumper, ecx, eax
	push	edx

	invoke	SetScrollPos, [Dumper.hwnd], SB_HORZ, eax, TRUE
	pop	edx
	invoke	SetScrollPos, [Dumper.hwnd], SB_VERT, edx, TRUE

	invoke	InvalidateRect, [Dumper.hwnd], NULL, FALSE
	pop	ebx edi esi
	ret
endp




proc TDumper.DrawCell .hdc, .ptrRect, .Col, .Row, .dummy, .fSelected, .fFixed
	push	esi ebx

; Choose text color
	mov	eax, [edi+TDataGrid.clTextFixed]
	cmp	[.fFixed], 0
	jne	.colorok

	mov	eax, [edi+TDataGrid.clTextSelected]
	cmp	[.fSelected], FALSE
	jne	.colorok

	mov	eax, [edi+TDataGrid.clTextCell]

.colorok:
	invoke	SetTextColor, [.hdc], eax

	cmp	[.fFixed], 0
	jne	.fixed

; Data...maybe
	cmp	[Dumper.ptrData], 0
	je	.drawok

	cmp	[.Col], 17	; ASCII field
	je	.ascii

; yes it is data cell
	mov	eax, [.Row]
	mov	ecx, [.Col]
	dec	eax
	dec	ecx
	shl	eax, 4
	add	ecx, [Dumper.ptrData]

	movzx	eax, byte [eax+ecx]

	mov	ah, al
	and	al, $0f
	shr	ah, 4

	cmp	al,$0a
	sbb	al,$69
	das

	xchg	al, ah

	cmp	al,$0a
	sbb	al,$69
	das

	mov	dword [DumperAddress], eax
	mov	eax, DumperAddress
	jmp	.drawit

.ascii:
	push	edi

	mov	esi, [.Row]
	dec	esi
	shl	esi, 4
	add	esi, [Dumper.ptrData]
	mov	edi, DumperAddress
	mov	ecx, $10

.copy:	lodsb
	cmp	al, ' '
	jae	@f
	mov	al, '.'
@@:	stosb
	loop	.copy

	pop	edi

	invoke	DrawText, [.hdc], DumperAddress, 16, [.ptrRect], DT_VCENTER or DT_CENTER or DT_SINGLELINE or DT_NOPREFIX
	jmp	.drawok

.fixed:
	cmp	[.Row], 0
	je	.offset

; address...
	mov	eax, [.Row]
	dec	eax
	shl	eax, 4	; *16 bytes on row
	add	eax, [edi+TDumper.pointer]
	stdcall NumToStr, eax, DumperAddress, 1, ntsHex or ntsUnsigned or ntsZeroTerminated or ntsFixedWidth or 8
	mov	[DumperAddress], '$'
	mov	eax, DumperAddress
	jmp	.drawit

.offset:
	mov	eax, [.Col]
	lea	eax, [DumperOffsets+4*eax]

.drawit:
	invoke	DrawText, [.hdc], eax, -1, [.ptrRect], DT_VCENTER or DT_CENTER or DT_SINGLELINE

.drawok:
	pop	ebx esi
	ret
endp



iglobal
  DumperOffsets  dd	 0, '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F', 'ASCI', 'I'
endg

uglobal
  DumperAddress rb $20
endg
