;
;  IDEparams.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;
; This module is responsible for creating/reading/saving of common for
; the whole IDE parameters from the main .ini file. It uses prefs.asm
; library.
;

iglobal
  ; Fresh user preferences
  pref_list FreshPreferences,\
  \       ; name                 type            default     name        group
            prefCC,              preftypeFlag,   1,              'CC', 'Editor Settings',\
            prefHelpFile,        preftypeString, 0,        'HelpFile', 'IDE Settings',\
            prefDebugger,        preftypeString, 0,        'Debugger', 'IDE Settings',\
            prefLine,            preftypeString, 0,            'LINE', 'IDE Settings',\
            prefViewPalette,     preftypeFlag,   0,    'show palette', 'IDE Settings',\
            prefPropWndVisible,  preftypeFlag,   0,         'visible', 'Properties Editor',\
            prefProjManVisible,  preftypeFlag,   1,         'visible', 'Project Manager',\
            prefFloatingWindows, preftypeFlag,   0,'floating windows', 'IDE Settings',\
            prefMainWindowX,     preftypeInt,    0,               'x', 'Main Window',\
            prefMainWindowY,     preftypeInt,    0,               'y', 'Main Window',\
            prefMainWindowW,     preftypeInt,  500,               'w', 'Main Window',\
            prefMainWindowH,     preftypeInt,  400,               'h', 'Main Window'
endg

proc CheckParams
local buf[1024]:BYTE
        ; Check ini file consistency...
        stdcall StrPtr, [hIniFileName]
        mov     esi, eax
        stdcall FileExists, esi
        jnz     .load_prefs
        stdcall FirstTimeInstall       ; try to create .ini file automatically
   .load_prefs:
        stdcall LoadPrefs, [hIniFileName], FreshPreferences

        stdcall GetPref, prefViewPalette,0
        stdcall ActionSetCheck, actViewPalette,eax

        stdcall GetPref, prefPropWndVisible,0
        stdcall ActionSetCheck, actViewPropEditor,eax

        stdcall GetPref, prefProjManVisible,0
        stdcall ActionSetCheck, actViewProjMan,eax

        ret
endp


;-------------------------------------
; It tries to create proper ini file
; in the case that ini file does not
; exists.
;-------------------------------------
proc FirstTimeInstall
locals
.str rb 1024
.ptr dd ?
endl
        mov     ebx, cPathsTableInc

.findloop:
        mov     ecx, [ebx]
        test    ecx, ecx
        jz      .notfound1
        lea     edi, [.str]
        lea     eax, [.ptr]
        invoke  GetFullPathName, ecx, 1024, edi, eax
        stdcall FileExists, edi
        lea     ebx, [ebx+4]
        test    eax, eax
        jz      .findloop
        mov     eax, [.ptr]
        mov     byte [eax-1], 0
        invoke  WritePrivateProfileString, _section_environment, cIncludeKey, edi, esi
.findmain:
        mov     ebx, cPathsTableMain
.findloop2:
        mov     ecx, [ebx]
        test    ecx, ecx
        jz      .notfound2

        lea     edi, [.str]
        lea     eax, [.ptr]
        invoke  GetFullPathName, ecx, 1024, edi, eax
        stdcall FileExists, edi
        lea     ebx, [ebx+4]
        test    eax, eax
        jz      .findloop2

        mov     eax, [.ptr]
        mov     byte [eax-1], 0
        invoke  WritePrivateProfileString, _section_environment, cFreshKey, edi, esi
        ret

.notfound1:
        invoke  MessageBox, [hApplication], cNotFoundMsg1, cNotFoundTtl, MB_ICONWARNING or MB_OK
        jmp     .findmain

.notfound2:
        invoke  MessageBox, [hApplication], cNotFoundMsg2, cNotFoundTtl, MB_ICONWARNING or MB_OK
        ret
endp


iglobal
  cPathsTableInc dd cAutoPath1, cAutoPath2, 0
  cAutoPath2  db '..\'
  cAutoPath1  db 'include'
  cFincTestFile db '\finc.txt',0

  cIncludeKey db 'finc',0

  cPathsTableMain dd cAutoPath3, cAutoPath4, 0
  cAutoPath3  db '..\'
  cAutoPath4  db 'Fresh.exe',0

  cFreshKey db 'Fresh',0

  cNotFoundMsg1 db "First time run procedure can't find include path automanically. Try to set it manualy in 'options|IDE params' menu."
                db "You have to use 'finc' name for this path.",0
  cNotFoundMsg2 db "First time run procedure can't find main program directory. Try to set it manually in 'options|IDE params' menu."
                db "You have to use 'Fresh' name for this path.",0
  cNotFoundTtl db 'Auto setup error',0
endg




proc AddCheckMessage .ptrMessage, .msgType
local .msg:TFreshMsg
        mov     eax, [.msgType]
        mov     ecx, [.ptrMessage]
        mov     [.msg.type], eax
        mov     [.msg.hMessage], ecx
        mov     [.msg.link.hFileName], 0
        lea     eax, [.msg]
        invoke  SendMessage, [frmMsg], MWM_ADDMESSAGE, eax, TVI_ROOT
        ret
endp

;------------------------------------------------
; Parces command line and opens the files.
; Still not parse the name with spaces.
;------------------------------------------------
proc ParseCommandLine
local .num:DWORD
        stdcall StrNew
        mov     ebx, eax
        invoke  GetCommandLine
        stdcall StrCopy, ebx, eax
        stdcall StrLen, ebx
        mov     ecx, eax
        stdcall StrPtr, ebx
        mov     edi, eax
        mov     [.num], 0
.foundspc:
        call    .skipspc
        call    .findquoted
        jc      .found
        call    .findnotquoted
        jnc     .finish
.found:
        inc     [.num]
        cmp     [.num], 1
        je      .foundspc
        push    ecx
        stdcall OpenFileByName, esi, NULL, TRUE
        pop     ecx
        jmp     .foundspc
.finish:
        stdcall StrDel, ebx
        ret
.findquoted:
        jecxz   .false
        mov     al, '"'
        cmp     [edi], al
        jne     .false
        inc     edi
        dec     ecx
        jecxz   .false
        mov     esi, edi
        repnz   scasb
        jnz     .false
        mov     byte [edi-1], 0
        stc
        retn
.false:
        clc
        retn
.findnotquoted:
        jecxz   .false
        mov     al, ' '
        mov     esi, edi
        repnz scasb
        jnz     .ok
        mov     byte [edi-1], 0
.ok:    stc
        retn
.skipspc:
        cmp     byte [edi], ' '
        jne     .endskip
        inc     edi
        loop    .skipspc
.endskip:
        retn
endp