;
;  Prefs.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;

iglobal
  ; preference types

  ; simple string type - for edit box
  preftypeString:
    dd prefLoadString
    dd prefSaveString
    dd prefGetString
    dd prefSetString
    dd prefFreeString
    dd prefStringToWnd
    dd prefWndToString

  ; integer - for edit box
  preftypeInt:
    dd prefLoadInt
    dd prefSaveInt
    dd prefGetInt
    dd prefSetInt
    dd prefFreeInt
    dd prefIntToWnd
    dd prefWndToInt

  ; TRUE/FALSE - almost like integer, for check box
  preftypeFlag:
    dd prefLoadInt
    dd prefSaveInt
    dd prefGetInt
    dd prefSetInt
    dd prefFreeInt
    dd prefFlagToWnd
    dd prefWndToFlag
endg


;--------------------------------------------------------------------------
; string preferences
;--------------------------------------------------------------------------
proc prefLoadString uses ebx edx, pszIniFile, pref
local buf[128]:BYTE
local def:DWORD
        mov     [def],0
        mov     ebx,[pref]
        or      ebx,ebx
        jz      .qfalse
        lea     eax,[buf]
        lea     edx,[def]
        invoke  GetPrivateProfileString, [ebx+TPref.pszGroup],[ebx+TPref.pszName],\
                                         edx,eax,128,[pszIniFile]
        lea     eax,[buf]
        stdcall StrDup, eax
        mov     [ebx+TPref.value],eax
        ret
   .qfalse:
        xor     eax,eax
        ret
endp


proc prefSaveString uses ebx, pszIniFile,pref
        mov     ebx,[pref]
        or      ebx,ebx
        jz      .qfalse

        stdcall StrPtr, [ebx+TPref.value]
        invoke  WritePrivateProfileString,[ebx+TPref.pszGroup],[ebx+TPref.pszName],\
                                         eax,[pszIniFile]
        ret
    .qfalse:
        xor     eax,eax
        ret
endp

proc prefGetString, pref, pData
        mov     eax,[pref]
        stdcall StrPtr, [eax+TPref.value]
        ret
endp

proc prefSetString, pref, pData
        mov     eax,[pref]
        stdcall StrCopy, [eax+TPref.value],[pData]
endp

proc prefFreeString, pref
        mov     eax,[pref]
        push    [eax+TPref.value]
        mov     [eax+TPref.value],0
        stdcall StrDel  ; arg from stack
        ret
endp

proc prefStringToWnd, pref, hwnd
        mov     eax,[pref]
        test    eax,eax
        jz      .finish
        stdcall StrPtr, [eax+TPref.value]
        test    eax,eax
        jz      .finish
        invoke  SendMessage, [hwnd],WM_SETTEXT,0,eax
.finish:
        ret
endp

proc prefWndToString uses ebx, pref,hwnd
local buf[1024]:BYTE
        lea     ebx,[buf]
        mov     dword[ebx],0
        invoke  SendMessage, [hwnd],WM_GETTEXT,1024,ebx
        mov     eax,[pref]
        or      eax,eax
        jz      .finish
        stdcall StrCopy, [eax+TPref.value],ebx
    .finish:
        ret
endp

;--------------------------------------------------------------------------
; Integer preferences
;--------------------------------------------------------------------------
proc prefLoadInt uses ebx, pszIniFile, pref
local buf[128]:BYTE
        mov     byte[buf],0
        mov     ebx,[pref]
        or      ebx,ebx
        jz      .qfalse
        lea     eax,[buf]
        invoke  GetPrivateProfileInt, [ebx+TPref.pszGroup],[ebx+TPref.pszName],\
                                         [ebx+TPref.value],[pszIniFile]
        mov     [ebx+TPref.value],eax
        ret
   .qfalse:
        xor     eax,eax
        ret
endp


proc prefSaveInt uses ebx, pszIniFile,pref
        mov     ebx,[pref]
        or      ebx,ebx
        jz      .qfalse
        stdcall NumToStr, [ebx+TPref.value],0,0,ntsDec
        push    eax
        stdcall StrPtr,eax
        invoke  WritePrivateProfileString,[ebx+TPref.pszGroup],[ebx+TPref.pszName],\
                                         eax,[pszIniFile]
        stdcall StrDel   ; arg from stack
        ret
    .qfalse:
        xor     eax,eax
        ret
endp

proc prefGetInt, pref, pData
        mov     eax,[pref]
        mov     eax,[eax+TPref.value]
        ret
endp

proc prefSetInt, pref, iValue
        mov     eax,[pref]
        push    [iValue]
        pop     [eax+TPref.value]
        ret
endp

proc prefFreeInt, pref
        ret
endp

proc prefIntToWnd, pref, hwnd
        mov     eax,[pref]
        test    eax,eax
        jz      .finish
        stdcall NumToStr, [eax+TPref.value],0,0,ntsDec
        push    eax
        stdcall StrPtr, eax
        invoke  SendMessage, [hwnd],WM_SETTEXT,0,eax
        stdcall StrDel   ; arg from stack
.finish:
        ret
endp

proc prefWndToInt uses ebx, pref,hwnd
local buf[1024]:BYTE
        lea     ebx,[buf]
        mov     dword[ebx],0
        invoke  SendMessage, [hwnd],WM_GETTEXT,1024,ebx
        mov     eax,[pref]
        or      eax,eax
        jz      .finish
        stdcall StrToNum, ebx
        mov     [eax+TPref.value],eax
    .finish:
        ret
endp

;--------------------------------------------------------------------------
; Flag preferences
;--------------------------------------------------------------------------

proc prefFlagToWnd, pref, hwnd
        mov     eax,[pref]
        test    eax,eax
        jz      .finish
        mov     eax,[TPref.value]
        test    eax,eax
        jz      @f
        stdcall SendMessage, [hwnd],BM_SETCHECK,BST_CHECKED,0
        jmp     .finish
    @@: stdcall SendMessage, [hwnd],BM_SETCHECK,BST_UNCHECKED,0
.finish:
        ret
endp

proc prefWndToFlag uses edx, pref,hwnd
        stdcall SendMessage, [hwnd],BM_GETCHECK,0,0
        mov     edx,1
        cmp     eax,BST_CHECKED
        je      @f
        xor     edx,edx
    @@: mov     eax,[pref]
        test    eax,eax
        jz      .finish
        mov     [eax+TPref.value],edx
.finish:
        ret
endp

;--------------------------------------------------------------------------


; loads all preferences (from the list) from INI file
proc LoadPrefs uses ebx, pszIniFile, pPrefList
        mov     ebx,[pPrefList]   ; pPrefList - array of TPref structures, ended by dword 0
        stdcall StrPtr, [pszIniFile]
        mov     [pszIniFile],eax
    .load_preference:
        mov     eax,[ebx]
        or      eax,eax
        jz      .finished
        stdcall [eax+TPrefType.pfnLoad], [pszIniFile],ebx
        add     ebx,sizeof.TPref
        jmp     .load_preference
    .finished:
        ret
endp


proc SavePrefs uses ebx, pszIniFile, pPrefList
        mov     ebx,[pPrefList]   ; pPrefList - array of TPref structures, ended by dword 0
        stdcall StrPtr, [pszIniFile]
        mov     [pszIniFile],eax
    .load_preference:
        mov     eax,[ebx]
        or      eax,eax
        jz      .finished
        stdcall [eax+TPrefType.pfnSave], [pszIniFile],ebx
        add     ebx,sizeof.TPref
        jmp     .load_preference
    .finished:
        ret
endp

; returns value of given preference
; pData - usually can be NULL, is required
;   when preference is not of a simple type,
;   but for example it is a RECT structure
proc GetPref uses ebx, pref,pData
        mov     ebx,[pref]
        mov     eax,[ebx+TPref.pPrefType]
        stdcall [eax+TPrefType.pfnGet],ebx,[pData]
        ret
endp

proc SetPref uses ebx, pref,pData
        mov     ebx,[pref]
        mov     eax,[ebx+TPref.pPrefType]
        stdcall [eax+TPrefType.pfnSet],ebx,[pData]
        ret
endp

proc PrefToWnd uses ebx, pref,hwnd
        mov     ebx,[pref]
        mov     eax,[ebx+TPref.pPrefType]
        stdcall [eax+TPrefType.pfnToWnd],ebx,[hwnd]
        ret
endp

proc WndToPref uses ebx, pref,hwnd
        mov     ebx,[pref]
        mov     eax,[ebx+TPref.pPrefType]
        stdcall [eax+TPrefType.pfnFromWnd],ebx,[hwnd]
        ret
endp


proc PrefToDlgItem uses ebx, pref,hwnd,id
        invoke  GetDlgItem, [hwnd],[id]
        mov     [hwnd],eax
        mov     ebx,[pref]
        mov     eax,[ebx+TPref.pPrefType]
        stdcall [eax+TPrefType.pfnToWnd],ebx,[hwnd]
        ret
endp

proc DlgItemToPref uses ebx, pref,hwnd,id
        invoke  GetDlgItem, [hwnd],[id]
        mov     [hwnd],eax
        mov     ebx,[pref]
        mov     eax,[ebx+TPref.pPrefType]
        stdcall [eax+TPrefType.pfnFromWnd],ebx,[hwnd]
        ret
endp


;--------------------------------------------------------------------------
;  misc. functions
;--------------------------------------------------------------------------
proc PrefToWndPos uses ebx, pref,hwnd
local x:DWORD,y:DWORD,w:DWORD,h:DWORD
        mov     ebx,[pref]
        stdcall GetPref, ebx,0
        mov     [x],eax
        add     ebx,sizeof.TPref
        stdcall GetPref, ebx,0
        mov     [y],eax
        add     ebx,sizeof.TPref
        stdcall GetPref, ebx,0
        mov     [w],eax
        add     ebx,sizeof.TPref
        stdcall GetPref, ebx,0
        mov     [h],eax
        cmp     [w],0
        je      .finish
        cmp     [h],0
        je      .finish
        invoke  MoveWindow, [hwnd],[x],[y],[w],[h],TRUE
   .finish:
        ret
endp

proc WndPosToPref uses ebx, pref,hwnd
local rect:RECT
        mov     ebx,[pref]
        lea     eax,[rect]
        invoke  GetWindowRect, [hwnd],eax
        stdcall SetPref, ebx, [rect.left]
        add     ebx,sizeof.TPref
        stdcall SetPref, ebx, [rect.top]
        mov     eax,[rect.right]
        sub     eax,[rect.left]
        add     ebx,sizeof.TPref
        stdcall SetPref, ebx, eax
        mov     eax,[rect.bottom]
        sub     eax,[rect.top]
        add     ebx,sizeof.TPref
        stdcall SetPref, ebx, eax
        ret
endp