;
;  ProjectFile.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;

iglobal
  ; error messages
  cCannotOpenProject db 'Cannot open project', 0
  cProjectFileNotFound db 'Project file not found.', 0
  cInvalidProjectFileFormat db 'File has invalid format.', 0
  msgPUC1 db 'The file: "',0
  msgPUC2 db '" is part of saved project, but it doesn''t exists on the disk.', 13, 10
          db 'Do you want to add it to the project anyway?',0
  msgPUCtitle db  'Error adding file to the project.',0
endg


; loads project from FPR file.
; returns: pointer to TProject structure on success,
;          or 0 on failure.
proc LoadProjectFromFile uses ebx ecx edx esi edi, hFile, hFileName
local project:DWORD, source:DWORD
        stdcall FileExists, [hFileName]
        or      eax,eax
        jz      .error_not_found
        stdcall ChangeDir, [hFileName]
        stdcall LoadBinaryFile, [hFileName]
        or      ecx,ecx
        jz      .error_invalid_format
        mov     esi,eax
        mov     [source],esi
        ; Check the file singature
        lodsd
        cmp     eax, 'FRPR'   ; Fresh project?
        jne     .error_invalid_format

        mov     eax,[hFile]
        or      eax,eax
        jz      .finish
        mov     [project],eax
.getoptions:
        lodsd
        cmp     eax, -1
        jne     .oldformat
        ; New format options
        ; Get memory size:
        lodsd
        mov     [poMemory], eax
        ; Get bin name check.
        xor     eax, eax
        lodsb
        mov     [poBinNameAuto], eax
        ; Get bin name...
        stdcall StrLen, esi
        mov     ebx, eax
        stdcall StrCopy, [poBinName], esi
        lea     esi, [esi+ebx+1]
        jmp     .getmainfile
.oldformat:
        mov     [poMemory], eax
        ; Create the main file of the project.
.getmainfile:
        stdcall StrLen, esi
        mov     ebx, eax
        test    ebx, ebx
        jz      .getcategories
        stdcall CreateFStruct, esi, ftMainSource
        mov     [eax+TOpenFile.projectID],TRUE
        mov     edi, eax
        stdcall GetModified, edi
        test    eax, eax
        jz      .addmain
        stdcall ProjectUnConsistent, edi
        cmp     eax, IDNO
        jne     .addmain
        stdcall FreeFStruct, edi
        mov     [fProjModified], TRUE
        jmp     .getcategories
.addmain:
        mov     eax,[project]
        mov     [eax+TProject.hMainFile],edi
        stdcall AddFileToCategory, edi,[eax+TProject.pMainCategory]
        mov     [fProjModified], FALSE
.getcategories:
        lea     esi, [esi+ebx+1]        ; goto to the first category
        ; Main category loop
.maincatloop:
        lodsd   ; image index
        cmp     eax, -1  ; Is this is the end of file?
        je      .endopenproj
        lea     esi, [esi+4]    ;skip the flags for now.
        mov     edx,[project]
        stdcall GetCategoryTree, esi, [edx+TProject.pMainCategory], eax
        mov     ebx, eax                ; Current category tree item.
.filesloop:
        stdcall StrLen, esi
        lea     esi, [esi+eax+1]        ; goto to next file
        cmp     byte [esi], 0
        je      .endoffiles
        stdcall DetectFileType, esi
        stdcall CreateFStruct, esi, eax
        mov     [eax+TOpenFile.projectID],TRUE
        mov     edi, eax
        stdcall GetModified, edi
        test    eax, eax
        jz      .addtoproject
        stdcall ProjectUnConsistent, edi
        cmp     eax, IDNO
        je      .adderror
.addtoproject:
        stdcall AddFileToCategory, edi,ebx
        mov     [fProjModified], FALSE
        test    eax, eax
        jnz     .filesloop
.adderror:
        mov     [fProjModified], TRUE
        stdcall FreeFStruct, edi   ; error add file...
        jmp     .filesloop
.endoffiles:
        inc     esi     ; next category.
        jmp     .maincatloop
.endopenproj:
        ; free data loaded from file
        invoke  GetProcessHeap
        invoke  HeapFree, eax,0,[source]
        mov     eax,[project]
.finish:
        ret
.error_not_found:
        invoke  MessageBox, [hMainWindow],cProjectFileNotFound, cCannotOpenProject,MB_OK+MB_ICONERROR
        jmp     .error
.error_invalid_format:
        invoke  MessageBox, [hMainWindow],cInvalidProjectFileFormat, cCannotOpenProject,MB_OK+MB_ICONERROR
;        jmp     .error
.error:
        xor     eax,eax
        ret
endp

proc ProjectUnConsistent .ptrOpenFile
        push    ebx
        mov     ebx, [.ptrOpenFile]
        stdcall StrDup, msgPUC1
        stdcall StrCat, eax, [ebx+TOpenFile.hFileName]
        stdcall StrCat, eax, msgPUC2
        push    eax
        stdcall StrPtr, eax
        invoke  MessageBox, [hApplication], eax, msgPUCtitle, MB_YESNO or MB_ICONWARNING
        cmp     eax,IDYES
        je      .ok
        mov     [fProjModified],1
    .ok:
        stdcall StrDel ; from the stack
        pop     ebx
        ret
endp

;-------------------------------------------------------------------------
; Returns the handle of treeview item that follow the tree of direcories
; of [hFileName]. If there is no some (or any) of the directories of the
; path, it will be created.
; The function checks for "/" and "\" as items separator.
; Arguments:
; hFileName - handle or pointer to the path.
; hRoot     - is root TVItem for category. For examples if category
;             have to be created in the root of the TreeView, hParent
;             should be TVI_ROOT
; Returns:
;   eax - Handle to treeview item that should be parent of the file.
;   edx - pointer to the name of the file without path.
;-------------------------------------------------------------------------
proc GetCategoryTree .hFileName, .hRoot, .iImage
        push    esi edi ebx
        stdcall StrPtr, [.hFileName]
        mov     esi, eax
        stdcall StrLen, [.hFileName]
        mov     ecx, eax
; find the filename:
.findloop:
        cmp     byte [esi+ecx], '/'
        je      .foundname
        cmp     byte [esi+ecx], '\'
        je      .foundname
        dec     ecx
        jns     .findloop
; Returns the [hRoot] if there is no directories names.
        mov     eax, [.hRoot]
        mov     edx, esi
        pop     ebx edi esi
        ret
.foundname:
        lea     edi, [esi+ecx+1]
        pushd   [edi]
        and     dword [edi], 0
        stdcall _DoCategoryTree, esi, [.hRoot], [.iImage]    ; track/create path items.
        popd    [edi]
        mov     edx, edi
        pop     ebx edi esi
        ret
endp

proc _DoCategoryTree .ptrName, .hTVRoot, .iImage
        push    esi edi ebx
        mov     esi, [.ptrName]
        mov     ebx, [.hTVRoot]
.outloop:
        cmp     byte [esi], 0   ; is it the end?
        je      .endloop
        mov     edi, esi
.loop:
        lodsb
        cmp     al,'\'
        je      .foundslash
        cmp     al,'/'
        jne     .loop
.foundslash:
        mov     byte [esi-1], 0
        stdcall _DoGetOneItem, edi, ebx, [.iImage]
        mov     ebx, eax
        mov     byte [esi-1], '/'
        jmp     .outloop
.endloop:
        mov     eax, ebx
        pop     ebx edi esi
        ret
endp

proc _DoGetOneItem .ptrItem, .hParent, .iImage
locals
.tvi TVITEM
.tvins TVINSERTSTRUCT
.str rb 256
endl
        push    esi edi ebx
        mov     eax,[.hParent]
        mov     eax,[eax+TProjectCategory.pSubCategory]
        test    eax, eax
        jz      .notfound
.itemscan:
        mov     ebx, eax
        stdcall StrCompNoCase, [ebx+TProjectCategory.hName], [.ptrItem]
        jc      .found
        mov     eax,[ebx+TProjectCategory.pNext]
        test    eax, eax
        jnz     .itemscan
.notfound:
        ; create new category
        stdcall CreateProjectCategory, [.ptrItem],[.iImage]
        mov     ecx, [.hParent]
        mov     edx,[ecx+TProjectCategory.pSubCategory]
        or      edx,edx
        jne     .first_child_ok
        mov     [ecx+TProjectCategory.pSubCategory],eax
        jmp     .added
    .first_child_ok:

    .find_last_child:
        cmp     [edx+TProjectCategory.pNext],0
        je      .got_last_one
        mov     edx,[edx+TProjectCategory.pNext]
        jmp     .find_last_child
    .got_last_one:
        mov     [eax+TProjectCategory.pPrev],edx
        mov     [edx+TProjectCategory.pNext],eax
    .added:
        pop     ebx edi esi
        ret
.found:
        mov     eax, ebx
        pop     ebx edi esi
        ret
endp



; converts project to data stream that can be written onto disk
; returns size in ecx
proc ProjectToStream uses ebx esi edi, hProject
local buf:DWORD
        invoke  GetProcessHeap
        invoke  HeapAlloc, eax, 0, 65536    ; Maximal size of the project file.
        test    eax, eax
        jz      .finish
        mov     [buf],eax
        mov     edi,eax
.finish:
        mov     eax,'FRPR'
        stosd
        ; New project format save
        xor     eax, eax
        dec     eax
        stosd
        ; save memory...
        mov     eax, [poMemory]
        stosd
        ; save binary name auto flag.
        mov     al, byte [poBinNameAuto]
        stosb
        ; save binary name
        stdcall GetRelativePath, [poBinName]
        push    eax
        call    pts_StrToStream
        stdcall StrDel ; from the stack
        xor     al, al
        stosb
        ; Main project file:
        mov     eax,[hProject]
        mov     eax,[eax+TProject.hMainFile]
        test    eax, eax
        jz      .mainok
        stdcall GetRelativePath, [eax+TOpenFile.hFileName]
        push    eax
        call    pts_StrToStream
        stdcall StrDel ; from the stack
        xor     al,al
   .mainok:
        stosb   ; end of main file filename.
        ; now go through all categories
        stdcall StrNew   ; temp. string for holding category name
        push    eax
        mov     ebx,eax
        mov     eax,[hProject]
        mov     eax,[eax+TProject.pMainCategory]
        mov     eax,[eax+TProjectCategory.pSubCategory]
        or      eax,eax
        jz      .no_category_tree
        stdcall pts_SaveCategory, eax,ebx
    .no_category_tree:
        stdcall StrDel ; arg from stack
        xor     eax, eax
        dec     eax
        stosd                   ; end of categories.
        mov     eax,[buf]
        mov     ecx,edi
        sub     ecx,[buf]
        ret
endp

;--------------------------------------
; edi - pointer to the stream buffer
; eax - handle/pointer to the string
;--------------------------------------
proc pts_StrToStream
        stdcall StrPtr, eax
        xor     ecx, ecx
.copyname:
        mov     cl, [eax]
        inc     eax
        jecxz   .endofname
        mov     [edi], cl
        inc     edi
        jmp     .copyname
.endofname:
        ret
endp

; edi has to point to destination memory
proc pts_SaveCategory uses ebx esi, hCategory,hParentString
        mov     esi,[hCategory]
        mov     eax,[esi+TProjectCategory.iIcon]  ; save icon
        stosd
        xor     eax,eax                           ; reserved dword
        stosd
        stdcall StrDup, [hParentString]           ; category name
        mov     ebx,eax
        stdcall StrCat,ebx,[esi+TProjectCategory.hName]
        stdcall StrCharCat,ebx,'\'
        mov     eax,ebx
        call    pts_StrToStream
        xor     al,al
        stosb
        ; save files
        mov    edx,[esi+TProjectCategory.pFileList]
  .save_file:
        or     edx,edx
        jz     .end_of_files
        stdcall GetRelativePath, [edx+TOpenFile.hFileName]
        push    eax
        call    pts_StrToStream
        xor     al,al
        stosb
        stdcall StrDel ; arg from stack
        mov     edx,[edx+TOpenFile.pNext]
        jmp     .save_file
  .end_of_files:
        xor     al,al
        stosb
        ; process subcategories
        cmp      [esi+TProjectCategory.pSubCategory],0
        je       .no_subcategory
        stdcall  pts_SaveCategory, [esi+TProjectCategory.pSubCategory],ebx
   .no_subcategory:
        ; process next category
        cmp      [esi+TProjectCategory.pNext],0
        je       .no_next_category
        stdcall  pts_SaveCategory, [esi+TProjectCategory.pNext],[hParentString]
   .no_next_category:
        stdcall StrDel, ebx
        ret
endp