;
;  ProjManager.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;

uglobal
  hProjManager    dd 0  ; handle of project manager
  fProjModified   dd 0  ; flag "modified" for current open project.
  poMemory        dd 0  ; The memory for the compiler
  poBinNameAuto   dd 0  ; The flag whether the binary name will be auto created of fixed
  poBinName       dd 0  ; Handle to the binary name.
  imlCategories   dd 0  ; Image list for tree-view control
  ptrMRUProj      dd 0  ; pointer to MRU list for projects.
  hTreeCatPopup   dd 0
  hTreeRootPopup  dd 0
  hProject        dd 0  ; currently opened project
  hDragImageList  dd ?  ; image list for drag and drop operation
  hDragItem       dd ?  ; dragged item
endg

iglobal
  cProjTitle      db      'Project manager', 0
  CoolMenu mnuProjCategory,           \
    mfNormal, actAddCategory,     \
    mfNormal, actAddFile,\
    mfSeparator, NONE,            \
    mfNormal, actMoveCategoryUp,  \
    mfNormal, actMoveCategoryDn

  CoolMenu mnuProjRoot,           \
    mfNormal, actAddCategory,     \
    mfSeparator, NONE,            \
    mfNormal, actProjectOptions
endg

;----------------------------------------
; Creates project manager common objects.
; Call this only once.
;----------------------------------------
proc InitProjectManager, owner
; Read project MRU list
        stdcall LoadMRUFromIni, 16, [hIniFileName], cMRUProjSection
        mov     [ptrMRUProj], eax
        stdcall ImageList_LoadGif, [hInstance], resBmpProjMan, 16, FALSE
        mov     [imlCategories], eax
        stdcall CreateCoolMenu, mnuProjCategory, TRUE
        mov     [hTreeCatPopup], eax
        stdcall CreateCoolMenu, mnuProjRoot, TRUE
        mov     [hTreeRootPopup], eax
        mov     [poBinNameAuto], TRUE
        mov     [poMemory], 16384
        stdcall StrNew
        mov     [poBinName], eax

        mov     ebx,WS_CAPTION or WS_CLIPCHILDREN or WS_SIZEBOX or WS_SYSMENU
        stdcall GetPref, prefFloatingWindows,0
        test    eax,eax
        jnz     @f

        mov     ebx,WS_CHILD+WS_BORDER
    @@:
        stdcall GetPref, prefProjManVisible,0
        test    eax,eax
        jz      .create
        or      ebx,WS_VISIBLE
   .create:
        invoke  CreateWindowEx, WS_EX_TOOLWINDOW, cFormClassName, cProjTitle,  \
                                ebx, 0, 0, 180, 400,                           \
                                [owner], 0, [hInstance], 0
        mov     [hProjManager], eax
        stdcall SubclassWindow, eax, ProjManProc

        ret
endp


proc FinalizeProjectManager
        cmp     [hProject], NULL
        je      .closeproj
        stdcall CloseFile, [hProject]
.closeproj:
        stdcall SaveMRUToIni, [ptrMRUProj], [hIniFileName], cMRUProjSection
        stdcall DestroyMRUList, [ptrMRUProj]
        invoke  ImageList_Destroy, [imlCategories]
        invoke  DestroyMenu, [hTreeCatPopup]
        invoke  DestroyWindow, [hProjManager]
        mov     [hProjManager],0
        ret
endp



proc CreateProjectManager .ptrOpenFile, .hTemplate
        push    ebx esi
        stdcall GetPref, prefProjManVisible,0
        test    eax,eax
        jnz     @f
        stdcall OnViewProjMan,actViewProjMan,0
    @@: xor     ebx, ebx
        mov     esi, [.ptrOpenFile]
        cmp     [hProject], ebx
        je      .onlyone
        stdcall CloseFile, [hProject]
        cmp     eax, cfrDestroyed
        jne     .fault
.onlyone:
        mov     [hProject], esi
; Initial project options... they could be changed after CEM_LOADFILE
        mov     [poMemory], 16384
        mov     [poBinNameAuto], TRUE
        cmp     [poBinName], 0
        je      @f
        stdcall StrDel, [poBinName]
@@:     stdcall StrNew
        mov     [poBinName], eax
        mov     eax,[hProjManager]
        mov     [esi+TOpenFile.hEditor], eax
        mov     [esi+TOpenFile.projectID], ebx
        invoke  SendMessage, [hProjManager], CEM_LOADFILE, [.hTemplate], ebx
        push    eax
        stdcall RearrangeWindows
        pop     eax
.finish:
        pop     esi ebx
        ret
.fault:
        mov     eax, errCanceled
        pop     esi ebx
        ret
endp

;-----------------------------------------------------
; Dialog proc for Project manager.
;-----------------------------------------------------
winproc ProjManProc
locals
.tvi    TVINSERTSTRUCT
.tvhit  TVHITTESTINFO
.item   TVITEM
.lvi    LVITEM
.col    LVCOLUMN
.count  dd ?
.str    rb 1024
.str2   rb 256
.ptr    dd ?
.pnt    POINT
endl

ondefault

        stdcall DockFormProc, [hwnd],[wmsg],[wparam],[lparam]
        ret

onmessage PMM_FILECHANGED
        cmp     [hProject],0
        je      .qfalse
        stdcall GetFileParent, [hProject], [wparam]
        test    eax,eax
        jz      .qfalse
        push    eax
        invoke  SendMessage, [hwnd], PMM_REMOVEFILE, [wparam], 0
        invoke  SendMessage, [hwnd], PMM_ADDFILE, [wparam] ; lparam from the stack
.endchanged:
        clc
        ret

onmessage PMM_ADDFILE
        cmp     [hProject],0
        je      .qfalse
        mov     [fProjModified], TRUE
        mov     eax, [wparam]
        mov     esi,eax
        cmp     [eax+TOpenFile.projectID], FALSE
        je      .projok
        invoke  SendMessage, [hwnd], PMM_REMOVEFILE, eax, 0
.projok:
        mov     eax,[lparam]   ; check parent category
        cmp     eax,-1
        jne     @f
        invoke  SendMessage, [hwnd],PMM_GETCURCATEGORY,0,0
        jmp     .category_ok
    @@: or      eax,eax
        jnz     .category_ok
        ; Add as main file.
        cmp     [esi+TOpenFile.ptrType], ftMainSource
        je      .typeok
        cmp     [esi+TOpenFile.ptrType], ftCommonSource
        jne     .qfalse
        mov     [esi+TOpenFile.ptrType], ftMainSource   ; change file type.
   .typeok:
        mov     eax,[hProject]
        mov     [eax+TProject.hMainFile],esi
        mov     eax,[eax+TProject.pMainCategory]
.category_ok:
        mov     [esi+TOpenFile.projectID], TRUE
        mov     ecx,[hProject]
        mov     ecx,[ecx+TProject.pMainCategory]
        cmp     eax,ecx
        jne     .not_main_cat     ; can't add files other than main file to main category
        cmp     [lparam],0        ; set main file?
        jne     .finish
   .not_main_cat:
        stdcall AddFileToCategory, esi,eax
        invoke  GetDlgItem, [hwnd],ProjTreeCtrlID
        stdcall ProjectToTreeView, [hProject],eax
        stdcall FileChanged, esi, fcEditor
.finish:
        clc
        ret

onmessage PMM_REMOVEFILE
        mov     [fProjModified], TRUE
        mov     esi, [wparam]
        mov     [esi+TOpenFile.projectID], FALSE
        stdcall FileChanged, esi, fcEditor
        stdcall RemoveFileFromProject, [hProject],esi
        invoke  GetDlgItem, [hwnd],ProjTreeCtrlID
        stdcall ProjectToTreeView, [hProject],eax
        jmp     .qfalse

onmessage PMM_GETCURCATEGORY
        invoke  SendDlgItemMessage, [hwnd],ProjTreeCtrlID,TVM_GETNEXTITEM,TVGN_CARET,0
        or      eax,eax
        jnz     .get_parent_category
    .return_main_category:
        mov     eax,[hProject]
        mov     eax,[eax+TProject.pMainCategory]
        jmp     .current_cat_ok
    .get_parent_category:
        stdcall GetTVItemParam, [hwnd],ProjTreeCtrlID,eax
        or      eax,eax
        jz      .return_main_category
        test    eax,1
        jnz     .fix_pointer
        ; try again
        invoke  SendDlgItemMessage, [hwnd],ProjTreeCtrlID,TVM_GETNEXTITEM,TVGN_PARENT,[.item.hItem]
        or      eax,eax
        jz      .return_main_category
        jmp     .get_parent_category
    .fix_pointer:
        btr     eax,0
    .current_cat_ok:
        clc
        ret

onmessage CEM_LOADFILE
        mov     ebx, [hProject]
        cmp     [wparam], 0
        jne     .loadit
        cmp     [ebx+TOpenFile.fNeverSaved], FALSE
        je      .loadit
        mov     [fProjModified], FALSE
        xor     eax, eax
        clc
        ret
.loadit:
        mov     eax, [wparam]
        test    eax, eax
        jnz     @f
        mov     eax, [ebx+TOpenFile.hFileName]
   @@:  stdcall LoadProjectFromFile, ebx,eax
        or      eax,eax
        jz      .finish
        mov     [hProject],eax
        invoke  GetDlgItem, [hwnd],ProjTreeCtrlID
        stdcall ProjectToTreeView, [hProject],eax
        stdcall SetMainTitle
        mov     ebx, [hProject]
        stdcall AddToMRUList, [ptrMRUProj], [ebx+TOpenFile.hFileName]
        invoke  SendDlgItemMessage, [hwnd], ProjMainCtrlID, LVM_SETCOLUMNWIDTH, 0, LVSCW_AUTOSIZE
        xor     eax, eax
        clc
        ret

onmessage _CEM_GETTEXT
        mov     eax, [hProject]
        cmp     [lparam],0
        je      .errorgettext
        stdcall ChangeDir, [eax+TOpenFile.hFileName]
        stdcall ProjectToStream, [hProject]
        or      eax,eax
        jz      .finish
        mov     esi, [lparam]
        mov     [esi+TInMemoryFile.ptrBuffer], eax
        mov     [esi+TInMemoryFile.fileSize], ecx
        mov     [esi+TInMemoryFile.fAscii], 0
        mov     eax,TRUE  ; return with success
        clc
        ret
.errorgettext:
        xor     eax, eax
        clc
        ret

onmessage  _CEM_CHANGENOTIFY
; change directory to the new name...
        mov     ebx, [hProject]
        stdcall ChangeDir, [ebx+TOpenFile.hFileName]
        stdcall SetMainTitle
        stdcall AddToMRUList, [ptrMRUProj], [ebx+TOpenFile.hFileName]
        clc
        ret

onmessage _CEM_FOCUSFILE
        invoke  SetFocus, [hwnd]
        clc
        ret

onmessage _CEM_CANACTION
        xor     eax, eax
        clc
        ret

onmessage _CEM_EDITACTION
        clc
        ret

onmessage _CEM_GETFILE
        mov     eax, [hProject]
        clc
        ret

onmessage _CEM_GETMODIFIED
        mov     eax, [fProjModified]
        clc
        ret

onmessage _CEM_SETMODIFIED
        push    [wparam]
        pop     [fProjModified]
        clc
        ret

onmessage WM_ACTIVATE
        stdcall FocusFile, [hProject]
        stc
        ret

onmessage WM_DROPFILES
        lea     esi, [.str]
        invoke  DragQueryFile, [wparam], -1, esi, 1024
        mov     ebx, eax
        xor     edi, edi
.opendroploop:
        invoke  DragQueryFile, [wparam], edi, esi, 1024
        stdcall OpenFileByName, esi, NULL, FALSE
        test    eax, eax
        jz      .next
        invoke  SendMessage, [hwnd], PMM_ADDFILE, eax, -1
.next:
        inc     edi
        dec     ebx
        jnz     .opendroploop
        invoke  DragFinish, [wparam]
        xor     eax, eax
        clc
        ret

onmessage WM_COMMAND
        cmp     [wparam], IDOK
        jne     .tomain
        jmp     .ondefault
.tomain:
        invoke  PostMessage, [hMainWindow], [wmsg], [wparam], [lparam]
        jmp     .ondefault

onmessage PMM_GETMAINFILE
        mov     eax,[hProject]
        or      eax,eax
        jz      .qfalse
        mov     eax,[eax+TProject.hMainFile]
        clc
        ret

onmessage WM_NOTIFY
        mov     eax, [wparam]
        mov     esi, [lparam]
        cmp     eax, ProjTreeCtrlID
        je      .treenotify
        clc
        ret

;--- Category tree notification messages ---------
.treenotify:
        cmp     [esi+NMHDR.code], TVN_KEYDOWN
        je      .keydown
        cmp     [esi+NMHDR.code], NM_RCLICK
        je      .tvrightclick
        cmp     [esi+NMHDR.code], NM_DBLCLK
        je      .tvndblclick
        cmp     [esi+NMHDR.code], TVN_ENDLABELEDIT
        je      .endlabeledit
        cmp     [esi+NMHDR.code], TVN_BEGINLABELEDIT
        je      .beginlabeledit
        cmp     [esi+NMHDR.code], TVN_ITEMEXPANDED
        je      .itemexpanded
        cmp     [esi+NMHDR.code], TVN_SELCHANGED
        je      .selchanged
        cmp     [esi+NMHDR.code], TVN_BEGINDRAG
        jne     .qfalse

.begindrag:
        invoke  GetDlgItem, [hwnd], ProjTreeCtrlID
        mov     ebx,eax
        mov     eax,[esi+NMTREEVIEW.itemNew.hItem]
        mov     [hDragItem],eax
        invoke  SendMessage, ebx,TVM_CREATEDRAGIMAGE,0,[esi+NMTREEVIEW.itemNew.hItem]
        mov     [hDragImageList],eax
        invoke  ImageList_BeginDrag, eax,0,0,0
        invoke  ImageList_DragEnter, ebx,[esi+NMTREEVIEW.ptDrag.x],[esi+NMTREEVIEW.ptDrag.y]
        invoke  SetCapture, [hwnd]
        jmp     .qfalse

.selchanged:
        invoke  SendDlgItemMessage, [hwnd],ProjTreeCtrlID,TVM_SELECTITEM,TVGN_CARET,[esi+NMTREEVIEW.itemOld]
        jmp     .qfalse

; category was expanded
.itemexpanded:
        mov     eax,[esi+NMTREEVIEW.itemNew.lParam]
        btr     eax,0
        jnc     .qfalse   ; file?
        btc     [eax+TProjectCategory.iFlags],PCF_EXPANDED shr 1   ; toggle expaned flag
        jmp     .qfalse

; begin of label editing
.beginlabeledit:
        mov     eax,[lparam]
        test    [eax+TVDISPINFO.item.lParam],1
        jnz     .qfalse
        xor     eax,eax
        dec     eax
        clc
        ret

; end of label editing
.endlabeledit:
        mov     esi, [lparam]
        cmp     [esi+TVDISPINFO.item.pszText], 0
        je      .qfalse
        stdcall StrLen, [esi+TVDISPINFO.item.pszText]
        or      eax,eax
        jz      .qfalse    ; category name can't be empty string
        mov     eax,[esi+TVDISPINFO.item.lParam]
        test    eax,1
        jz      .qfalse    ; can't change file name
        btr     eax,0
        stdcall StrCopy, [eax+TProjectCategory.hName],[esi+TVDISPINFO.item.pszText]
        mov     [esi+TVDISPINFO.item.mask], TVIF_TEXT
        lea     esi, [esi+TVDISPINFO.item]
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID, TVM_SETITEM, 0, esi
        mov     [fProjModified], TRUE
.endedit:
       clc
       ret

;---------- double click on treeview control --------------
; if user clicked on file item, then open it in editor.
.tvndblclick:
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID,TVM_GETNEXTITEM,TVGN_CARET,0
        stdcall GetTVItemParam, [hwnd],ProjTreeCtrlID,eax
        or      eax,eax
        jz      .qfalse
        test    eax,1   ; file?
        jnz     .qfalse ; category, ignore
        stdcall LoadFile, eax, NULL
        jmp     .qfalse

;---------- right click on treeview control --------------
.tvrightclick:
        lea     ebx,[.tvhit.pt]
        invoke  GetCursorPos, ebx
        invoke  ScreenToClient,[hwnd],ebx
        lea     eax,[.tvhit]
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID,TVM_HITTEST,0,eax
        or      eax,eax
        jz      .qfalse
   .ready_for_menu:
        push    eax
        mov     [.item.hItem],eax
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID,TVM_SELECTITEM, TVGN_CARET, eax
        stdcall GetTVItemParam, [hwnd],ProjTreeCtrlID ; last arg from stack
        or      eax,eax
        jnz     @f
        lea     eax, [.pnt]
        invoke  GetCursorPos, eax
        invoke  TrackPopupMenu, [hTreeRootPopup], TPM_LEFTALIGN or TPM_RIGHTBUTTON, [.pnt.x], [.pnt.y], 0, [hwnd], NULL
        jmp     .qfalse
    @@: test    eax,1   ; file?
        jz      .qfalse ; yes, ignore
        lea     eax, [.pnt]
        invoke  GetCursorPos, eax
        invoke  TrackPopupMenu, [hTreeCatPopup], TPM_LEFTALIGN or TPM_RIGHTBUTTON, [.pnt.x], [.pnt.y], 0, [hwnd], NULL
        jmp     .qfalse

;------------- key down notification -----------------
.keydown:
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CARET, 0

        or      eax,eax
        jz      .qfalse
        mov     [.item.hItem],eax
        mov     [.item.mask],TVIF_PARAM+TVIF_HANDLE
        lea     eax,[.item]
        invoke  SendDlgItemMessage, [hwnd],ProjTreeCtrlID,TVM_GETITEM,0,eax
        or      eax,eax
        jz      .qfalse
        cmp     [esi+TVKEYDOWN.wVKey], VK_APPS
        je      .apps_key
        cmp     [esi+TVKEYDOWN.wVKey], VK_RETURN
        je      .tv_enter_key
        cmp     [esi+TVKEYDOWN.wVKey], VK_DELETE
        je      .tv_del_key
        jmp     .qfalse
  .apps_key:
        invoke  SendDlgItemMessage, [hwnd],ProjTreeCtrlID,TVM_GETNEXTITEM,TVGN_CARET,0
        or      eax,eax
        jz      .qfalse
        jmp     .ready_for_menu
  .tv_del_key:
        mov     esi,[.item.lParam]
        or      esi,esi    ; if lParam is 0, then user pressed DEL on
        jz      .qfalse    ; main project category, which cannot be removed
        test    esi,1
        jz      .remove_file
        ; remove category
        invoke  MessageBox, [hApplication], cDelCatQuery, cDelCatTitle, MB_ICONQUESTION or MB_YESNO
        cmp     eax, IDYES
        jne     .qfalse
        btr     esi,0
        invoke  SendDlgItemMessage, [hwnd],ProjTreeCtrlID,TVM_DELETEITEM,0,[.item.hItem]
        stdcall RemoveCategoryFromProject, [hProject],esi
        stdcall DestroyProjectCategory, [hProject],esi
        jmp     .qfalse
   .remove_file:
        ; remove file
        invoke  MessageBox, [hApplication], cDelFileQuery, cDelFileTitle, MB_ICONQUESTION or MB_YESNO
        cmp     eax, IDYES
        jne     .qfalse
        invoke  SendMessage, [hwnd], PMM_REMOVEFILE, esi, 0
        jmp     .qfalse

   .tv_enter_key:
        mov     eax,[.item.lParam]
        test    eax,1
        jnz     .qfalse
        stdcall LoadFile, eax, NULL
        jmp     .qfalse

onmessage WM_LBUTTONUP
        mov     esi,[hDragImageList]
        or      esi,esi
        jz      .finish
        invoke  GetDlgItem, [hwnd],ProjTreeCtrlID
        mov     ebx,eax
        invoke  ImageList_DragLeave, ebx
        invoke  ImageList_EndDrag
        invoke  ImageList_Destroy,esi
        invoke  SendMessage, ebx,TVM_GETNEXTITEM,TVGN_DROPHILITE,0
        mov     edi,eax
        invoke  SendMessage, ebx,TVM_SELECTITEM,TVGN_CARET,eax
        invoke  SendMessage, ebx,TVM_SELECTITEM,TVGN_DROPHILITE,0
        invoke  ReleaseCapture
        mov     [hDragImageList],0
        ; now move category
        mov     [.item.hItem],edi
        mov     [.item.mask],TVIF_HANDLE+TVIF_PARAM
        lea     eax,[.item]
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID,TVM_GETITEM,0,eax
        or      eax,eax
        jz      .qfalse
        mov     edi,[.item.lParam]
        test    edi,edi
        jz      .drop_at_main
        test    edi,1   ; file?
        jz      .qfalse ; yes, ignore
        btr     edi,0
    .drop_at_main:
        mov     eax,[hDragItem]
        mov     [.item.hItem],eax
        lea     eax,[.item]
        invoke  SendDlgItemMessage, [hwnd], ProjTreeCtrlID,TVM_GETITEM,0,eax
        or      eax,eax
        jz      .qfalse
        mov     esi,[.item.lParam]
        test    esi,esi
        jz      .qfalse

        test    esi,1
        jz      .drop_file
        btr     esi,0  ; clear first bit

        cmp     esi,edi   ; can't drop item on itself
        je      .qfalse

        stdcall RemoveCategoryFromProject, [hProject],esi
        stdcall InsertProjectCategory, [hProject],edi,esi
        stdcall ProjectToTreeView, [hProject],ebx
        mov     [fProjModified], TRUE
        jmp     .qfalse
  .drop_file:
        or      edi,edi    ; can't add file to main category right now
        jz      .qfalse
        stdcall RemoveFileFromProject, [hProject],esi
        stdcall AddFileToCategory, esi,edi
        stdcall ProjectToTreeView, [hProject],ebx
        mov     [fProjModified], TRUE
        jmp     .qfalse

onmessage WM_MOUSEMOVE
        cmp     [hDragImageList],0
        je      .finish
        invoke  GetDlgItem, [hwnd],ProjTreeCtrlID
        mov     ebx,eax
        movzx   eax,word[lparam]
        mov     [.tvhit.pt.x],eax
        movzx   ecx,word[lparam+2]
        mov     [.tvhit.pt.y],ecx
        invoke  ImageList_DragMove, eax,ecx
        invoke  ImageList_DragShowNolock, 0
        lea     eax,[.tvhit]
        invoke  SendMessage, ebx,TVM_HITTEST,0,eax
        or      eax,eax
        jz      @f
        invoke  SendMessage, ebx,TVM_SELECTITEM,TVGN_DROPHILITE,eax
    @@: invoke  ImageList_DragShowNolock, 1
        jmp     .qfalse

iglobal
  cDelCatQuery db 'Do you really want to remove this category and all category files from the project?', 0
  cDelCatTitle db 'Remove category from project.', 0
  cDelFileQuery db 'Do you really want to remove selected file from the project?',0
  cDelFileTitle db 'Remove file from project.',0
endg


onmessage WM_CLOSE
        stdcall OnViewProjMan, actViewProjMan,0
        invoke  ShowWindow, [hProjManager],SW_HIDE
        jmp     .qfalse

onmessage WM_DESTROY
        mov     [hProjManager], 0
        stdcall StrDel, [poBinName]
        mov     [poBinNameAuto], TRUE
        jmp     .qfalse

iglobal
  cMRUProjSection db 'MRU projects',0
endg

onmessage AM_INITWINDOW
        xor     ebx, ebx
        invoke  CreateWindowEx, WS_EX_CLIENTEDGE,                                                       \
                                cTreeViewClassName, ebx,                                                \
                                WS_CHILD or WS_VISIBLE or TVS_HASBUTTONS or                             \
                                TVS_SHOWSELALWAYS or TVS_HASLINES or TVS_EDITLABELS or WS_TABSTOP,      \
                                ebx, ebx, ebx, 100,                                                     \
                                [hwnd], ProjTreeCtrlID, [hInstance], ebx
        mov     edi, eax
        stdcall SetAlign, eax, waClient
        invoke  SendMessage, edi, TVM_SETIMAGELIST, TVSIL_NORMAL, [imlCategories]
        invoke  DragAcceptFiles, [hwnd], TRUE

;        stdcall ProjectToTreeView, [hProject],edi

        mov     [hDragItem],0  ; no item is being dragged
        mov     [hDragImageList],0

.qfalse:
        xor     eax,eax
        clc
        ret
endwp

;***********************************************************
; Set the filename of the project and updates the title of
; the main window.
;***********************************************************
proc SetMainTitle
        push    esi
        xor     eax, eax
        mov     esi, [hProject]
        test    esi, esi
        jz      @f
        stdcall StrPtr, [esi+TProject.hFileName]
@@:     invoke  SendMessage, [hMainWindow], WM_SETTEXT, 0, eax
        pop     esi
        ret
endp

;***********************************************
; Handler of action actMoveCategory
;***********************************************
proc OnMoveCategory uses ebx ecx edx esi, wparam, lparam
local item:TVITEM
local buf[128]:BYTE
        xor     ebx, ebx
        cmp     [wparam], actMoveCategoryUp
        jne     @f
        inc     ebx
@@:     invoke  GetDlgItem, [hProjManager], ProjTreeCtrlID
        mov     esi, eax
        invoke  SendMessage, esi, TVM_GETNEXTITEM, TVGN_CARET, 0
        stdcall GetTVItemParam, [hProjManager],ProjTreeCtrlID,eax
        test    eax,1
        jz      .finish   ; file?
        btr     eax,0
        stdcall MoveCategory, [hProject],eax,ebx
        or      eax,eax
        jz      .finish
        stdcall ProjectToTreeView, [hProject],esi
        mov     [fProjModified], TRUE
        jmp     .finish
.finish:
        ret
endp