;
;  SourceEditor.asm
;  Fresh IDE
;  Copyright (c) 2003-2006, Fresh Developement Team
;

uglobal
  hEditorMenu     dd 0                        ; source popup menu item.
  hBookmarkIcon   dd 0
  hCCList         dd 0
  CurrentSettings dd ?
  lpszReadLine    dd ?
endg


iglobal

defaultSettings:
Theme                                                                   \
  'Fixedsys', -12, DEFAULT_CHARSET,                                     \
  AES_AUTOINDENT or AES_SMARTTABS or AES_CONSOLECARET,                  \
  $FFFFFF, $800000, $000000, $C0C0C0, $C0C0C0, $C0C0C0, 30,             \
  $000000, $ffff00,                                                     \
  $D0D0D0, $FFFF00, $FFFF00, $A0A0A0, $00FF00, $00FFFF


CoolMenu EditorPopup,                   \
    mfNormal, actGotoDefinition,        \
    mfNormal, actOpenFileAtCursor,      \
    mfNormal, actInclude,               \
    mfSeparator, NONE,                  \
    mfNormal, actCut,                   \
    mfNormal, actCopy,                  \
    mfNormal, actPaste,                 \
    mfSeparator, NONE,                  \
    mfNormal,    actIndent,             \
    mfNormal,    actOutdent,            \
    mfSeparator, NONE,                  \
    mfNormal,    actComment,            \
    mfNormal,    actUncomment,          \
    mfSeparator, NONE,                  \
    mfSubmenu, '&Bookmarks',            \
      mfNormal,    actToggleBM,         \
      mfSeparator, NONE,                \
      mfNormal,    actNextBM,           \
      mfNormal,    actPrevBM,           \
      mfSeparator, NONE,                \
      mfNormal,    actClearBM,          \
    mfEnd, END
endg


CCM_UPDATECCPOS = WM_USER+$f000


iglobal
  cCurrentEditorSettings db 'Current.theme',0
  aeicons AEICONS
endg


;----------------------------------------
; Creates sourceeditor window.
; Call this only once.
;----------------------------------------
proc InitSourceEditor
        stdcall CreateCoolMenu, EditorPopup, TRUE
        mov     [hEditorMenu], eax

        ; prepare line buffer
        invoke  VirtualAlloc, 0,0x10000,MEM_COMMIT,PAGE_READWRITE
        mov     [lpszReadLine],eax

        ; new icons
        invoke  LoadImage, [hInstance], resBookmarkIcon, IMAGE_ICON, 16, 16, LR_SHARED
        mov     [aeicons.hBookmarkIcon],eax
        invoke  LoadImage, [hInstance], resBreakpointIcon, IMAGE_ICON, 16, 16, LR_SHARED
        mov     [aeicons.hBreakpointIcon],eax
        invoke  LoadImage, [hInstance], resActiveIcon, IMAGE_ICON, 16, 16, LR_SHARED
        mov     [aeicons.hActiveIcon],eax

; Load editor color profile...
        stdcall LoadEditorSettings, cCurrentEditorSettings
        test    eax, eax
        jnz     .profileok

        stdcall ThemeDuplicate, defaultSettings

.profileok:
        mov     [CurrentSettings], eax
        ret
endp


proc FinalizeSourceEditor
        stdcall SaveEditorSettings, cCurrentEditorSettings, [CurrentSettings], TRUE
        stdcall DestroySettings, [CurrentSettings]

        invoke  VirtualFree, [lpszReadLine],0,MEM_RELEASE

        invoke DestroyMenu, [hEditorMenu]
        invoke DestroyIcon, [hBookmarkIcon]
        ret
endp


;--------------------------------------------------------
; Creates the editor window and returns handle to it.
;--------------------------------------------------------
proc CreateSourceEditor .ptrOpenFile, .hTemplate
        push    ebx esi edi


        mov     esi, [.ptrOpenFile]
        mov     edi, [CurrentSettings]

        xor     eax, eax
        mov     ecx, WS_CHILD or WS_HSCROLL or WS_VSCROLL or ES_NOHIDESEL or ES_WANTRETURN
        or      ecx, [edi+TEditorSettings.AsmEditStyle]
        invoke  CreateWindowEx, WS_EX_CLIENTEDGE or WS_EX_ACCEPTFILES, cAsmEditClassName,    \
                eax, ecx, eax, eax, eax, eax, [hEditorsHost], eax,      \
                [hInstance], eax

        test    eax, eax
        jz      .errorcreate

        mov     ebx, eax
        mov     [esi+TOpenFile.hEditor], ebx

        invoke  SetWindowLong, ebx, ofsOpenFileStruct, esi

; adjust the look and feel for the editor window.
        invoke  CreateFont, [edi+TEditorSettings.FontSize], 0, 0, 0, FW_NORMAL,                 \
                            0, 0, 0, [edi+TEditorSettings.FontCharset],                         \
                            OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,           \
                            FIXED_PITCH or FF_DONTCARE, [edi+TEditorSettings.ptrFontName]

        invoke  SendMessage, ebx, WM_SETFONT, eax, FALSE
        invoke  SendMessage, ebx, AEM_SETTHEME, 0, [edi+TEditorSettings.ptrTheme]
        mov     eax, [esi+TOpenFile.ptrType]
        invoke  SendMessage, ebx, AEM_SETSYNTAXHIGHLIGHT, 0, [eax+TFileType.lParam]
        invoke  SendMessage, ebx, AEM_SETRIGHTCLICKMENU, [hEditorMenu],[hEditorsHost]
        invoke  SendMessage, ebx, AEM_SETICONS, 0,aeicons

        stdcall SubclassWindow, ebx, SourceEditProc
        invoke  SendMessage, ebx, CEM_LOADFILE, [.hTemplate], 0
        test    eax, eax
        jnz     .errorload

        invoke  SendMessage, [hEditorsHost], EHM_DOCKWINDOW, ebx, dfClient


        xor     eax, eax

        pop     edi esi ebx
        ret

.errorcreate:
        invoke  GetLastError
        pop     edi esi ebx
        ret

.errorload:
        push    eax
        invoke  DestroyWindow, ebx
        pop     eax
        pop     edi esi ebx
        ret
endp





winproc SourceEditProc
locals
.aepos AEPOS
.textlen dd ?
.written dd ?
.memoryfile TInMemoryFile
endl

        invoke  GetWindowLong, [hwnd], ofsOpenFileStruct
        mov     edi, eax

ondefault
        stc
        ret

;------- Custom message handlers ------------------

onmessage _CEM_GETTEXT
        mov     ebx, [lparam]   ; pointer to TInMemoryFile
        invoke  SendMessage, [hwnd], WM_GETTEXTLENGTH, 0, 0
        lea     esi, [eax+1]
        invoke  HeapAlloc, [hHeap], 0, esi
        test    eax, eax
        jz      .errorgettext
        mov     [ebx+TInMemoryFile.ptrBuffer], eax
        mov     [ebx+TInMemoryFile.fAscii], TRUE
        invoke  SendMessage, [hwnd], WM_GETTEXT, esi, [ebx+TInMemoryFile.ptrBuffer]
        mov     [ebx+TInMemoryFile.fileSize], eax
        mov     eax,TRUE
        clc
        ret

.errorgettext:
        clc
        ret

;----------------------------------------------------
onmessage CEM_LOADFILE

        mov     eax, [wparam]
        test    eax, eax
        jnz     .loadit

        cmp     [edi+TOpenFile.fNeverSaved], 0
        jne     .skiploading

; Load the file in ASMEDIT window.
        mov     eax, [edi+TOpenFile.hFileName]

.loadit:
        stdcall LoadBinaryFile, eax
        test    eax, eax
        jz      .errorload

        push    eax
        invoke  SendMessage, [hwnd], WM_SETTEXT, 0, eax
        invoke  HeapFree, [hHeap], 0        ; address from the stack.

        stdcall AddToMRUList, [ptrMRUFiles], [edi+TOpenFile.hFileName]

.skiploading:
        invoke  GetParent, [hwnd]
        invoke  SendMessage, eax, EHM_FILECHANGED, edi, 0   ; Set the parameters of the tab.

        xor     eax, eax
        clc
        ret

.errorload:
        invoke  GetLastError
        clc
        ret

;---------------------------------------------------------------------
onmessage _CEM_CHANGENOTIFY

        invoke  GetParent, [hwnd]
        invoke  SendMessage, eax, EHM_FILECHANGED, edi, 0
        clc
        ret

;---------------------------------------------------------------------
onmessage _CEM_FOCUSFILE

        invoke  GetParent, [hwnd]
        invoke  SendMessage, eax, EHM_FOCUSWINDOW, [hwnd], 0
        clc
        ret

;---------------------------------------------------------------------
onmessage _CEM_CANACTION

        mov     eax, [lparam]
        cmp     eax, WM_CUT
        je      .checkcopycut
        cmp     eax, WM_COPY
        je      .checkcopycut
        cmp     eax, WM_CLEAR
        je      .checkcopycut
        cmp     eax, WM_UNDO
        je      .checkundo
        cmp     eax, EM_UNDO
        je      .checkundo
        cmp     eax, WM_PASTE
        je      .checkpaste
        xor     eax, eax
        clc
        ret

.checkcopycut:
        lea     esi, [.aepos]
        invoke  SendMessage, [hwnd], AEM_GETPOS, esi, 0
        mov     eax, [esi+AEPOS.caretPosition]
        mov     ecx, [esi+AEPOS.caretLine]
        cmp     [esi+AEPOS.selectionPosition], eax
        jne     .can_copy
        cmp     [esi+AEPOS.selectionLine], ecx
        jne     .can_copy
        xor     eax, eax
        clc
        ret

.can_copy:
        xor     eax, eax
        inc     eax
        clc
        ret

.checkundo:
        invoke  SendMessage, [hwnd], EM_CANUNDO, 0, 0
        clc
        ret

.checkpaste:
        invoke  IsClipboardFormatAvailable, CF_TEXT
        clc
        ret

;---------------------------------------------------------------------
onmessage _CEM_EDITACTION

        invoke  SendMessage, [hwnd], [lparam], 0, 0
        clc
        ret

;---------------------------------------------------------------------
onmessage _CEM_GETMODIFIED

        invoke  SendMessage, [hwnd], AEM_GETMODIFIED, 0, 0
        clc
        ret

;---------------------------------------------------------------------
onmessage _CEM_SETMODIFIED

        invoke  SendMessage, [hwnd], AEM_SETMODIFIED, [wparam], [lparam]
        clc
        ret

;---------------------------------------------------------------------
onmessage _CEM_GETFILE

        mov     eax, edi
        clc
        ret

;---------------------------------------------------------------------
onmessage _CEM_GETASMEDIT

        mov     eax, [hwnd]
        clc
        ret
;---------------------------------------------------------------------
onmessage WM_KILLFOCUS

        cmp     [hCCList], 0
        je      .ondefault

        mov     eax, [wparam]
        cmp     eax, [hCCList]
        je      .ondefault

        invoke  PostMessage, [hCCList], WM_CLOSE, 0, 0
        jmp     .ondefault

;---------------------------------------------------------------------

onmessage CCM_UPDATECCPOS
        stdcall UpdateCCPosition, [hwnd], [hCCList]
        clc
        ret

;-----------------------------------------------------------------------
onmessage WM_LBUTTONDOWN
onmessage WM_RBUTTONDOWN

        cmp     [hCCList], 0
        je      .ondefault

        invoke  PostMessage, [hCCList], WM_CLOSE, 0, 0
        jmp     .ondefault
;-----------------------------------------------------------------------
onmessage WM_DROPFILES
locals
  .filename rb 1024
  .drop     POINT
  .incl     dd ?
endl

iglobal
  cIncludeCommand db 'include "', 0
  cIncludeEnd     db '"'
  cIncludeEnd2    db $0d, $0a, 0
endg
        lea     eax, [.drop]
        invoke  DragQueryPoint, [wparam], eax

        mov     ebx, [.drop.y]
        shl     ebx, 16
        mov     bx, word [.drop.x]
        invoke  SendMessage, [hwnd], WM_LBUTTONDOWN, 0, ebx
        invoke  SendMessage, [hwnd], WM_LBUTTONUP, 0, ebx

        lea     esi, [.aepos]
        invoke  SendMessage, [hwnd], AEM_GETPOS, esi, 0
        mov     [esi+AEPOS.caretPosition], 1
        mov     [esi+AEPOS.selectionPosition], 1
        invoke  SendMessage, [hwnd], AEM_SETPOS, esi, 0

        invoke  SendMessage, [hwnd], AEM_GETMODE, 0, 0
        push    eax

        invoke  SendMessage, [hwnd], AEM_SETMODE, 0, 0

        lea     esi, [.filename]

        invoke  DragQueryFile, [wparam], -1, esi, 1024
        mov     ebx, eax
        xor     edi, edi

        stdcall StrNew
        mov     [.incl], eax

.opendroploop:
        invoke  DragQueryFile, [wparam], edi, esi, 1024

        stdcall StrCat, [.incl], cIncludeCommand

        stdcall _RelativeToFinc, esi
        stdcall StrCat, [.incl], eax
        stdcall StrDel, eax

        stdcall StrCat, [.incl], cIncludeEnd

        inc     edi
        dec     ebx
        jnz     .opendroploop

        invoke  DragFinish, [wparam]

        stdcall StrCat, [.incl], cIncludeEnd2
        stdcall StrPtr, [.incl]
        invoke  SendMessage, [hwnd], EM_REPLACESEL, TRUE, eax
        stdcall StrDel, [.incl]

        pop     eax
        invoke  SendMessage, [hwnd], AEM_SETMODE, eax, 0

        lea     esi, [.aepos]
        invoke  SendMessage, [hwnd], AEM_GETPOS, esi, 0
        mov     eax, [esi+AEPOS.selectionPosition]
        mov     ecx, [esi+AEPOS.selectionLine]
        mov     [esi+AEPOS.caretPosition], eax
        mov     [esi+AEPOS.caretLine], ecx
        invoke  SendMessage, [hwnd], AEM_SETPOS, esi, 0

        xor     eax, eax
        clc
        ret

;-----------------------------------------------------------------------
onmessage WM_KEYDOWN
        cmp     [hCCList], 0
        jne     .list_opened
        cmp     [wparam], VK_SPACE
        jne     .ondefault
        invoke  GetKeyState, VK_CONTROL
        test    eax, $8000
        jz      .ondefault
        jmp     .open_cc_list
   .list_opened:
        mov     ebx, [wparam]
        call    JumpTo
        MessageList                                      \
          VK_DOWN,   .sendtocc,                          \
          VK_UP,     .sendtocc,                          \
          VK_PRIOR,  .sendtocc,                          \
          VK_NEXT,   .sendtocc,                          \
          VK_HOME,   .sendtocc,                          \
          VK_END,    .sendtocc,                          \
          VK_ESCAPE, .closecc,                           \
          VK_LEFT,   .closekey,                          \
          VK_RIGHT,  .closekey,                          \
          VK_RETURN, .inserttext

        invoke MapVirtualKey, [wparam], 2
        test   eax, eax
        jz     .ondefault
        cmp     al, ' '
        je      .closekey
        push    eax
        cmp     al, '-'
        jne     @f
        invoke  GetKeyState, VK_SHIFT
        test    eax, eax
        jns     @f
        mov     dword [esp], '_'
@@:     pop     eax
        call    recognize_character
        jnc     .endchar
        call    .insert_text_proc
        cmp     [wparam], $be
        jne     .closekey

.endchar:
        invoke  PostMessage, [hwnd], CCM_UPDATECCPOS, 0, 0
        stc
        ret
.closekey:
        invoke  PostMessage, [hCCList], WM_CLOSE, 0, 0
        stc
        ret
.sendtocc:
        invoke  SendMessage, [hCCList], [wmsg], [wparam], [lparam]
        clc
        ret
.inserttext:
        call    .insert_text_proc
.closecc:
        invoke  PostMessage, [hCCList], WM_CLOSE, 0, 0
        clc
        ret

;----------------------------------------------------------------------
.insert_text_proc:
        invoke  SendMessage, [hCCList], LB_GETCURSEL, 0, 0
        cmp     eax, LB_ERR
        je      .finishinsert
        push    eax

        mov     ebx,[lpszReadLine]
        invoke  SendMessage, [hwnd], AEM_GETWORDATCARET,0x1ffff,ebx

        mov     [.aepos.selectionLine], 0
        mov     [.aepos.caretLine], 0
        movzx   ecx, ax
        shr     eax, 16

        inc     ecx
        inc     eax

        mov     [.aepos.selectionPosition], eax
        mov     [.aepos.caretPosition], ecx
        lea     eax, [.aepos]
        invoke  SendMessage, [hwnd], AEM_SETPOS, eax, 0

        pop     eax
        invoke  SendMessage, [hCCList], LB_GETTEXT,eax,ebx
        invoke  SendMessage, [hwnd], EM_REPLACESEL, FALSE, ebx

        lea     ebx, [.aepos]
        invoke  SendMessage, [hwnd], AEM_GETPOS, ebx, 0
        push    [.aepos.selectionPosition] [.aepos.selectionLine]
        pop     [.aepos.caretLine] [.aepos.caretPosition]
        invoke  SendMessage, [hwnd], AEM_SETPOS, ebx, 0

.finishinsert:
        retn


;-------------------------------------------------------------------

onmessage WM_CHAR   ; check for dot
        cmp     [hCCList], 0
        jne     .ondefault
        cmp     [wparam],'.'
        jne     .ondefault
.check_dot:
        lea     eax,[.aepos]
        invoke  SendMessage, [hwnd],AEM_GETPOS,eax,0
        dec     [.aepos.caretPosition]
        jz      .ondefault
        dec     [.aepos.caretPosition]
        jz      .ondefault
        invoke  SendMessage, [hwnd],AEM_GETLINELENGTH,0,0
        cmp     eax,[.aepos.caretPosition]
        jb      .ondefault
        mov     eax,[lpszReadLine]
        invoke  SendMessage, [hwnd],AEM_GETLINE,0,eax
        mov     eax,[lpszReadLine]
        add     eax,[.aepos.caretPosition]
        mov     al,[eax-2]
        call    recognize_character
        jc      .ondefault
.open_cc_list:
        invoke  CreateWindowEx, 0,  cListboxClassName,  0,              \
                WS_POPUP or LBS_HASSTRINGS or WS_VSCROLL or WS_SIZEBOX or LBS_NOINTEGRALHEIGHT or LBS_SORT, \
                0, 0, 300, 120, [hwnd], NULL, [hInstance], 0
        mov     [hCCList], eax
        invoke  GetStockObject, DEFAULT_GUI_FONT
        invoke  SendMessage, [hCCList], WM_SETFONT, eax, TRUE
        stdcall SubclassWindow, [hCCList], CCListProc
        cmp     [ptrLabels], 0
        je      .listboxok
        mov     esi, [ptrLabels]
        mov     ebx, [esi+TArray.Count]
        lea     esi, [esi+TArray.Items]
        test    ebx,ebx
        jz      .listboxok
.addloop:
        mov     eax, [ptrNames]
        add     eax, [esi+TLabel.iName]
        invoke  SendMessage, [hCCList], LB_ADDSTRING, 0, eax
        add     esi, sizeof.TLabel
        dec     ebx
        jnz     .addloop
.listboxok:
        invoke  PostMessage, [hwnd], CCM_UPDATECCPOS, 0, 0
;        stdcall UpdateCCPosition, [hwnd], [hCCList]
        cmp     [wparam], $be
        je      .ondefault
        clc
        ret
endwp


; code completion list window procedure
winproc CCListProc
        mov     ebx, [wmsg]
ondefault
        stc
        ret

onmessage WM_LBUTTONDBLCLK
        invoke  SendMessage, [hEditorsHost], EHM_GETCURRENTASMEDIT, 0, 0
        invoke  PostMessage, eax, WM_KEYDOWN, VK_RETURN, 0
        clc
        ret

onmessage WM_KEYDOWN
        mov     ebx, [wparam]
        call    JumpTo
        MessageList                                     \
          VK_DOWN,   .ondefault,                        \
          VK_UP,     .ondefault,                        \
          VK_PRIOR,  .ondefault,                        \
          VK_NEXT,   .ondefault,                        \
          VK_HOME,   .ondefault,                        \
          VK_END,    .ondefault
        invoke  SendMessage, [hEditorsHost], EHM_GETCURRENTASMEDIT, 0, 0
        invoke  PostMessage, eax, [wmsg], [wparam], [lparam]
        clc
        ret

onmessage WM_CLOSE
        invoke  DestroyWindow, [hwnd]
        mov     [hCCList], 0
        clc
        ret
endwp


proc UpdateCCPosition .aedit, .listbox
locals
.aepos AEPOS
.caret AECARETXY
.rect  RECT
.workrect RECT
.str rb 0x100
endl
        cmp     [.listbox], 0
        je      .finish
        lea     ebx, [.caret]
        invoke  SendMessage, [.aedit], AEM_GETCARETXY, ebx, 0
        invoke  ClientToScreen, [.aedit], ebx
        add     ebx, 8
        invoke  ClientToScreen, [.aedit], ebx
        sub     ebx, 8
        lea     eax, [.rect]
        invoke  GetWindowRect, [.listbox], eax
        mov     eax, [.rect.right]
        mov     ecx, [.rect.bottom]
        sub     eax, [.rect.left]
        sub     ecx, [.rect.top]
        mov     [.rect.right], eax
        mov     [.rect.bottom], ecx
        lea     ebx, [.workrect]
        invoke  SystemParametersInfo, SPI_GETWORKAREA,0,ebx,0
        mov     eax, [.workrect.right]
        sub     eax, [.rect.right]
        mov     ecx, [.caret.x0]
        cmp     eax, ecx
        jle     @f
        mov     eax, ecx
@@:     mov     edx, [.workrect.bottom]
        sub     edx, [.rect.bottom]
        mov     ecx, [.caret.y1]
        add     ecx, 1
        cmp     edx, ecx
        jg      @f
        mov     ecx, [.caret.y0]
        sub     ecx, [.rect.bottom]
        sub     ecx, 2
@@:     invoke  SetWindowPos, [.listbox], NULL, eax, ecx, 0, 0, SWP_NOSIZE or SWP_NOZORDER or SWP_HIDEWINDOW or SWP_NOACTIVATE
        lea     ebx, [.str]
        invoke  SendMessage, [.aedit], AEM_GETWORDATCARET, $100, ebx
        invoke  SendMessage, [hCCList], LB_RESETCONTENT, 0, 0
        stdcall SearchTree, ebx, ptrLabels
        test    eax, eax
        jz      .content_ok
        mov     esi, [eax]
        test    esi, esi
        jz      .content_ok
        mov     ebx, [esi+TArray.Count]
        lea     esi, [esi+TArray.Items]
        test    ebx, ebx
        jz      .content_ok
        mov     edi, edx        ; The last word
 .fillloop:
        mov     eax, [ptrNames]
        add     eax, [esi+TLabel.iName]
        stdcall StringMatch, eax, edi
        jnc     .next
        invoke  SendMessage, [hCCList], LB_ADDSTRING, 0, eax
.next:
        add     esi, sizeof.TLabel
        dec     ebx
        jnz     .fillloop
        stdcall StrPtr, edi
        invoke  SendMessage, [hCCList], LB_SELECTSTRING, -1, eax
        stdcall StrDel, edi
        cmp     eax, LB_ERR
        jne     .content_ok
        invoke  SendMessage, [hCCList], LB_SETCURSEL, 0, 0
.content_ok:
        invoke  SendMessage, [hCCList], LB_GETCOUNT, 0, 0
        test    eax, eax
        jz      .destroycc
        cmp     eax, LB_ERR
        je      .destroycc
        invoke  SetWindowPos, [.listbox], NULL, 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE or SWP_NOZORDER or SWP_SHOWWINDOW or SWP_NOACTIVATE
.finish:
        ret
.destroycc:
        invoke  MessageBeep, MB_ICONASTERISK
        invoke  PostMessage, [hCCList], WM_CLOSE, 0, 0
        ret
endp


proc StringMatch .str1, .str2
        push    esi edi eax
        stdcall StrPtr, [.str1]
        mov     esi, eax
        stdcall StrPtr, [.str2]
        mov     edi, eax
.loop:
        mov     al, [esi]
        inc     esi
        mov     ah, [edi]
        inc     edi
        test    ah,ah
        jz      .match
        test    al,al
        jz      .notmatch
        or      eax, $2020
        cmp     al,ah
        je      .loop
.notmatch:
        clc
        pop     eax edi esi
        ret
.match:
        stc
        pop     eax edi esi
        ret
endp


proc OnGotoDefinition uses esi ebx, wparam, lparam
        invoke  SendMessage, [hEditorsHost], EHM_GETCURRENTASMEDIT, 0, 0
        test    eax, eax
        jz      .finish
        mov     ebx,eax
        mov     esi,[lpszReadLine]
        invoke  SendMessage, ebx, AEM_GETWORDATCARET, 0xffff, esi
        stdcall SearchTreeExact, esi, ptrLabels
        test    eax, eax
        jz      .finish
        stdcall FindAddressPosition, [eax+TLabel.ValueLo]
        test    eax, eax
        jz      .finish
        movzx   ecx, [edx+TDebugLine.LineNum]
        stdcall PosEditor, eax, ecx
.finish:
        ret
endp