proc CreateArray, .ItemSize
begin
        push    ecx edx
        mov     ecx, [.ItemSize]
        imul    ecx, cInitialCapacity
        add     ecx, sizeof.TArray
        push    ecx
        invoke  GetProcessHeap
        invoke  HeapAlloc, eax, HEAP_ZERO_MEMORY

        push    [.ItemSize]
        pop     [eax+TArray.ItemSize]
        mov     [eax+TArray.Capacity], cInitialCapacity
        pop     edx ecx
        return
endp



;*******************************************************
;  proc AddArrayItem - adds new item at the end of TArray
;                      dynamic array.
;
;  Arguments:
;    ptrVarArray - pointer to dword containing pointer
;                  to TArray structure.
;
;  Returns:
;    eax - pointer to new created element.
;    edx - pointer to TArray of the element.
;*******************************************************
proc AddArrayItem, .ptrVarArray
begin
        push    esi edi ebx

        mov     esi, [.ptrVarArray]
        cmp     dword [esi], 0
        jne     @f

; Error creating item.
        xor     eax, eax
        pop     ebx edi esi
        return

@@:
        mov     edi, [esi]

        mov     ebx, [edi+TArray.Count]
        inc     [edi+TArray.Count]

        cmp     ebx, [edi+TArray.Capacity]
        jl      .memallocated

        mov     eax, [edi+TArray.Capacity]
        mov     ecx, eax
        shr     ecx, 2
        add     eax, ecx
        inc     eax
        mov     [edi+TArray.Capacity], eax
        imul    eax, [edi+TArray.ItemSize]
        add     eax, sizeof.TArray
        push    eax
        push    edi

        invoke  GetProcessHeap
        invoke  HeapReAlloc, eax, 0
        mov     [esi], eax
        mov     edi, eax

.memallocated:
        imul    ebx, [edi+TArray.ItemSize]
        lea     eax, [edi+ebx+TArray.Items]
        mov     edx, edi

        pop     ebx edi esi
        return
endp



;**************************************************************
;  proc DeleteArrayItem - deletes the item with the specified
;                         index from TArray dynamic array.
;
;  Arguments:
;    ptrVarArray - pointer to dword containing pointer
;                  to TArray structure.
;    iElement - index of the element to be deleted.
;
;  Returns: Nothing
;*******************************************************
proc DeleteArrayItem, .ptrVarArray, .iElement
begin
        push    esi edi ebx

        mov     esi, [.ptrVarArray]
        mov     esi, [esi]
        test    esi, esi
        jz      .exit           ; empty pointer - exit without error...?

        mov     ebx, [.iElement]
        mov     eax, [esi+TArray.Count]
        cmp     ebx, eax
        jae     .exit                   ; index greater or equal to the count...

        dec     eax
        cmp     ebx, eax
        je      .movedone               ; if deleted element is the last, don't move

        mov     eax, [esi+TArray.ItemSize]
        mov     ecx, [esi+TArray.Count]
        imul    ebx, eax                ; offset of the element deleted
        imul    ecx, eax                ; offset of the end of the array

        lea     edx, [ebx+eax]          ; offset of the next element

        sub     ecx, edx                ; count of the bytes have to be moved.
        test    ecx, 3                  ; check alignment
        jz      @f
        add     ecx, 4
@@:
        shr     ecx, 2                  ; count of dword have to be moved

        push    esi

        lea     edi, [esi+ebx+TArray.Items]
        lea     esi, [esi+edx+TArray.Items]

        rep movsd

        pop     esi

.movedone:
        dec     [esi+TArray.Count]
        mov     eax, [esi+TArray.Capacity]
        sub     eax, [esi+TArray.Count]
        cmp     eax, cMaxReserveDelete
        jbe     .exit

        mov     ebx, [esi+TArray.Count]
        add     ebx, cMinReserveDelete
        mov     [esi+TArray.Capacity], ebx

        imul    ebx, [esi+TArray.ItemSize]
        add     ebx, sizeof.TArray           ; New size of the array...

        invoke  GetProcessHeap
        invoke  HeapReAlloc, eax, 0, esi, ebx
        test    eax, eax
        jz      .exit           ; reallocate fail, but delete is OK

        mov     ecx, [.ptrVarArray]
        mov     dword [ecx], eax        ; update the variable...

.exit:
        pop     ebx edi esi
        return
endp