;************************************************************************************************
;* Fresh.asm
;*
;* This file is main file of the:
;*
;* ..::Fresh::.. - assembler RAD IDE with built-in FASM assembler.
;*
;* Credits:
;*   Tomasz Grysztar (aka Privalov) - for exelent work with FASM - the base for Fresh.
;*
;* The team: (everyone is welcome)
;*    John Found
;*    Fredrik Klasson (aka scientica)
;*    Tommy Lillehagen
;*    Mateusz Tymek (aka decard)
;*    Victor Loh (aka roticv)
;*    Yunus Sina Gulsen (aka VeSCeRa)
;*
;*    Tomasz Grysztar - FASM compiler and AsmEdit control.
;*
;* Contact the authors: http:/board.flatassembler.net/
;* FASM main site:      http://www.flatassembler.net/
;*
;************************************************************************************************
format PE GUI 4.0
stack 0x10020           ; 64K stack for HtmlHelp :/

entry start

include '../include/win32/win32a.inc'

options.ShowUserMessages = 1
options.ShowSkipped = 1
options.FastEnter = 0
options.CheckArguments = 0
options.ShowSizes = 0

include 'version.inc'

include '../include/win32/equates/htmlhelp.inc'

include '../include/libs/msgutils.inc'
include '../include/libs/arrays.inc'
include '../include/libs/strlib.inc'
include '../include/libs/giflib.inc'
include '../include/libs/gui.inc'
include '../include/libs/parents.inc'
include '../include/libs/tform.inc'
include '../include/libs/templates.inc'
include '../include/libs/application.inc'
include '../include/libs/TDataGrid.inc'
include '../include/libs/buttonedit.inc'

include '../include/libs/lists.inc'

include 'asmedit/asmedit.inc'
include 'lister.inc'
include 'dumper.inc'
include 'designtime.inc'
include 'virtfiles.inc'
include 'macroses.inc'
include 'resourceid.inc'
include 'projmanager.inc'
include 'Controls.inc'
include 'editorhost.inc'
include 'msgwindow.inc'
include '../include/libs/splitter.inc'
include 'propgrid.inc'
include 'MainWindow.inc'
include 'fasm/Version.inc'
include 'LabelsList.inc'
include 'debug.inc'
include 'debugger.inc'
include 'dbgserver.inc'
include 'propeditor.inc'
include 'formsDT.inc'
include 'filemanager.inc'
include 'FormEditor.inc'
include 'SizerClass.inc'

section '.code' code readable executable

  include 'asmedit/asmedit.asm'
  include 'asmedit/syntax.asm'

  include 'errors.asm'

  include 'LabelsList.asm'
  include 'fasm.asm'
  include 'virtfiles.asm'

  include 'LabelsExplorer.frm'

  include '../include/libs/parents.asm'
  include '../include/libs/tform.asm'
  include '../include/libs/templates.asm'
  include '../include/libs/application.asm'

  include '../include/libs/giflib.asm'

  include 'Controls.asm'
  include 'SizerClass.asm'

  include '../include/libs/TDataGrid.asm'

  include 'MainWindow.asm'

  include 'ActionEvents.asm'
  include 'actNewFile.asm'
  include 'formutils.asm'
  include 'opendialog.asm'

  include 'propeditor.asm'

  include 'projmanager.asm'

  include 'msgwindow.asm'

  include 'palette.asm'

  include '../include/libs/splitter.asm'
  include '../include/libs/buttonedit.asm'
  include 'propgrid.asm'

  include 'progresswnd.frm'

  include 'findreplace.asm'
  include 'editorhost.asm'
  include 'sourceeditor.asm'
  include 'CCList.asm'
  include 'PAHint.asm'

  include 'fileutils.asm'
  include 'mrulist.asm'

  include 'IDEparams.asm'

  include 'HelpFilesOptions.frm'

  include 'EditPath.frm'
  include 'EditPathsOptions.frm'
  include 'IDEBehaviourOptions.frm'
  include 'IDEOptions.frm'

  include 'formsDT.asm'
  include 'help.asm'

  include 'lister.asm'
  include 'dumper.asm'

  include 'dbgserver.asm'
  include 'debugger.asm'
  include 'debug.asm'

  include 'dialogs.asm'
  include '../include/libs/lists.asm'
  include 'FormEditor.asm'

  include 'filemanager.asm'
  include 'watcheditor.frm'


  include '../include/libs/msgutils.asm'
  include '../include/libs/gui.asm'

  include '../include/libs/memstream.asm'
  include '../include/libs/arrays.asm'
  include '../include/libs/strlib.asm'
  include '../include/libs/scrollbars.asm'

  include 'EDITOptions.frm'
  include 'EditorBehaviour.frm'
  include 'EditorColors.frm'

  include 'frmStyleEditor.frm'
  include 'ProjectOptions.frm'
  include 'GraphicEditor.frm'
  include 'templates.frm'
  include 'ExternalOptions.frm'

; Common initialized globals
iglobal
  ; Standard class names
  cEditClassName          db      'EDIT',0
  cStaticClassName        db      'STATIC', 0
  cButtonClassName        db      'BUTTON', 0
  cComboClassName         db      'COMBOBOX', 0
  cScrollClassName        db      'SCROLLBAR', 0
  cListboxClassName       db      'LISTBOX', 0

  ; some CommCtrl class names
  cTabClassName           db      TABCONTROL_CLASS,0
  cTreeViewClassName      db      TREEVIEW_CLASS, 0
  cStatusBarClassName     db      STATUS_CLASS,0
  cProgressbarClassName   db      PROGRESS_CLASS, 0
  cListviewClassName      db      LISTVIEW_CLASS,0
endg


uglobal
  hInstance       dd 0  ; Instance handle for common use.
  hHeap           dd 0
  hFreshDirectory dd 0  ; string containing Fresh directory.
  ComCtrlsFlags INITCOMMONCONTROLSEX
endg



start:
; Main init sequence
        invoke  GetModuleHandle,0
        mov     [hInstance],eax
        invoke  GetProcessHeap
        mov     [hHeap], eax

; Init ALL common controls
        mov     [ComCtrlsFlags.dwSize], sizeof.INITCOMMONCONTROLSEX
        mov     [ComCtrlsFlags.dwICC], $1fff
        invoke  InitCommonControlsEx, ComCtrlsFlags           InitCommonControlsEx

; call all "initialize" procedures in the included modules...

        InitializeAll

; Startup windows creation
        stdcall InitApplication, resAppIcon, cMainTitle

; Accelerator table creating
        invoke  CreateAcceleratorTable, Accelerators, [Accelerators.Count]

        invoke  SetProp, [hApplication], [propAccelerators], eax
        invoke  SetProp, [hApplication], [propOnIdle], OnIdle

        stdcall InitMainWindow, [hApplication]  ; Create main window.
        stdcall AutoArrangeWindows

; Create the tool palette
        stdcall  OnReloadVCL, 0, [hMainWindow]

; Command line parameters.
        stdcall  ParseCommandLine

        stdcall StrNew
        mov     ebx,eax
        mov     [hFreshDirectory],eax
        stdcall StrSetLength, eax,MAX_PATH
        stdcall StrPtr, ebx
        invoke  GetCurrentDirectory, MAX_PATH, eax

; Main message loop
Run:
        call    ProcessMessages
        jc      .terminate

if defined AppLib
        invoke  SendMessage, [hApplication], AM_ONIDLE, 0, 0
end if
        invoke  WaitMessage
        jmp     Run

.terminate:
        push    eax
        FinalizeAll
        invoke  ExitProcess ; exit code from the stack.



section '.data' data readable writeable
  IncludeAllGlobals


section '.idata' import data readable writeable

        ImportLib       \
            kernel32,   \
            user32,     \
            gdi32,      \
            comctl32,   \
            comdlg32,   \
            shell32,    \
            ole32

section '.res' resource from 'images\freshnew.res' data readable
