cNamesInitialSize = 65536

uglobal
  ptrLabels dd ?        ; ptr to Root items TArray structure.
  ptrNames  dd ?        ; pointer to memory buffer with name strings

  ptrPreprocessed dd ?  ; pointer to memory with preprocessed source.
endg



proc CopyPreprocessedSource, .ptrFAS
begin
        pushad

        cmp     [ptrPreprocessed], 0
        je      .bufferok

        invoke  HeapFree, [hHeap], 0, [ptrPreprocessed]

.bufferok:
        mov     ebx, [.ptrFAS]
        test    ebx, ebx
        jz      .exiterror

        mov     esi, [ebx+TDumpHeader.ofsPreprocessed]
        test    esi, esi
        jz      .exiterror

        add     esi, ebx
        invoke  HeapAlloc, [hHeap], 0, [ebx+TDumpHeader.lenPreprocessed]
        mov     edi, eax
        mov     [ptrPreprocessed], eax

        test    eax, eax
        jz      .exiterror

        cld

        mov     ecx, [ebx+TDumpHeader.lenPreprocessed]
        shr     ecx, 2
        rep movsd
        mov     ecx, [ebx+TDumpHeader.lenPreprocessed]
        and     ecx, 3
        rep movsb

.exitok:
        clc
        popad
        return

.exiterror:
        stc
        popad
        return
endp








proc CreateLabelsTree, .ptrFAS
.symend dd ?
.strend dd ?
.buffer rb 1024
begin
        pushad

        cmp     [ptrLabels], 0
        je      @f
        stdcall DestroyLabelsTree, ptrLabels
@@:

; Allocate memory for label names.
        stdcall CreateMemoryStream, ptrNames, cNamesInitialSize

        mov     esi, [.ptrFAS]
        cmp     esi, 0
        je      .endofsymbols   ; there is no FAS file loaded.

        mov     edi, [esi+TDumpHeader.ofsSymbolTable]
        cmp     edi, 0
        je      .endofsymbols   ; there is no symbols table in the fas file.

        mov     ebx, [esi+TDumpHeader.lenSymbolTable]
        add     edi, esi    ; pointer to the symbol table
        add     ebx, edi    ; pointer to the end of symbol table

.symbol_loop:
        cmp     edi, ebx
        jae     .endofsymbols

; get name
        mov     edx, [edi+TIntLabel.ofsName]
        test    edx, edx
        js      .instrings

; in preprocessed source; pascal string copy
        add     edx, [esi+TDumpHeader.ofsPreprocessed]
        add     edx, esi

        movzx   ecx, byte [edx]
        mov     [.buffer+ecx], 0

.copypas:
        dec     ecx
        js      .nameok

        mov     al, [edx+ecx+1]
        cmp     al, '?'
        je      .nextlabel              ; this is internal generated label.
        mov     [.buffer+ecx], al
        jmp     .copypas


; in strings table; asciiz string;
.instrings:
        and     edx, $7fffffff
        add     edx, [esi+TDumpHeader.ofsStringTable]
        add     edx, esi
        xor     ecx, ecx

.copy:
        mov     al, byte [edx+ecx]
        mov     [.buffer+ecx], al
        inc     ecx

        cmp     al, '?'
        je      .nextlabel

        cmp     al, 0
        jne     .copy

.nameok:
        lea     eax, [.buffer]
        stdcall AddToTree, eax, ptrLabels       ; returns pointer to TLabel structure

        mov     dx, [edi+TIntLabel.Flags]
        mov     [eax+TLabel.flags], dx

        mov     dl, [edi+TIntLabel.DataSize]
        mov     [eax+TLabel.DataSize], dl
        mov     dl, [edi+TIntLabel.type]
        mov     [eax+TLabel.type], dl

        mov     edx, [edi+TIntLabel.SIBEx]
        mov     [eax+TLabel.SIBEx], edx

        mov     edx, dword [edi+TIntLabel.Value]
        mov     [eax+TLabel.ValueLo], edx

        mov     edx, dword [edi+TIntLabel.Value+4]
        mov     [eax+TLabel.ValueHi], edx

        mov     edx, [edi+TIntLabel.ofsDefinition]
        mov     [eax+TLabel.ofsDefinition], edx

.nextlabel:
        add     edi, sizeof.TIntLabel
        jmp     .symbol_loop

.endofsymbols:
        popad
        return
endp


proc AddToNames, .ptrString
begin
        push    esi edi ebx

        stdcall StrLen, [.ptrString]
        lea     eax, [eax+5]
        and     eax, $fffffffc
        stdcall GetStreamRoom, ptrNames, eax

        mov     esi, [.ptrString]

        push    edi
        lea     edi, [ebx+edi]

.copyloop:
        lodsb
        stosb
        test    al, al
        jnz     .copyloop

        pop     eax
        pop     ebx edi esi
        return
endp



;---------------------------------------------------------------
; Returns:
;    eax - The name of the label without the first parent name.
;    edx - First parent.
;---------------------------------------------------------------
proc GetLabelParent, .ptrLabel
begin
        push    esi ebx

        mov     esi, [.ptrLabel]
        push    esi

        cmp     word [esi], '..'        ; Begins with ".."
        jne     .search

        add     esi, 2

.search:
        mov     al, [esi]
        test    al,al
        jz      .notfound
        cmp     al, '.'
        je      .found

        inc     esi
        jmp     .search

.found:
        mov     byte [esi], 0

        lea     eax, [esi+1]
        pop     edx

        pop     ebx esi
        return

.notfound:
        pop     eax
        xor     edx, edx

        pop     ebx esi
        return
endp


proc SearchCreateLabel, .ptrVarArray, .ptrLabel
.hash dd ?
begin
        push    esi edi ebx edx

        stdcall StrHash, [.ptrLabel], 0
        mov     [.hash], eax

        mov     edx, [.ptrVarArray]
        mov     esi, [edx]
        test    esi, esi
        jnz     .arrayok

        stdcall CreateArray, sizeof.TLabel
        mov     edx, [.ptrVarArray]
        mov     [edx],eax
        mov     esi, eax        ; pointer to TArray structure

        stdcall CreateArray, 8          ; Array for search tree.
        mov     [esi+TArray.lParam], eax
        lea     eax, [esi+TArray.lParam]
        stdcall AddArrayItem, eax       ; First empty element of the tree.
        mov     dword [eax], 0
        mov     dword [eax+4], 0

.arrayok:
        mov     edi, [esi+TArray.lParam]
        mov     ecx, 32         ; the depth of the tree
        mov     ebx, TArray.Items

.treeseek:
        xor     eax, eax
        ror     [.hash], 1
        adc     eax, 0
        lea     edx, [ebx+4*eax]

        mov     edx, [edi + edx]        ; eax is offset from the begining of the array.
        test    edx, edx
        jz      .notfound

        mov     ebx, edx
        loop    .treeseek

.found:                         ; the offset of the item is in the ebx register
        mov     eax, [edi+edx]
        lea     eax, [esi + eax]
        pop     edx ebx edi esi
        return


.addloop:
        xor     eax, eax
        ror     [.hash],1
        adc     eax, 0

.notfound:      ; edi - pointer to the tree array,  ebx - offset of the last found element.
                ; eax - 0 or 1 depending of the last bit checked. ecx - remaining bits count.
                ; edx = 0
        push    ecx eax

        lea     edx, [esi+TArray.lParam]
        stdcall AddArrayItem, edx
        mov     dword [eax], 0
        mov     dword [eax+4], 0

        mov     edi, edx                ; new base for the array.
        sub     eax, edx                ; offset of the new element

        xchg    eax, [esp]              ; this is the old eax
        lea     edx, [ebx+ 4*eax]

        pop     ebx
        mov     [edi+edx], ebx          ; the new offset is properly set.

        pop     ecx
        loop    .addloop

        push    eax
        mov     edx, [edi+edx]
        lea     ebx, [edi+edx]  ; address of the last element.

        stdcall AddArrayItem, [.ptrVarArray]

        mov     esi, eax
        sub     eax, edx        ; offset of the TLabel item.

        pop     ecx
        mov     [ebx+4*ecx], eax  ; stored.

        xor     eax, eax
        mov     [esi+TLabel.Children], eax
        mov     [esi+TLabel.type], al
        mov     [esi+TLabel.flags], ax
        mov     [esi+TLabel.ValueLo], eax
        mov     [esi+TLabel.ValueHi], eax
;        mov     [esi+TLabel.ptrIntLabel], eax

; This should be the only place where the .iName and .hashName of TLabel is set.
        stdcall AddToNames, [.ptrLabel]
        mov     [esi+TLabel.iName], eax
        push    [.hash]
        pop     [esi+TLabel.hashName]

        mov     eax, esi
        pop     edx ebx edi esi
        return

endp

;------------------------------------------------------
; Returns: eax - pointer to new created TLabel
;          [hLabelName] content of the string will be
;          destroyed.
;------------------------------------------------------
proc AddToTree, .ptrLabelName, .ptrVarTree
begin
        push    esi edi ebx ecx

        mov     edi, [.ptrVarTree]
        mov     ebx, [.ptrLabelName]

.addloop:
        stdcall GetLabelParent, ebx
        test    edx, edx
        jz      .addhere

        mov     ebx, eax
        stdcall SearchCreateLabel, edi, edx

        lea     edi, [eax+TLabel.Children]
        jmp     .addloop

.addhere:
        stdcall SearchCreateLabel, edi, eax

        pop     ecx ebx edi esi
        return
endp




proc SearchLabel, .ptrVarArray, .hLabel
.hash dd ?
begin
        push    esi edi ebx edx

        mov     edx, [.ptrVarArray]
        mov     esi, [edx]
        test    esi, esi
        jz      .notfound

        stdcall StrHash, [.hLabel], 0
        mov     [.hash], eax

        mov     edi, [esi+TArray.lParam]
        mov     ecx, 32         ; the depth of the tree
        mov     ebx, TArray.Items

.treeseek:
        xor     eax, eax
        ror     [.hash], 1
        adc     eax, 0
        lea     edx, [ebx+4*eax]

        mov     edx, [edi + edx]        ; eax is offset from the begining of the array.
        test    edx, edx
        jz      .notfound

        mov     ebx, edx
        loop    .treeseek

.found:                         ; the offset of the item is in the ebx register
        mov     eax, [edi+edx]
        lea     eax, [esi + eax]

        pop     edx ebx edi esi
        return

.notfound:
        xor     eax, eax
        pop     edx ebx edi esi
        return
endp



proc SearchTree, .hLabelName, .ptrVarTree
begin
        push    edi ebx ecx

        mov     edi, [.ptrVarTree]
        mov     ebx, [.hLabelName]

.searchloop:
        stdcall GetLabelParent, ebx
        test    edx, edx
        jz      .here

        mov     ebx, eax
        stdcall SearchLabel, edi, edx
        test    eax, eax
        jz      .finish

        lea     edi, [eax+TLabel.Children]
        jmp     .searchloop

.here:
        mov     edx, eax
        mov     eax, edi

.finish:
        pop     ecx ebx edi
        return
endp

; returns pointer to TLabel or NULL
proc SearchTreeExact, .hLabelName, .ptrVarTree
begin
        push    edi ebx ecx

        mov     edi, [.ptrVarTree]
        mov     ebx, [.hLabelName]

.searchloop:
        stdcall GetLabelParent, ebx
        test    edx, edx
        jz      .here

        mov     ebx, eax
        stdcall SearchLabel, edi, edx
        test    eax, eax
        jz      .finish

        lea     edi, [eax+TLabel.Children]
        jmp     .searchloop

.here:
        stdcall SearchLabel, edi, eax

.finish:
        pop     ecx ebx edi
        return
endp





proc DestroyLabelsTree, .ptrVarArray
begin
        push    esi ebx

        mov     eax, [.ptrVarArray]
        mov     esi, [eax]              ; pointer to TArray
        test    esi, esi
        jz      .finish

        mov     ebx, [esi+TArray.Count]
        lea     esi, [esi+TArray.Items]

.clearlist:
        cmp     dword [esi], 0
        je      .next
        stdcall DestroyLabelsTree, esi          ; for .Children field of the TLabel
.next:
        add     esi, sizeof.TLabel
        dec     ebx
        jnz     .clearlist

        mov     eax, [.ptrVarArray]
        xchg    ebx, [eax]

        invoke  HeapFree, [hHeap], 0, [ebx+TArray.lParam]
        invoke  HeapFree, [hHeap], 0, ebx

.finish:
        cmp     [ptrNames], 0
        je      .quit

        invoke  HeapFree, [hHeap], 0, [ptrNames]
        mov     [ptrNames], 0

.quit:
        pop     ebx esi
        return
endp






; Doesn't work properly!
; to be fixed.
;********************************
proc ShowNotUsed
.frm TFreshMsg
begin
        push    esi edi ebx

        mov     esi, [ptrLabels]
        test    esi, esi
        jz      .finish

        mov     ebx, [esi+TArray.Count]
        test    ebx, ebx
        jz      .finish

        lea     esi, [esi+TArray.Items]

.valuesloop:
        test    [esi+TLabel.flags], lfCanRedefine
        jnz     .next

        test    [esi+TLabel.flags], lfDefined
        jz      .next

        test    [esi+TLabel.flags], lfUsed
        jnz     .next

        cmp     [esi+TLabel.iName], 0
        je      .next

        stdcall StrNew
        push    eax

        stdcall StrCopy, eax, cHintPrefix
        mov     ecx, [esi+TLabel.iName]
        add     ecx, [ptrNames]
        stdcall StrCat, eax, ecx
        stdcall StrCat, eax, cHintSuffix

        mov     [.frm.type], mtInfo
        mov     [.frm.link.hFileName], 0
        mov     [.frm.hMessage], eax

        lea     eax, [.frm]
        invoke  SendMessage, [frmMsg], MWM_ADDMESSAGE, eax, 0
        stdcall StrDel ; from the stack.

.next:
        add     esi, sizeof.TLabel
        dec     ebx
        jnz     .valuesloop

.finish:
        pop     ebx edi esi
        return
endp

iglobal
  cHintPrefix db 'The label "',0
  cHintSuffix db '" is defined but not used.',0
endg

