uglobal
align 4
  LastAddr      dd ?
  LastPtrLine   dd ?
  LastVirt      db ?

align 4
  ptrLastFilename dd ?
  iLastFilename   dd ?

  ptrDebugInfo  dd ?   ; Dynamic array with TLineInfo elements.
  ptrFileNames  dd ?   ; Fixed size hash table, containing string handles to filenames. ????????
  ptrListing    dd ?   ; Pointer to TMemoryStream structures with the listing.

  AsmLines      dd ?   ; count of the lines assembled to code
  TotalLines    dd ?   ; count of all compiled lines
endg

cInitListingSize = 65536

proc InitDebugInfo
begin
        stdcall ClearDebugInfo

        stdcall CreateArray, sizeof.TDebugLine
        mov     [ptrDebugInfo], eax

        invoke  HeapAlloc, [hHeap], HEAP_ZERO_MEMORY, 32768 * 2
        mov     [ptrFileNames], eax

        stdcall CreateMemoryStream, ptrListing, cInitListingSize
        return
endp



proc  ClearDebugInfo
begin
        stdcall DestroyDebugInfo, ptrDebugInfo
        stdcall DestroyFilesList, ptrFileNames
        return
endp


proc DestroyDebugInfo, .ptrVarArray
begin
        push    esi ebx

        mov     esi, [.ptrVarArray]
        cmp     dword [esi], 0
        je      .finish

        xor     ebx, ebx
        xchg    ebx, [esi]

        test    ebx, ebx
        jz      .debugok

        invoke  HeapFree, [hHeap], 0, ebx

.debugok:
        invoke  HeapFree, [hHeap], 0, [ptrListing]

.finish:
        pop     ebx esi
        return
endp



proc DestroyFilesList, .ptrVarArray
begin
        push    esi ebx

        mov     esi, [.ptrVarArray]
        mov     esi, [esi]
        test    esi, esi
        jz      .finish

        mov     ebx, 32768

.freeloop:
        movzx   eax, word [esi]
        add     esi, 2
        test    eax, eax
        jz      .next
        stdcall StrDel, eax
.next:
        dec     ebx
        jnz     .freeloop

.endloop:
        mov     esi, [.ptrVarArray]
        xchg    ebx, [esi]
        invoke  HeapFree, [hHeap], 0, ebx

.finish:
        pop     ebx esi
        return
endp


proc FileNameSearchAdd, .hFileName
begin
        push    esi edi ebx

        stdcall HashFilename, [.hFileName]
        mov     edi, eax
        and     eax, $7fff
        shr     edi, 17
        xor     edi, eax        ; direct index in the filenames array.

        mov     esi, [ptrFileNames]

        movzx   eax, word [esi+2*edi]
        test    eax, eax
        jnz     .found

        stdcall StrNew
        mov     [esi+2*edi], ax
        stdcall StrCopy, eax, [.hFileName]

.found:
        mov     eax, edi
        pop     ebx edi esi
        return
endp




proc CaptureDebugInfo, .ptrFAS
.asmlen dd ?
begin
        pushad

        stdcall InitDebugInfo

        mov     ebx, [.ptrFAS]
        mov     esi, [ebx+TDumpHeader.ofsAsmDump]
        add     esi, ebx

        mov     eax, [ebx+TDumpHeader.lenAsmDump]
        mov     [.asmlen], eax

        mov     [TotalLines], 0
        mov     [AsmLines], 0

        mov     [ptrLastFilename], 0
        mov     [iLastFilename], 0

.linesloop:
        cmp     [.asmlen], 4
        jbe     .endofdump

        inc     [TotalLines]

        cmp     [esi+TAsmDumpElement.flagVirtual], 0
        jne     .nextline

        cmp     [esi+TAsmDumpElement.extSIB],0
        jne     .nextline

        cmp     [.asmlen], 4+sizeof.TAsmDumpElement       ; last line of the code...???
        ja      .notlast

        mov     eax, dword [esi+TAsmDumpElement.ofsBinary]
        cmp     eax, dword [esi+sizeof.TAsmDumpElement]
        je      .nextline
        jmp     .thisline

.notlast:
        cmp     [esi+TAsmDumpElement.extSIB+sizeof.TAsmDumpElement], 0
        jne     .nextline

        cmp     [esi+TAsmDumpElement.flagVirtual+sizeof.TAsmDumpElement], 0
        jne     .nextline

        mov     eax, dword [esi+TAsmDumpElement.ofsBinary]
        cmp     eax, dword [esi+TAsmDumpElement.ofsBinary+sizeof.TAsmDumpElement]
        je      .nextline

.thisline:
        inc     [AsmLines]

        mov     ebx, [esi+TAsmDumpElement.ofsPreprocessed]
        add     ebx, [ptrPreprocessed]

; process one line, pointed by ebx
        stdcall CreateLineString, ebx
        push    eax

.findbaseline:
        test    byte [ebx+TIntLine.LineNumber+3], $80
        jz      .lineok

        mov     eax, [ebx+TIntLine.ofsGenerator]
        add     eax, [ptrPreprocessed]
        cmp     word [eax+1], '__'      ; is it internal macro definition?
        je      .searchdown

        mov     ebx, [ebx+TIntLine.ofsSourceLine] ; get the line where the macro was invoked.
        add     ebx, [ptrPreprocessed]
        jmp     .findbaseline

.searchdown:
        mov     ebx, [ebx+TIntLine.ofsMacroLine]
        add     ebx, [ptrPreprocessed]
        test    byte [ebx+TIntLine.LineNumber+3], $80
        jnz     .searchdown

.lineok:
        stdcall AddArrayItem, ptrDebugInfo
        mov     edi, eax

        pop     [edi+TDebugLine.strLine]

        mov     eax, dword [esi+TAsmDumpElement.Address]
        mov     edx, dword [esi+TAsmDumpElement.Address+4]
        mov     dword [edi+TDebugLine.Address], eax
        mov     dword [edi+TDebugLine.Address+4], edx

        mov     ecx, dword [ebx+TIntLine.LineNumber] ; line number.
        mov     [edi+TDebugLine.LineNum], ecx

        mov     eax, [iLastFilename]
        mov     ecx, [ebx+TIntLine.ofsGenerator]

        test    ecx, ecx
        jnz     .notmain

        mov     ecx, [input_file]
        sub     ecx, [ptrPreprocessed]

.notmain:
        add     ecx, [ptrPreprocessed]
        cmp     [ptrLastFilename], ecx
        je      .filenameok

        mov     [ptrLastFilename], ecx
        stdcall FileNameSearchAdd, ecx       ; filename, returns index in the filename array.
        mov     [iLastFilename], eax

.filenameok:
        mov     [edi+TDebugLine.iFile], eax

.nextline:
        add     esi, sizeof.TAsmDumpElement
        sub     [.asmlen], sizeof.TAsmDumpElement
        jmp     .linesloop

.endofdump:
        popad
        return
endp








proc CreateLineString, .ptrLine
.char dd ?
begin
        push    esi edi ebx

        stdcall GetStreamRoom, ptrListing, $1000
        push    edi                             ; this will be returned.
        add     edi, ebx

        mov     esi, [.ptrLine]
        lea     esi, [esi+TIntLine.PreprocessedLine]

        mov     [.char], $09

.getloop:
        lodsb
        test    al, al
        jz      .eol
        cmp     al, $3b
        je      .ignore
        cmp     al, $1a         ; prefix of token
        je      .token
        cmp     al, $22         ; quoted text.
        je      .quoted

        stosb
        cmp     al,','
        jne     .getloop
        mov     al,$20
        stosb
        jmp     .getloop

.quoted:
        mov     al, "'"
        stosb

        lodsd   ; length
        mov     ecx, eax
        rep movsb

        mov     al, "'"
        stosb
        jmp     .getloop

.token:
        movzx   ecx, byte [esi]
        inc     esi
        rep movsb               ; copy the token to the output string.

        cmp     byte [esi], '['
        je      .tab
        cmp     byte [esi], $1a
        jne     .getloop
.tab:
        mov     al, byte [.char]
        mov     byte [.char], $20
        stosb
        jmp     .getloop

.ignore:
        mov     al, 0

.eol:
        stosb

        sub     edi, ebx
        mov     [ebx+TMemoryStream.offset], edi

        pop     eax
        pop     ebx edi esi
        return
endp




;--------------------------------------------------------------
; Find the file and line of the source, where some address
; is generate.
; Returns:
;   eax: handle of filename string - 0 means not found
;   edx: pointer to the TDebugLine record.
;--------------------------------------------------------
proc FindAddressPosition, .addr
begin
        push    esi edi ecx

        mov     esi, [ptrDebugInfo]
        mov     edi, [ptrFileNames]

        test    esi, esi
        jz      .notfound

        mov     ecx, [esi+TArray.Count]
        lea     esi, [esi+TArray.Items]

        mov     eax, [.addr]

.search:
        jecxz   .notfound
        cmp     eax, dword [esi+TDebugLine.Address]
        jbe     .found
        add     esi, sizeof.TDebugLine
        dec     ecx
        jmp     .search

.notfound:
        xor     eax, eax
        xor     edx, edx
        jmp     .finish

.found:
        movzx   eax, word [esi+TDebugLine.iFile]
        and     eax, $7fff
        mov     eax, [edi+2*eax]
        mov     edx, esi

.finish:
        pop     ecx edi esi
        return
endp







;--------------------------------------------------
; Positions the editor to the file/line
;--------------------------------------------------
uglobal
  fSetCurrent dd ?
endg


proc PosEditor, .hFileName, .LineNum
.aepos AEPOS
begin
        push    eax ebx ecx edx
        stdcall ConvertPath, [.hFileName]
        push    eax

        stdcall OpenFileByName, eax, NULL, TRUE
        stdcall StrDel ; from the stack.
        test    eax, eax
        jz      .finish

        mov     ebx, [eax+TOpenFile.hEditor]
        mov     [fSetCurrent], ebx

        invoke  SendMessage, ebx, AEM_SETFOCUSLINE, [.LineNum], 0
        invoke  RedrawWindow, ebx, NULL, NULL, RDW_UPDATENOW

        invoke  SetForegroundWindow, ebx

; workaround, because AEM_SETFOCUSLINE do not move the caret.
        mov     [.aepos.selectionPosition], 1
        mov     [.aepos.caretPosition], 1
        mov     eax, [.LineNum]
        mov     [.aepos.selectionLine], eax
        mov     [.aepos.caretLine], eax
        lea     eax, [.aepos]
        invoke  SendMessage, ebx, AEM_SETPOS, eax, 0

; Work around, because of some bug in AsmEdit, that make caret to dissapear, even when
; the control have an input focus.
        invoke  SetFocus, 0
        invoke  SetFocus, ebx

        mov     [fSetCurrent], 0
.finish:
        pop     edx ecx ebx eax
        return
endp



;----------------------------------------------------------
; Returns the address in binary of some line from file
;----------------------------------------------------------
proc FindAddress, .hFilename, .linenum
begin

endp