; This module is responsible for creating/reading/saving of common for
; the whole IDE parameters from the main .ini file.
;

iglobal
  cParamSection db 'IDE settings',0
  cDebugerKey   db 'Debugger',0
  cHelpSection  db 'HelpFiles',0

  cAutoHideKey db 'AutoHideEnable', 0
  cAutoHideTime db 'AutoHideTime', 0

  cErrStr2    db 'ERR!'
              db 0
endg


uglobal
  flagAutoHide dd ?
  timeAutoHide dd ?
  blockAutoHide dd ?
endg



proc CheckParams
begin
        invoke  SendMessage, [frmMsg], MWM_CLEARMESSAGES, 0, 0

; Check ini file consistency...
        stdcall StrPtr, [hIniFileName]
        mov     esi, eax

        stdcall FileExists, esi
        jnz     .doCheckParams

        stdcall FirstTimeInstall       ; try to create .ini file automatically

.doCheckParams:

        stdcall CheckFilesSection, cHelpSection, cHelpFileNotSet, cHelpFileNotExists
        stdcall CheckFileParam, cDebugerKey, cDebuggerNotSet, cDebuggerNotExists

; Load IDE options from the .ini file
        invoke  GetPrivateProfileInt, cParamSection, cAutoHideKey, 0, esi
        mov     [flagAutoHide], eax
        invoke  GetPrivateProfileInt, cParamSection, cAutoHideTime, 5000, esi
        mov     [timeAutoHide], eax

        stdcall RebuildHelpFilesList

        call SetAutoHideTimer
        return

iglobal
  cHelpFileNotExists db "Some of configured help files doesn't exists.",0
  cHelpFileNotSet    db "Not any help file is set in IDE options.",0

  cDebuggerNotExists db "Specified external debugger doesn't exists.",0
  cDebuggerNotSet    db "No external debugger is set in IDE options.",0
endg

endp



proc CheckFileParam, .ptrKeyName, .msgNotSet, .msgNotExists
begin
        stdcall GetParamFromINI, [.ptrKeyName]
        jc      .filenotset

        push    eax
        stdcall FileExists, eax
        stdcall StrDel ; from the stack

        test    eax, eax
        jz      .notexists
        return

.notexists:
        stdcall AddCheckMessage, [.msgNotExists], mtError
        return

.filenotset:
        stdcall AddCheckMessage, [.msgNotSet], mtWarning
        return
endp


proc CheckFilesSection, .ptrSectionName, .msgNotSet, .msgNotExists
.count dd ?
.bad   dd ?
begin
        mov     [.count], 0
        mov     [.bad], 0
        lea     eax, [.count]
        stdcall ReadSectionFromIni, [hIniFileName], [.ptrSectionName], _OnCheckOneFile, eax

        cmp     [.count], 0
        je      .filenotset

        cmp     [.bad], 0
        jne     .notexists

.finish:
        return

.notexists:
        stdcall AddCheckMessage, [.msgNotExists], mtError
        return

.filenotset:
        stdcall AddCheckMessage, [.msgNotSet], mtWarning
        return
endp


proc _OnCheckOneFile, .ptrFileName, .ptrCounts
begin
        mov     esi, [.ptrCounts]
        inc     dword [esi]

        stdcall FileExists, [.ptrFileName]
        jnz     .finish

        inc     dword [esi+4]

.finish:
        clc
        return
endp



;-------------------------------------
; It tries to create proper ini file
; in the case that ini file does not
; exists.
;-------------------------------------
proc FirstTimeInstall
.str rb 1024
.ptr dd ?
begin
        mov     ebx, cPathsTableInc

.findloop:
        mov     ecx, [ebx]
        test    ecx, ecx
        jz      .notfound1

        lea     edi, [.str]
        lea     eax, [.ptr]
        invoke  GetFullPathName, ecx, 1024, edi, eax
        stdcall FileExists, edi
        lea     ebx, [ebx+4]
        test    eax, eax
        jz      .findloop

        mov     eax, [.ptr]
        mov     byte [eax-1], 0
        invoke  WritePrivateProfileString, _section_environment, cIncludeKey, edi, esi

.findmain:
        mov     ebx, cPathsTableMain

.findloop2:
        mov     ecx, [ebx]
        test    ecx, ecx
        jz      .notfound2

        lea     edi, [.str]
        lea     eax, [.ptr]
        invoke  GetFullPathName, ecx, 1024, edi, eax
        stdcall FileExists, edi
        lea     ebx, [ebx+4]
        test    eax, eax
        jz      .findloop2

        mov     eax, [.ptr]
        mov     byte [eax-1], 0
        invoke  WritePrivateProfileString, _section_environment, cFreshKey, edi, esi

        return

.notfound1:
        invoke  MessageBox, [hApplication], cNotFoundMsg1, cNotFoundTtl, MB_ICONWARNING or MB_OK
        jmp     .findmain

.notfound2:
        invoke  MessageBox, [hApplication], cNotFoundMsg2, cNotFoundTtl, MB_ICONWARNING or MB_OK
        return
endp


iglobal
  cPathsTableInc dd cAutoPath1, cAutoPath2, 0
  cAutoPath2  db '..\'
  cAutoPath1  db 'include'
  cFincTestFile db '\finc.txt',0

  cIncludeKey db 'finc',0

  cPathsTableMain dd cAutoPath3, cAutoPath4, 0
  cAutoPath3  db '..\'
  cAutoPath4  db 'Fresh.exe',0

  cFreshKey db 'Fresh',0

  cNotFoundMsg1 db "First time run procedure can't find include path automanically. Try to set it manualy in 'options|IDE params' menu."
                db "You have to use 'finc' name for this path.",0
  cNotFoundMsg2 db "First time run procedure can't find main program directory. Try to set it manually in 'options|IDE params' menu."
                db "You have to use 'Fresh' name for this path.",0
  cNotFoundTtl db 'Auto setup error',0
endg




proc AddCheckMessage, .ptrMessage, .msgType
.msg TFreshMsg
begin
        mov     eax, [.msgType]
        mov     ecx, [.ptrMessage]
        mov     [.msg.type], eax
        mov     [.msg.hMessage], ecx
        mov     [.msg.link.hFileName], 0

        lea     eax, [.msg]
        invoke  SendMessage, [frmMsg], MWM_ADDMESSAGE, eax, TVI_ROOT
        return
endp




;------------------------------------------------
; Parces command line and opens the files.
;------------------------------------------------
proc ParseCommandLine
.num dd ?
begin
        stdcall StrNew
        mov     ebx, eax

        invoke  GetCommandLine
        stdcall StrCopy, ebx, eax
        stdcall StrLen, ebx
        mov     ecx, eax

        stdcall StrPtr, ebx
        mov     edi, eax

        mov     [.num], 0
.foundspc:
        call    .skipspc
        call    .findquoted
        jc      .found

        call    .findnotquoted
        jnc     .finish

.found:
        inc     [.num]
        cmp     [.num], 1
        je      .foundspc
        push    ecx
        stdcall OpenFileByName, esi, NULL, TRUE
        pop     ecx
        jmp     .foundspc

.finish:
        stdcall StrDel, ebx
        return

.findquoted:
        jecxz   .false
        mov     al, '"'
        cmp     [edi], al
        jne     .false

        inc     edi
        dec     ecx
        jecxz   .false

        mov     esi, edi
        repnz scasb
        jnz     .false

        mov     byte [edi-1], 0
        stc
        ret

.false:
        clc
        ret

.findnotquoted:
        jecxz   .false
        mov     al, ' '

        mov     esi, edi
        repnz scasb
        jnz     .ok

        mov     byte [edi-1], 0
.ok:
        stc
        ret

.skipspc:
        cmp     byte [edi], ' '
        jne     .endskip

        inc     edi
        loop    .skipspc
.endskip:
        ret
endp



proc GetParamFromINI, .ptrKeyName
.buff rb 1024
begin
        stdcall StrPtr, [hIniFileName]

        lea     ecx, [.buff]
        invoke  GetPrivateProfileString, cParamSection, [.ptrKeyName], cErrStr2, ecx, 1024, eax
        cmp     dword [.buff], 'ERR!'
        je      .error

        lea     eax, [.buff]
        stdcall StrDup, eax
        clc
        return

.error:
        xor     eax, eax
        stc
        return
endp





proc ReadSectionFromIni, .hIniFile, .hSectionName, .procCallback, .lparam
.ptrmem dd ?
begin
        push    ebx esi edi

        invoke  HeapAlloc, [hHeap], HEAP_ZERO_MEMORY, 65536     ; should be enough
        mov     esi, eax
        mov     [.ptrmem], eax

        stdcall StrPtr, [.hIniFile]
        push    eax
        push    65536
        push    esi
        stdcall StrPtr, [.hSectionName]
        push    eax
        invoke  GetPrivateProfileSection ; parameters from the stack.

.addloop:
        cmp     byte [esi], 0
        je      .endofread

        pushad
        add     esi, 4  ; skip the key name
        stdcall [.procCallback], esi, [.lparam]
        popad
        jc      .endofread

        stdcall StrLen, esi
        lea     esi, [esi+eax+1]
        jmp     .addloop

.endofread:
        pushf
        invoke  HeapFree, [hHeap], 0, [.ptrmem]
        popf
        pop     edi esi ebx
        return
endp