;****************************************
; Creates and manages MRU lists.
; This file was written very quick, so
; it will be commented later. Maybe...
;****************************************

proc CreateNewMRUList, .count
begin
        mov     eax, [.count]
        add     eax, 2          ; Two additional cells for capacity and count.
        shl     eax, 2          ;
        invoke  HeapAlloc, [hHeap], HEAP_ZERO_MEMORY, eax
        pushd   [.count]
        popd    [eax]      ; capacity of the MRU list.
        return
endp


proc AddToMRUList, .ptrList, .hString
begin
        push    esi ebx

        mov     esi, [.ptrList]
        mov     ebx, [esi+4]    ; the count of MRU items.

.search:
        dec     ebx
        js      .notfound
        stdcall StrCompNoCase, [esi+4*ebx+8], [.hString]
        jnc     .search

; Found
        pushd   [esi+4*ebx+8]   ; The found one.
.move:
        inc     ebx
        cmp     ebx, [esi+4]
        je      .save

        pushd   [esi+4*ebx+8]
        popd    [esi+4*ebx+4]
        jmp     .move

.save:
        popd    [esi+4*ebx+4]   ; Move to the top.
        pop     ebx esi
        return

.notfound:
        mov     ebx, [esi+4]    ; Count
        cmp     ebx, [esi]      ; Capacity
        jl      .insertlast

        stdcall RemoveFromMRUList, [.ptrList], 0
        dec     ebx

.insertlast:
        stdcall StrNew
        mov     [esi+4*ebx+8], eax      ; On the last free place
        stdcall StrCopy, eax, [.hString]
        inc     dword [esi+4]

        pop     ebx esi
        return
endp


proc RemoveFromMRUList, .ptrList, .iItem
begin
        push    esi ebx

        mov     esi, [.ptrList]
        mov     ebx, [.iItem]
        cmp     ebx, [esi+4]      ; Count
        jae     .finish

        dec     dword [esi+4]
        stdcall StrDel, [esi+4*ebx+8]

.shift:
        cmp     ebx, [esi+4]
        je      .zero

        pushd   [esi+4*ebx+12]
        popd    [esi+4*ebx+8]
        inc     ebx
        jmp     .shift
.zero:
        mov     dword [esi+4*ebx+8], 0        ; Zero the last element.

.finish:
        pop     ebx esi
        return
endp


proc ValidateMRUList, .ptrList, .ptrCallback
begin
        push    esi ebx
        mov     esi, [.ptrList]
        mov     ebx, [esi+4]    ; count;

.checkloop:
        dec     ebx
        js      .endcheck
        invoke  .ptrCallback, [esi+4*ebx+8]      ; Check from the end to the begin
        test    eax, eax
        jnz     .checkloop
        stdcall RemoveFromMRUList, [.ptrList], ebx
        jmp     .checkloop
.endcheck:
        pop     ebx esi
        return
endp


proc DestroyMRUList, .ptrList
begin
        push    esi ebx
        mov     esi, [.ptrList]
        mov     ebx, [esi+4]      ; Count

.freestrings:
        dec     ebx
        js      .freelist
        stdcall StrDel, [esi+4*ebx+8]
        jmp     .freestrings

.freelist:
        invoke  HeapFree, [hHeap], 0, esi
        pop     ebx esi
        return
endp


proc SaveMRUToIni, .ptrList, .hIniName, .hSection
.ptrini dd ?
begin
        push    ebx esi edi
        mov     esi, [.ptrList]

        stdcall StrPtr, [.hIniName]
        mov     [.ptrini], eax

        stdcall StrPtr, [.hSection]
        mov     edi, eax

        stdcall NumToStr, [esi+4], cMRUKey, 4, ntsUnsigned or ntsZeroTerminated or ntsDec

        ; Write the count key
        invoke  WritePrivateProfileString, edi, cMRUCountKey, cMRUKey+4, [.ptrini]

        xor     ebx, ebx
.itemsloop:
        cmp     ebx, [esi+4]
        jae     .endsave

        stdcall NumToStr, ebx, cMRUKey, 4, ntsUnsigned or ntsZeroTerminated or ntsDec
        stdcall StrPtr, [esi+4*ebx+8]
        invoke  WritePrivateProfileString, edi, cMRUKey, eax, [.ptrini]
        inc     ebx
        jmp     .itemsloop
.endsave:
        pop     edi esi ebx
        return
endp

iglobal
  cMRUCountKey  db 'MRUCount',0    ; max 9999 items.
  cMRUKey       db 'Item.....',0    ;
  cMRUErr       db 'ERR#',0
endg


proc LoadMRUFromIni, .capacity, .hIniName, .hSection
.ptrini dd ?
.str    rb 1024
.count  dd ?
begin
        push    esi edi ebx

        stdcall CreateNewMRUList, [.capacity]
        mov     esi, eax

        stdcall StrPtr, [.hIniName]
        mov     [.ptrini], eax

        stdcall StrPtr, [.hSection]
        mov     edi, eax

        invoke  GetPrivateProfileInt, edi, cMRUCountKey, 0, [.ptrini]
        test    eax, eax
        jz      .finish

        mov     [.count], eax
        xor     ebx, ebx

.loadloop:
        cmp     ebx, [.count]
        jae     .finish

        stdcall NumToStr, ebx, cMRUKey, 4, ntsUnsigned or ntsZeroTerminated or ntsDec

        lea     eax, [.str]
        invoke  GetPrivateProfileString, edi, cMRUKey, cMRUErr, eax, 1024, [.ptrini]
        cmp     dword [.str], 'ERR#'
        je      .next

        lea     eax, [.str]
        stdcall AddToMRUList, esi, eax
.next:
        inc     ebx
        jmp     .loadloop

.finish:
        mov     eax, esi
        pop     ebx edi esi
        return
endp



proc FillMRUMenu, .hMenu, .ptrList, .ItemID, .fromSeparator
.mii  MENUITEMINFO
begin
        push    esi edi ebx

        mov     [.mii.cbSize], sizeof.MENUITEMINFO
        mov     [.mii.fMask], MIIM_TYPE or MIIM_ID

        xor     edi, edi

        invoke  GetMenuItemCount, [.hMenu]
        mov     ebx, eax
        test    ebx, ebx
        jz      .endscan

.scanloop:
        lea     ecx, [.mii]
        invoke  GetMenuItemInfo, [.hMenu], edi, TRUE, ecx
        cmp     [.mii.fType], MF_SEPARATOR
        jne     .next
        dec     [.fromSeparator]
        jz      .delloop
.next:
        inc     edi
        cmp     edi, ebx
        jae     .endscan
        jmp     .scanloop

; found needed separator, delete to the next separator.
.delloop:
        invoke  DeleteMenu, [.hMenu], edi, MF_BYPOSITION
        dec     ebx
        cmp     edi, ebx
        jae     .endscan

        lea     ecx, [.mii]
        invoke  GetMenuItemInfo, [.hMenu], edi, TRUE, ecx
        cmp     [.mii.fType], MF_SEPARATOR
        jne     .delloop

.endscan:
        mov     [.mii.fType], MFT_STRING

        mov     esi, [.ptrList]
        xor     ebx, ebx
        mov     eax, [esi+4]
        add     [.ItemID], eax
        stc                     ; One for the separator.
        adc     [.ItemID], edi
.loop:
        cmp     ebx, [esi+4]
        jae     .endinsert

        stdcall StrPtr, [esi+4*ebx+8]
        mov     [.mii.dwTypeData], eax

        dec     [.ItemID]
        push    [.ItemID]
        pop     [.mii.wID]

        lea     eax, [.mii]
        invoke  InsertMenuItem, [.hMenu], edi, TRUE, eax
        inc     ebx
        jmp     .loop

.endinsert:
; Put separator at the begin.
        mov     [.mii.fMask], MIIM_TYPE
        mov     [.mii.fType], MFT_SEPARATOR
        mov     [.mii.wID], NULL
        lea     eax, [.mii]
        invoke  InsertMenuItem, [.hMenu], edi, TRUE, eax

.finish:
        pop     ebx edi esi
        return
endp